/****************************************************************
 FastMarkovChain.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "FastMarkovChain.H"
#include <math.h>
#include <iostream>
#include <fstream>
#include "tigr++/TigrProteinTrans.H"


extern Alphabet alphabet;



FastMarkovChain::FastMarkovChain(TigrFile &file)
  : alphabetSize(alphabet.getNumElements()),
    state(0),
    transitionMatrix(0,0),
    probabilityVector(0),
    statesPerOrder(0),
    revComp(NULL)
{
  load(file);
}



FastMarkovChain::FastMarkovChain(const TigrString &filename)
  : alphabetSize(alphabet.getNumElements()),
    state(0),
    transitionMatrix(0,0),
    probabilityVector(0),
    statesPerOrder(0),
    revComp(NULL)
{
  load(filename);
}



FastMarkovChain::FastMarkovChain(int order,ContentType contentType)
  : transitionMatrix(0,0),
    probabilityVector(0),
    statesPerOrder(order+1),
    order(order),
    alphabetSize(alphabet.getNumElements()),
    state(0),
    revComp(NULL)
{
  setContentType(contentType);
  computeNumberOfStates();

  probabilityVector.resize(numStates);
  transitionMatrix.resize(numStates,alphabetSize);  
}



void FastMarkovChain::computeNumberOfStates()
{
  numStates=1; // initial state corresponds with no combination
  int numCombinations=1; // allow N^order combinations
  for(int i=0 ;i<=order ; ++i)
    {
      numCombinations*=alphabetSize; // N^order
      statesPerOrder[i]=numCombinations;
      numStates+=numCombinations;
      //cout << numCombinations << " -> " << numStates << endl;
    }
}



bool FastMarkovChain::save(const TigrString &filename)
{
  TigrFile file(filename,"w");
  return save(file);
}



void FastMarkovChain::reset()
{
  state=0;
}



void FastMarkovChain::load(const TigrString &filename)
{
  TigrFile file(filename);
  load(file);
}



void FastMarkovChain::load(TigrFile &file)
{
  loadNonRecurs(file);
  ContentType contentType=getContentType();
  if(contentType!=INTERGENIC)
    {
      revComp=new FastMarkovChain(order,::reverseComplement(contentType));
      revComp->revComp=this;
      revComp->loadNonRecurs(file);
    }
}



void FastMarkovChain::loadNonRecurs(TigrFile &file)
{
  TigrString modelType;
  file >> modelType;
  if(modelType!="FastMC")
    throw TigrString("Attempt to load model of type \"")+modelType+
      "\" into FastMarkovChain from binary file";

  int contentTypeCode;
  file >> order >> contentTypeCode;
  setContentType(static_cast<ContentType>(contentTypeCode));
  setStrand(::getStrand(static_cast<ContentType>(contentTypeCode)));
  statesPerOrder.resize(order+1);
  computeNumberOfStates();
  transitionMatrix.loadBytes(file);
  probabilityVector.loadBytes(file);
}



bool FastMarkovChain::save(TigrFile &file)
{
  if(!saveNonRecurs(file)) return false;
  if(getContentType()!=INTERGENIC)
    {
      reverseComplement();
      return revComp->saveNonRecurs(file);
    }
  return true;
}



bool FastMarkovChain::saveNonRecurs(TigrFile &file)
{
  file << "FastMC" << order << static_cast<int>(getContentType());
  transitionMatrix.saveBytes(file);
  probabilityVector.saveBytes(file);
  return true;
}



double FastMarkovChain::scoreSingleBase(const Sequence &,const TigrString &,
					int index,Symbol s,char)
{
  //cout<<state<<"x"<<alphabet.lookup(s)<<"="<<transitionMatrix[state][s]<<"="<<stateToString(transitionMatrix[state][s])<<endl;
  state=transitionMatrix[state][s];
  return probabilityVector[state];
}



void FastMarkovChain::scoreSingleBase(const Sequence &,const TigrString &,int index,
				      Symbol s,char,double &scorePhase0,
				      double &scorePhase1,double &scorePhase2)
{
  state=transitionMatrix[state][s];
  scorePhase0=scorePhase1=scorePhase2=probabilityVector[state];
}



double FastMarkovChain::scoreSubsequence(const Sequence &seq,const TigrString &str,
					 int begin,int length,int phase)
{
  char c;
  double score=0;
  int end=begin+length;
  for(int pos=begin ; pos<end ; ++pos)
    score+=scoreSingleBase(seq,str,pos,seq[pos],c);
  return score;
}



bool FastMarkovChain::save(ostream &os)
{
  throw "attempt to save FastMarkovChain in non-binary file";
}



void FastMarkovChain::setTrans(int state1,Symbol s,int state2)
{
  transitionMatrix[state1][s]=state2;
}



void FastMarkovChain::setProb(int state,double prob)
{
  probabilityVector[state]=prob;
}



int FastMarkovChain::statesOfOrder(int order)
{
  return statesPerOrder[order];
}



int FastMarkovChain::stringToState(const TigrString &s)
{
  int len=s.length();
  if(len==0) return 0;
  int state=1;
  for(int i=0 ; i<len-1 ; ++i)
    state+=statesOfOrder(i);
  int alphabetSize=alphabet.getNumElements();
  int factor=1;
  for(int i=0 ; i<len ; ++i)
    {
      state+=(alphabet.lookup(s[len-i-1]) * factor);
      factor*=alphabetSize;
    }
  return state;
}



TigrString FastMarkovChain::stateToString(int state)
{
  int alphabetSize=alphabet.getNumElements();
  int order=0;
  int firstStateOfOrder=1;
  for(; 1 ; ++order)
    {
      if(state<firstStateOfOrder) break;
      firstStateOfOrder+=statesOfOrder(order);
    }
  --order;
  firstStateOfOrder=1;
  for(int i=0 ; i<order ; ++i)
    firstStateOfOrder+=statesOfOrder(i);
  state-=firstStateOfOrder;
  TigrString str;
  for(int i=0 ; i<=order ; ++i)
    {
      int factor=static_cast<int>(pow((float)alphabetSize,order-i));
      for(int s=alphabetSize-1 ; s>=0 ; --s)
	{
	  if(state>=s*factor)
	    {
	      str+=alphabet.lookup(s);
	      state-=s*factor;
	      break;
	    }
	}
    }
  return str;
}



TigrString FastMarkovChain::getState()
{
  return stateToString(state)+"("+state+")";
}



void FastMarkovChain::gotoState(const TigrString &stateLabel)
{
  state=stringToState(stateLabel);
}



ContentSensor *FastMarkovChain::reverseComplement()
{
  if(!revComp)
    {
      throw "bad!";

      // ### THIS BLOCK IS OBSOLETE

      revComp=new FastMarkovChain(order,::reverseComplement(getContentType()));
      for(int state=0 ; state<numStates ; ++state)
	{
	  TigrString stateLabel=stateToString(state);
	  TigrString rcStateLabel=
	    TigrProteinTrans::reverseComplement(stateLabel);
	  int rcState=stringToState(rcStateLabel);
	  
	  // Copy probabilities into reverse-complementary states:
	  revComp->probabilityVector[rcState]=probabilityVector[state];
	  
	  // The transition function does NOT change -- this is an
	  // invariant property of our encoding system between state
	  // labels and their integer codes, which is not affected by
	  // reverse complementation:
	  for(Symbol s=0 ; s<alphabetSize ; ++s)
	    revComp->transitionMatrix[state][s]=transitionMatrix[state][s];
	}
    }
  return revComp;
}


