/****************************************************************
 GffPathFromParseGraph.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "GffPathFromParseGraph.H"
#include "tigrscan.H"
#ifdef EXPLICIT_GRAPHS
#include "TigrScan.H"
#include "ParseGraph.H"
#include <iostream>
using namespace std;
#include "tigr++/TigrVectorSorter.H"


GffPathFromParseGraph::GffPathFromParseGraph(TigrScan &tigrscan)
  : tigrscan(tigrscan)
{
  // ctor
}



SignalPtr GffPathFromParseGraph::getSignal(SignalType signalType,
					   int pos,Sequence &seq,
					   const TigrString &seqStr,
					   BOOL &found)
{
  // First, see if the signal is in the parse graph (and therefore 
  // already scored)
  ParseGraph &graph=tigrscan.getParseGraph();
  int index=graph.findSignal(signalType,pos);
  if(index>-1) 
    {
      found=true;
      return graph.getIthVertex(index);
    }

  // Otherwise, we have to invoke the appropriate signal sensor
  // to score this putative signal
  SignalSensor &sensor=tigrscan.getSignalSensor(signalType);
  double logP=sensor.getLogP(seq,seqStr,pos);
  int contextWindowPos=pos-sensor.getConsensusOffset();
  found=false;
  SignalPtr p=new Signal(contextWindowPos,logP,sensor,tigrscan.getGC());
  p->setIndex(-1);
  return p;
}



TigrVector<SignalPtr> *GffPathFromParseGraph::getPathFromGff(
  TigrVector<TigrGffTranscript*> &transcripts,Sequence &seq,
  const TigrString &seqStr,TigrVector<BOOL> &found)
{
  TigrVector<SignalPtr> &path=*new TigrVector<SignalPtr>;
  int numTranscripts=transcripts.size();
  for(int i=0 ; i<numTranscripts ; ++i)
    {
      TigrGffTranscript &transcript=*transcripts[i];
      if(transcript.getSubstrate()!=tigrscan.getSubstrateId()) continue;
      int numExons=transcript.getNumExons();
      for(int j=0 ; j<numExons ; ++j)
	{
	  TigrGffExon &exon=transcript.getIthExon(j);
	  Strand strand=exon.getStrand();
	  ContentType contentType=
	    ::exonTypeToContentType(exon.getExonType(),strand);
	  SignalType leftSignal=::leftSignal(contentType);
	  SignalType rightSignal=::rightSignal(contentType);
	  int begin=exon.getBegin(), end=exon.getEnd(), pos1, pos2;
	  exonCoordsToSigPos(contentType,begin,end,pos1,pos2);
	  found.push_back(false);
	  SignalPtr signal1=
	    getSignal(leftSignal,pos1,seq,seqStr,found[found.size()-1]);
	  path.push_back(signal1);
	  found.push_back(false);
	  SignalPtr signal2=
	    getSignal(rightSignal,pos2,seq,seqStr,found[found.size()-1]);
	  path.push_back(signal2);
	}
    }

  class SignalCmp : public TigrComparator<SignalPtr>
  {
    bool equal(SignalPtr &a,SignalPtr &b) 
      {return a->getConsensusPosition()==b->getConsensusPosition();}
    bool less(SignalPtr &a,SignalPtr &b) 
      {return a->getConsensusPosition()<b->getConsensusPosition();}
    bool greater(SignalPtr &a,SignalPtr &b) 
      {return a->getConsensusPosition()>b->getConsensusPosition();}
  } cmp;
  DirectComparator<BOOL> cmp2;
  TigrVectorSorter<SignalPtr> sorter(path,cmp);
  TigrVectorSorter<BOOL> boolSorter(found,cmp2);
  //sorter.sortAscendInPlace();
  TigrVector<int> *indices=sorter.sortAscendByIndex();
  sorter.applyIndexMap(*indices,path);
  boolSorter.applyIndexMap(*indices,found);
  delete indices;
  return &path;
}



void GffPathFromParseGraph::exonCoordsToSigPos(ContentType contentType,
					       int begin,int end,
					       int &pos1,int &pos2)
{
  switch(contentType)
    {
    case INITIAL_EXON:             // ATG - GT
      pos1=begin; pos2=end; break;
    case INTERNAL_EXON:            // AG - GT
      pos1=begin-2; pos2=end; break;
    case FINAL_EXON:               // AG - TAG
      pos1=begin-2; pos2=end-3; break;
    case SINGLE_EXON:              // ATG - TAG
      pos1=begin; pos2=end-3; break;
    case NEG_INITIAL_EXON:         // NEG_GT - NEG_ATG
      pos1=begin-2; pos2=end-3; break;
    case NEG_INTERNAL_EXON:        // NEG_GT - NEG_AG
      pos1=begin-2; pos2=end; break;
    case NEG_FINAL_EXON:           // NEG_TAG - NEG_AG
      pos1=begin; pos2=end; break;
    case NEG_SINGLE_EXON:          // NEG_TAG - NEG_ATG
      pos1=begin; pos2=end-3; break;
    }
}



TigrVector<int> *GffPathFromParseGraph::getSignalCoordinates(
  TigrVector<TigrGffTranscript*> &transcripts)
{
  TigrVector<int> &path=*new TigrVector<int>;
  int numTranscripts=transcripts.size();
  for(int i=0 ; i<numTranscripts ; ++i)
    {
      TigrGffTranscript &transcript=*transcripts[i];
      int numExons=transcript.getNumExons();
      for(int j=0 ; j<numExons ; ++j)
	{
	  TigrGffExon &exon=transcript.getIthExon(j);
	  Strand strand=exon.getStrand();
	  ContentType contentType=
	    ::exonTypeToContentType(exon.getExonType(),strand);
	  SignalType leftSignal=::leftSignal(contentType);
	  SignalType rightSignal=::rightSignal(contentType);
	  int begin=exon.getBegin(), end=exon.getEnd(), pos1, pos2;
	  exonCoordsToSigPos(contentType,begin,end,pos1,pos2);
	  path.push_back(pos1);
	  path.push_back(pos2);
	}
    }

  DirectComparator<int> cmp;
  TigrVectorSorter<int> sorter(path,cmp);
  sorter.sortAscendInPlace();
  return &path;
}


#endif
