/**************************************************************
IMM.C
bmajoros@tigr.org 1/1/2003

TIGR++ : C++ class template library for bioinformatics

Copyright (c) 2003, The Institute for Genomic Research (TIGR),
Rockville, Maryland, U.S.A.  All rights reserved.
***************************************************************/

// IMM.C

#include "IMM.H"
#include <iostream.h>
#include <fstream.h>
#include "NthOrderStringIterator.H"
#include <math.h>
#include "tigr++/TigrChi2FitTest.H"
#include "tigr++/TigrProteinTrans.H"
#include "exonomy.H"


IMM::IMM(const IMM &other)
  : N(other.N), phase(other.phase), alphabet(other.alphabet),
    alphabetSize(other.alphabetSize), minSampleSize(other.minSampleSize),
    strand(other.strand), VariableLengthSubmodel(IMM_MODEL)
{
  // copy ctor

  TigrVector<TigrStringMap<double>*> &otherModels=
    const_cast<TigrVector<TigrStringMap<double>*> &>(other.models);
  TigrVector<TigrStringMap<int>*> &otherCounts=
    const_cast<TigrVector<TigrStringMap<int>*> &>(other.counts);
  int n=otherModels.size();
  for(int i=0 ; i<n ; ++i)
    {
      models.push_back(new TigrStringMap<double>(*otherModels[i]));
      counts.push_back(new TigrStringMap<int>(*otherCounts[i]));
    }
}



IMM::IMM(Alphabet &a,int N,int minSampleSize,int phase,char strand)
  : alphabet(a), N(N), minSampleSize(minSampleSize), phase(phase),
    strand(strand), alphabetSize(a.getNumElements()), 
    VariableLengthSubmodel(IMM_MODEL)
{
  // ctor
}



IMM::IMM(TigrVector<Sequence*> &v,int N,Alphabet &alphabet,
	 int minSampleSize,int phase)
  : N(N), alphabet(alphabet), alphabetSize(alphabet.getNumElements()),
    minSampleSize(minSampleSize), phase(phase), strand('+'), 
    VariableLengthSubmodel(IMM_MODEL)
{
  // ctor

  buildModels(v);
}



IMM::IMM(const TigrString &filename,Alphabet &alphabet)
  : alphabet(alphabet), alphabetSize(alphabet.getNumElements()),
    strand('+'), VariableLengthSubmodel(IMM_MODEL)
{
  // ctor

  ifstream is(filename.c_str());
  if(!is.good()) throw TigrString("Error opening file ")+filename
		   +" in IMM::IMM()";
  TigrString modelType;
  is >> modelType;
  if(modelType!="IMM")
    throw TigrString("Attempt to load an object of type ")+modelType+
      " into an IMM";
  load(is);
}



IMM::IMM(istream &is,
	 Alphabet &alphabet)
  : alphabet(alphabet), alphabetSize(alphabet.getNumElements()),
    strand('+'), VariableLengthSubmodel(IMM_MODEL)
{
  // ctor

  load(is);
}



IMM::~IMM()
{
  // dtor

  int n=models.size();
  for(int i=0 ; i<n ; ++i)
    delete models[i];
}



void IMM::buildModels(TigrVector<Sequence*> &v)
{
  for(int i=0 ; i<=N ; ++i)
    {
      models.push_back(new TigrStringMap<double>(HASH_TABLE_SIZE));
      counts.push_back(new TigrStringMap<int>(HASH_TABLE_SIZE));
    }

  // Extract all windows up to size N and count the different
  // combinations that actually occur
  TigrVector<Sequence*>::iterator cur=v.begin(), end=v.end();
  for(; cur!=end ; ++cur)
    {
      Sequence &seq=**cur;
      TigrString &str=*seq.toString(alphabet);
      for(int order=0 ; order<=N ; ++order)
	updateCounts(str,order);
      delete &str;
    }

  // Now convert the counts into conditional probabilities
  for(int order=0 ; order<=N ; ++order)
    {
      TigrStringMap<double> &model=*models[order];
      TigrStringMap<int> &count=*counts[order];
      NthOrderStringIterator iterator(order,alphabet);
      while(!iterator.done())
	{
	  // Within this state (history), consider all emissions
	  TigrString history=iterator.getNextString();
	  int sampleSize=0;
	  for(Symbol s=0 ; s<alphabetSize ; ++s)
	    {
	      TigrString ngram=history+alphabet.lookup(s);
	      if(count.isDefined(ngram.c_str(),ngram.length()))
		sampleSize+=count.lookup(ngram.c_str(),ngram.length());
	    }
	  
	  // If sample size is insufficient, remove from the model
	  // (a lower-order model will always be available)
	  if(sampleSize<minSampleSize) undefine(history,count);

	  // Otherwise, normalize counts by sample size to produce a
	  // probability
	  else
	    for(Symbol s=0 ; s<alphabetSize ; ++s)
	      {
		TigrString ngram=history+alphabet.lookup(s);
		if(count.isDefined(ngram.c_str(),ngram.length()))
		  model.lookup(ngram.c_str(),ngram.length())=
		    count.lookup(ngram.c_str(),ngram.length())/
		    (float)sampleSize;
		else
		  model.lookup(ngram.c_str(),ngram.length())=0;
	      }
	}
    }
}



void IMM::undefine(TigrString &history,
		   TigrStringMap<int> &model)
{
  for(Symbol s=0 ; s<alphabetSize ; ++s)
    {
      TigrString ngram=history+alphabet.lookup(s);
      if(model.isDefined(ngram.c_str(),ngram.length())) 
	model.remove(ngram.c_str(),ngram.size());
    }
}



void IMM::updateCounts(TigrString &str,int order)
{
  TigrStringMap<int> &m=*counts[order];
  int windowSize=order+1;
  int len=str.length();

  int firstPos, periodicity;
  if(phase==NO_PHASE) 
    {
      firstPos=0;
      periodicity=1;
    }
  else 
    {
      firstPos=phase;
      periodicity=3;
    }

  for(int pos=firstPos ; pos<len ; pos+=periodicity)
    {
      int begin=pos-windowSize+1;
      if(begin<0) continue;
      TigrString sub=str.substring(begin,windowSize);
      if(m.isDefined(sub.c_str(),sub.length())) 
	++m.lookup(sub.c_str(),sub.length());
      else 
	m.lookup(sub.c_str(),sub.length())=1;
    }
}



double IMM::getLogP(Sequence &s,const char*str,int begin,int len)
{
  double score=0;
  int end=begin+len;
  int zero=(strand=='+' ? begin : end-1);
  for(int pos=begin ; pos<end ; ++pos)
    score+=log(getScore(str,pos,zero));
  return score;
}



double IMM::getScore(const char *str,int position,int zero)
{
  switch(strand)
    {
    case '+':
      {
	int roomForContext=position-zero;
	int maxOrder=(roomForContext>N ? N : roomForContext);
	for(int order=maxOrder ; order>=0 ; --order)
	  {
	    TigrStringMap<double> &model=*models[order];
	    int begin=position-order, len=order+1;
	    if(model.isDefined(str,begin,len))
	      return interpolate(order,str,begin);
	  }
      }

    case '-':
      {
	/*
	  On the minus strand we have to take our contexts from the
	  right (but only because we trained the model that way)
	 */
	int roomForContext=zero-position;
	int maxOrder=(roomForContext>N ? N : roomForContext);
	for(int order=maxOrder ; order>=0 ; --order)
	  {
	    TigrStringMap<double> &model=*models[order];
	    int begin=position, len=order+1;
	    if(model.isDefined(str,begin,len))
	      return interpolate(order,str,begin);
	  }
      }
    }
}



double IMM::interpolate(int order,const char *pStr,int begin)
{
  // ### THIS REALLY NEEDS TO BE OPTIMIZED -- THE DISTRIBUTIONS
  //     SHOULD BE COMPUTED AND STORED AT TRAINING TIME, OR
  //     AFTER THE MODEL IS LOADED

  TigrString str(pStr);
  double score=0, scale=1.0;
  char b=str[begin+order]; // the base we're trying to predict
  for(int ord=order ; ord>0 ; --ord)
  {
    TigrString ordContext=str.substring(order-ord,ord);
    TigrString ordMinusOneContext=str.substring(order-(ord-1),ord-1);
    TigrVector<int> ordCounts, ordMinusOneCounts;
    int ordTotal=0, ordMinusOneTotal=0;
    for(Symbol s=0 ; s<alphabetSize ; ++s)
      {
	char c=alphabet.lookup(s);
	TigrString ordPattern=ordContext+c;
	int ordCount=
	  counts[ord]->lookup(ordPattern.c_str(),ordPattern.length());
	ordTotal+=ordCount;
	ordCounts.push_back(ordCount);
	TigrString ordMinusOnePattern=ordMinusOneContext+c;
	int ordMinusOneCount=
	  counts[ord-1]->lookup(ordMinusOnePattern.c_str(),
				ordMinusOnePattern.length());
	ordMinusOneTotal+=ordMinusOneCount;
	ordMinusOneCounts.push_back(ordMinusOneCount);
      }
    TigrChi2FitTest test(ordCounts,ordMinusOneCounts,chiSquaredTable);
    double p=test.getP();
    double confidence=1-p; // our confidence that the distributions differ
    double samplePortion=// ordTotal/double(ordMinusOneTotal);
      ordTotal>=minSampleSize ? 1 : ordTotal/minSampleSize; 
    if(confidence<0.5) confidence=0.0;
    double lambda=confidence*samplePortion;
    TigrString ordContextPlusB=ordContext+b;
    score+=scale*lambda*models[ord]->lookup(ordContextPlusB.c_str(),
					    ordContextPlusB.length());
    scale*=(1-lambda);
  }
  TigrString justB=b;
  score+=scale*models[0]->lookup(justB.c_str(),justB.length());
  return score;
}



bool IMM::save(ostream &os)
{
  os << "IMM" << endl;
  os << N << "\t" << phase << "\t" << minSampleSize << endl;
  int numModels=models.size();
  os << numModels << endl;
  for(int i=0 ; i<numModels ; ++i)
    {
      TigrStringMap<double> &model=*models[i];
      os << model.size() << endl;
      TigrStringMap<double>::iterator cur=model.begin(), end=model.end();
      for(; cur!=end ; ++cur)
	{
	  const char *first=(*cur).first;
	  os << first << endl << (*cur).second 
	     << endl << counts[i]->lookup(first,strlen(first)) << endl;
	}
    }
  return true;
}



bool IMM::save(const TigrString &filename)
{
  ofstream os(filename.c_str());
  if(!os.good()) throw TigrString("Error creating file ")+filename+
		   "in IMM::save()";
  return save(os);
}



void IMM::load(istream &is)
{
  int numModels, numElements;
  TigrString str, pStr;

  is >> N >> phase >> minSampleSize >> numModels;
  for(int i=0 ; i<numModels ; ++i)
    {
      models.push_back(new TigrStringMap<double>(HASH_TABLE_SIZE));
      TigrStringMap<double> &model=*models[i];
      is >> numElements;
      for(int j=0 ; j<numElements ; ++j)
	{
	  is >> str >> pStr >> counts[i]->lookup(str.c_str(),str.length());
	  model.lookup(str.c_str(),str.length())=pStr.asDouble();
	}
    }
}



void IMM::describe(ostream &os)
{
  os << "IMM (Interpolated Markov Model):"
     << "\n\tmax order: " << N 
     << "\n\tmin sample size: " << minSampleSize 
     << endl;
}



void IMM::evaluate(TigrVector<ParseEdge*> &edges,
		   Sequence &sequence,const char *str)
{
  int numEdges=edges.size();
  ParseEdge &leftmostEdge=*edges[numEdges-1], &rightmostEdge=*edges[0];
  int leftmostPos=leftmostEdge.getBeginCoordinate();
  int rightmostPos=leftmostPos+leftmostEdge.getLength()-1;
  int nextEdgeIndex=0, nextBoundary=rightmostEdge.getBeginCoordinate();
  double logP=0;
  int pos=rightmostPos;
  ParseEdge *edge=&rightmostEdge;


  throw "IMM::evaluateInFrame() not implemented";
}



Submodel *IMM::reverseComplement()
{
  IMM *other=
    new IMM(alphabet,N,minSampleSize,phase,complement(strand));

  for(int order=0 ; order<=N ; ++order)
    {
      TigrStringMap<double> &model=*models[order];
      TigrStringMap<int> &count=*counts[order];
      TigrStringMap<double> *otherModel=new TigrStringMap<double>(HASH_TABLE_SIZE);
      other->models.push_back(otherModel);
      TigrStringMap<int> *otherCount=new TigrStringMap<int>(HASH_TABLE_SIZE);
      other->counts.push_back(otherCount);
      TigrStringMap<double>::iterator cur=model.begin(), end=model.end();
      for(; cur!=end ; ++cur)
	{
	  TigrStringMap<double>::ElemType &e=*cur;
	  TigrString revComp=TigrProteinTrans::reverseComplement(e.first);
	  otherModel->lookup(revComp.c_str(),0,revComp.length())=e.second;
	  otherCount->lookup(revComp.c_str(),0,revComp.length())=
	    count.lookup(revComp.c_str(),0,revComp.length());
	}
    }

  return other;
}



Submodel *IMM::clone() const
{
  return new IMM(*this);
}


