/****************************************************************
 IMM.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "IMM.H"
#include <iostream>
#include <fstream>
#include <math.h>
#include "tigr++/TigrProteinTrans.H"
#include "tigr++/Constants.H"
#include "tigr++/Alphabet.H"
#include "tigr++/TigrStacktrace.H"
#include "tigr++/TigrChi2FitTest.H"
#include "NthOrderStringIterator.H"

extern Alphabet alphabet;


inline int hashTableSize(int order)
{
  // these are all primes

  switch(order)
    {
    case 0: return 11; break;
    case 1: return 29; break;
    case 2: return 127; break;
    case 3: return 619; break;
    default:
    case 4: return 3121; break;
    }
}



IMM::IMM(const IMM &other)
  : N(other.N), phase(other.phase), alphabetSize(other.alphabetSize),
    revComp(NULL), models(new TigrVector<TigrStringMap<double>*>)
{
  for(int i=0 ; i<=N ; ++i)
    models->push_back(new TigrStringMap<double>(*(*other.models)[i]));
  setContentType(other.getContentType());
  setStrand(other.getStrand());
  if(getContentType()==INTERGENIC)
    revComp=this;
  else if(other.revComp && getStrand()==FORWARD_STRAND)
    revComp=new IMM(*other.revComp);
}



IMM::IMM(TigrVector<TrainingSequence*> &v,int order,
	 int minSampleSize,int phase,ContentType contentType,
	 Strand strand)
  : N(order),
    alphabetSize(alphabet.getNumElements()),
    phase(phase),
    revComp(NULL),
    models(new TigrVector<TigrStringMap<double>*>)
{
  setContentType(contentType);
  if(strand==EITHER_STRAND) strand=::getStrand(contentType);
  setStrand(strand);

  buildModels(v,minSampleSize);

  if(strand==FORWARD_STRAND)
    {
      TigrVector<TrainingSequence*> rcSeqs;
      revCompSeqs(v,rcSeqs);
      revComp=new IMM(rcSeqs,order,minSampleSize,phase,
		      ::reverseComplement(contentType),
		      REVERSE_STRAND);
      revComp->revComp=this;
    }
}



IMM::IMM(istream &is,Strand strand)
  : revComp(NULL), models(new TigrVector<TigrStringMap<double>*>)
{
  setStrand(strand);
  load(is);
}



IMM::IMM(const TigrString &filename)
  : revComp(NULL), models(new TigrVector<TigrStringMap<double>*>)
{
  ifstream is(filename.c_str());
  if(!is.good()) throw TigrString("Error opening file ")+filename
		   +" in IMM::IMM()";
  TigrString modelType;
  is >> modelType;
  if(modelType!="IMM")
    throw TigrString("Attempt to load an object of type ")+modelType+
      " into an IMM";
  load(is);
}



IMM::~IMM()
{
  int n=models->size();
  for(int i=0 ; i<n ; ++i)
    delete (*models)[i];
  delete models;

  /* ### THIS IS BAD!  THE REVCOMP WILL BE DELETED MULTIPLE TIMES...
         (UNLESS FORWARD_STRAND_ONLY IS DEFINED)
  if(getStrand()==FORWARD_STRAND && revComp!=this)
    delete revComp;
    */
}



double IMM::scoreSubsequence(const Sequence &seq,const TigrString &str,
				     int begin,int length,int)
{
  // This is slow -- please use FastIMM if you care about speed...

  Symbol dummySymbol;
  char dummyChar;
  double score=0;
  int end=begin+length;
  for(int pos=begin ; pos<end ; ++pos)
    score+=scoreSingleBase(seq,str,pos,dummySymbol,dummyChar);
  return score;
}



double IMM::scoreSingleBase(const Sequence &seq,const TigrString &str,
				    int index,Symbol s,char c)
{
  const char *p=str.c_str();
  switch(getStrand())
    {
    case PLUS_STRAND:
      {
	int maxOrder=(index>N ? N : index);
	for(int order=maxOrder ; order>=0 ; --order)
	  {
	    TigrStringMap<double> &model=*(*models)[order];
	    if(model.isDefined(p,index-order,order+1))
	      {
		double d=model.lookup(p,index-order,order+1);
		//cout<<"\t+"<<str.substring(index-order,order+1)<<endl;
		return d;
	      }
	  }
	throw TigrString("IMM::scoreSingleBase('+',")+
	  index+",strlen="+strlen(p)+",str="+
	  str.substring(index,maxOrder)+")";
      }

    case MINUS_STRAND:
      {
	/*
	  On the minus strand we have to take our contexts from the
	  right (but only because we trained the model that way)
	 */
	int seqLen=seq.getLength();
	int maxOrder=seqLen-index-1;
	if(maxOrder>N) maxOrder=N;
	for(int order=maxOrder ; order>=0 ; --order)
	  {
	    TigrStringMap<double> &model=*(*models)[order];
	    if(model.isDefined(p,index,order+1)) 
	      {
		double d=model.lookup(p,index,order+1);
		//cout<<"\t-"<<str.substring(index,order+1)<<endl;
		return d;
	      }
	  }
	throw TigrStacktrace(
          TigrString("IMM::scoreSingleBase('-',")+
	    index+",strlen="+strlen(p)+",str="+
	  str.substring(index,maxOrder)+")");
      }

    default: throw TigrString(__FILE__)+__LINE__;
    }
}



void IMM::scoreSingleBase(const Sequence &seq,const TigrString &str,
				  int index,Symbol s,char c,
				  double &scorePhase0,double &scorePhase1,
				  double &scorePhase2)
{
  // I'm not a 3-periodic markov chain -- I'm just a plain, ordinary,
  // homogeneous markov chain:
  scorePhase0=scorePhase1=scorePhase2=scoreSingleBase(seq,str,index,s,c);
}



ContentSensor *IMM::reverseComplement()
{
  return revComp;
}



bool IMM::save(const TigrString &filename)
{
  ofstream os(filename.c_str());
  if(!os.good()) throw TigrString("Error creating file ")+filename+
		   "in IMM::save()";
  return save(os);
}



bool IMM::save(ostream &os)
{
  os.precision(8);
  os << "IMM" << endl;
  os << getContentType() << endl;
  os << N << "\t" << phase << "\t" << endl;
  int numModels=models->size();
  os << numModels << endl;
  for(int i=0 ; i<numModels ; ++i)
    {
      TigrStringMap<double> &model=*(*models)[i];
      os << model.size() << endl;
      TigrStringMap<double>::iterator cur=model.begin(), end=model.end();
      for(; cur!=end ; ++cur)
	os << (*cur).first << endl << (*cur).second << endl;
    }
  if(getStrand()==FORWARD_STRAND) 
    revComp->save(os);
  return true;
}



void IMM::load(istream &is)
{
  int numModels, numElements;
  TigrString str, pStr;
  ContentType contentType;
  
  is >> contentType >> N >> phase >> numModels;
  setContentType(contentType);

  for(int i=0 ; i<numModels ; ++i)
    {
      models->push_back(new TigrStringMap<double>(hashTableSize(N)));
      TigrStringMap<double> &model=*(*models)[i];
      is >> numElements;
      for(int j=0 ; j<numElements ; ++j)
	{
	  is >> str >> pStr;
	  model.lookup(str.c_str(),str.length())=pStr.asDouble();
	}
    }

  if(getStrand()==FORWARD_STRAND)
    {
      TigrString modelType;
      is >> modelType;
      revComp=new IMM(is,REVERSE_STRAND);
      revComp->revComp=this;
    }
}



void IMM::buildModels(TigrVector<TrainingSequence*> &v,
			      int minSampleSize)
{
  /*
    This is the training procedure for Markov chains.
   */

  // Instantiate a separate hash table for each order
  counts=new TigrVector<TigrStringMap<int>*>;
  for(int i=0 ; i<=N ; ++i)
    {
      int size=hashTableSize(N);
      models->push_back(new TigrStringMap<double>(size));
      counts->push_back(new TigrStringMap<int>(size));
    }

  // Install pseudocounts
  for(int order=0 ; order<=N ; ++order)
    {
      TigrStringMap<int> &count=*(*counts)[order];
      NthOrderStringIterator iterator(order+1,alphabet);
      while(!iterator.done())
	{
	  TigrString ngram=iterator.getNextString();
	  count.lookup(ngram.c_str(),ngram.length())=1;
	}
    }

  // Extract all windows up to size N and count the different
  // combinations that actually occur
  TigrVector<TrainingSequence*>::iterator cur=v.begin(), end=v.end();
  for(; cur!=end ; ++cur)
    {
      TrainingSequence &seq=**cur;
      TigrString &str=*seq.toString(alphabet);
      for(int order=0 ; order<=N ; ++order)
	if(getStrand()==FORWARD_STRAND) 
	  updateCounts_fw(str,order,seq.getPhase(),seq.getBoostCount());
	else
	  updateCounts_rev(str,order,seq.getPhase(),seq.getBoostCount());
      delete &str;
    }
  
  // Now convert the counts into conditional probabilities
  if(getStrand()==FORWARD_STRAND) 
    {
      computeProbabilities_fw(minSampleSize);
      interpolate_fw();
    }
  else
    {
      computeProbabilities_rev(minSampleSize);
      interpolate_rev();
    }
  delete counts;
}



void IMM::computeProbabilities_fw(int minSampleSize)
{
  for(int order=0 ; order<=N ; ++order)
    {
      TigrStringMap<double> &model=*(*models)[order];
      TigrStringMap<int> &count=*(*counts)[order];
      NthOrderStringIterator iterator(order,alphabet);
      while(!iterator.done())
	{
	  // Within this state (history), consider all emissions
	  TigrString history=iterator.getNextString();
	  int sampleSize=0;
	  for(Symbol s=0 ; s<alphabetSize ; ++s)
	    {
	      TigrString ngram=history+alphabet.lookup(s);
	      if(count.isDefined(ngram.c_str(),ngram.length()))
		sampleSize+=int(count.lookup(ngram.c_str(),ngram.length()));
	    }
	  
	  // If sample size is insufficient, remove from the model
	  // (a lower-order model will always be available)
	  if(sampleSize<minSampleSize)
	    undefine_fw(history,count);

	  // Otherwise, normalize counts by sample size to produce a
	  // probability
	  else
	    for(Symbol s=0 ; s<alphabetSize ; ++s)
	      {
		TigrString ngram=history+alphabet.lookup(s);
		if(count.isDefined(ngram.c_str(),ngram.length()))
		  {
		    model.lookup(ngram.c_str(),ngram.length())=
		      log(count.lookup(ngram.c_str(),ngram.length())/
			  (double)sampleSize);
		  }
		else // ### this might be dangerous...
		  model.lookup(ngram.c_str(),ngram.length())=
		    NEGATIVE_INFINITY;
	      }
	}
    }
}



void IMM::computeProbabilities_rev(int minSampleSize)
{
  for(int order=0 ; order<=N ; ++order)
    {
      TigrStringMap<double> &model=*(*models)[order];
      TigrStringMap<int> &count=*(*counts)[order];
      NthOrderStringIterator iterator(order,alphabet);
      while(!iterator.done())
	{
	  // Within this state (future), consider all emissions
	  TigrString future=iterator.getNextString();
	  int sampleSize=0;
	  for(Symbol s=0 ; s<alphabetSize ; ++s)
	    {
	      TigrString ngram=alphabet.lookup(s)+future;
	      if(count.isDefined(ngram.c_str(),ngram.length()))
		sampleSize+=int(count.lookup(ngram.c_str(),ngram.length()));
	    }
	  
	  // If sample size is insufficient, remove from the model
	  // (a lower-order model will always be available)
	  if(sampleSize<minSampleSize)
	    undefine_rev(future,count);

	  // Otherwise, normalize counts by sample size to produce a
	  // probability
	  else
	    for(Symbol s=0 ; s<alphabetSize ; ++s)
	      {
		TigrString ngram=alphabet.lookup(s)+future;
		if(count.isDefined(ngram.c_str(),ngram.length()))
		  {
		    model.lookup(ngram.c_str(),ngram.length())=
		      log(count.lookup(ngram.c_str(),ngram.length())
			  /(double)sampleSize);
		  }
		else // ### this might be dangerous...
		  model.lookup(ngram.c_str(),ngram.length())=
		    NEGATIVE_INFINITY;
	      }
	}
    }
}



void IMM::undefine_fw(TigrString &history,
		      TigrStringMap<int> &model)
{
  /*
    This method removes from the given model all n-grams consisting
    of the given history followed by a single base.
   */

  for(Symbol s=0 ; s<alphabetSize ; ++s)
    {
      TigrString ngram=history+alphabet.lookup(s);
      if(model.isDefined(ngram.c_str(),ngram.length())) 
	model.remove(ngram.c_str(),ngram.length());
    }
}



void IMM::undefine_rev(TigrString &future,
		       TigrStringMap<int> &model)
{
  /*
    This method removes from the given model all n-grams consisting
    of the given future preceded by a single base.
   */

  for(Symbol s=0 ; s<alphabetSize ; ++s)
    {
      TigrString ngram=alphabet.lookup(s)+future;
      if(model.isDefined(ngram.c_str(),ngram.length())) 
	model.remove(ngram.c_str(),ngram.length());
    }
}



void IMM::updateCounts_fw(TigrString &str,int order,int seqPhase,
				  int boostCount)
{
  /*
    This method simply counts all the n-grams of length order+1.
   */

  TigrStringMap<int> &count=*(*counts)[order];
  int windowSize=order+1;
  int len=str.length();

  int firstPos, periodicity;
  if(phase==NO_PHASE) 
    {
      firstPos=0;
      periodicity=1;
    }
  else 
    {
      if(seqPhase==-1) 
	throw TigrString("seq phase=-1 in IMM::updateCounts(), "
	  "called as part of ThreePeriodicIMM:\n")+str;
      firstPos=(phase-seqPhase+3)%3;
      periodicity=3;
    }

  for(int pos=firstPos ; pos<len ; pos+=periodicity)
    {
      int begin=pos-order;
      if(begin<0) continue;
      if(count.isDefined(str.c_str(),begin,windowSize)) 
	count.lookup(str.c_str(),begin,windowSize)+=boostCount;
      else 
	count.lookup(str.c_str(),begin,windowSize)=boostCount;
    }
}



void IMM::updateCounts_rev(TigrString &str,int order,int seqPhase,
				   int boostCount)
{
  /*
    This method simply counts all the n-grams of length order+1.
   */

  TigrStringMap<int> &count=*(*counts)[order];
  int windowSize=order+1;
  int len=str.length();

  int firstPos, periodicity;
  if(phase==NO_PHASE) 
    {
      firstPos=0;
      periodicity=1;
    }
  else 
    {
      if(seqPhase==-1) 
	throw TigrString("seq phase=-1 in IMM::updateCounts(), "
	  "called as part of ThreePeriodicIMM (-):\n")+str;
      firstPos=(seqPhase-phase+3)%3;
      periodicity=3;
    }

  for(int pos=firstPos ; pos+order<len ; pos+=periodicity)
    {
      if(count.isDefined(str.c_str(),pos,windowSize)) 
	count.lookup(str.c_str(),pos,windowSize)+=boostCount;
      else 
	count.lookup(str.c_str(),pos,windowSize)=boostCount;
    }
}



void IMM::revCompSeqs(TigrVector<TrainingSequence*> &forwardSeqs,
		      TigrVector<TrainingSequence*> &revSeqs)
{
  TigrVector<TrainingSequence*>::iterator cur=forwardSeqs.begin(), 
    end=forwardSeqs.end();
  for(; cur!=end ; ++cur)
    revSeqs.push_back(
      static_cast<TrainingSequence*>((*cur)->reverseComplement(alphabet)));
}



void IMM::useLogOdds_anonymous(ContentSensor &nullModel)
{
  Strand strand=getStrand();
  for(int order=0 ; order<=N ; ++order)
    {
      TigrStringMap<double> *model=(*models)[order];
      TigrStringMap<double>::iterator cur=model->begin(),
	end=model->end();
      for(; cur!=end ; ++cur)
	{
	  const char *p=(*cur).first;
	  int len=strlen(p);
	  Sequence seq(p,alphabet);
	  int targetPos=(strand==FORWARD_STRAND ? len-1 : 0);
	  double nullScore=nullModel.scoreSingleBase(seq,p,targetPos,
						     seq[targetPos],
						     p[targetPos]);
	  cout << "="<<nullScore<<" vs "<<(*cur).second <<p<<endl;
	  (*cur).second-=nullScore;
	}
    }

  if(strand==FORWARD_STRAND && revComp && revComp!=this)
    revComp->useLogOdds_anonymous(*nullModel.reverseComplement());
}



void IMM::useLogOdds(ContentSensor &nullModel)
{
  IMM &other=dynamic_cast<IMM&>(nullModel);
  for(int order=0 ; order<=N ; ++order)
    {
      TigrStringMap<double> *model=(*models)[order];
      TigrStringMap<double> *otherModel=(*other.models)[order];
      TigrStringMap<double>::iterator cur=model->begin(),
	end=model->end();
      for(; cur!=end ; ++cur)
	{
	  const char *p=(*cur).first;
	  int len=strlen(p);
	  (*cur).second-=otherModel->lookup(p,0,len);
	}
    }

  if(getStrand()==FORWARD_STRAND && revComp && revComp!=this)
    revComp->useLogOdds(*nullModel.reverseComplement());
}



void IMM::interpolate_fw()
{
  TigrVector<TigrStringMap<double>*> interpolated;
  interpolated.push_back((*models)[0]);
  
  for(int order=1 ; order<=N ; ++order)
    {
      TigrStringMap<double> *newModel=
	new TigrStringMap<double>(hashTableSize(order));
      interpolated.push_back(newModel);
      TigrStringMap<double> *prevIMM=interpolated[order-1];
      TigrStringMap<double> *model=(*models)[order];
      TigrStringMap<int> *count=(*counts)[order];
      TigrStringMap<int> *smallerCount=(*counts)[order-1];
      NthOrderStringIterator iterator(order,alphabet);
      while(!iterator.done())
	{
	  TigrString longHistory=iterator.getNextString();
	  TigrString shortHistory=longHistory.substr(1,order-1);
	  TigrVector<int> longCounts, shortCounts; // distributions
	  int longTotal=0;
	  for(Symbol s=0 ; s<alphabetSize ; ++s)
	    {
	      char c=alphabet.lookup(s);
	      TigrString ngram=longHistory+c;
	      const char *p=ngram.c_str(); int len=ngram.length();
	      if(!count->isDefined(p,len)) continue;
	      int longCount=count->lookup(p,len);
	      longCounts.push_back(longCount);
	      longTotal+=longCount;
	      ngram=shortHistory+c;
	      int shortCount=
		smallerCount->lookup(ngram.c_str(),ngram.length());
	      shortCounts.push_back(shortCount);
	    }

	  //TigrChi2FitTest test(longCounts,shortCounts,chiSquaredTable);
	  //double p=test.getP();
	  double p=chiTest(longCounts,shortCounts);

	  double confidence=1-p;// confidence that the distributions differ
	  if(confidence<0.5) confidence=0.0;
	  double samplePortion=longTotal/double(400);
	  double lambda=
	    (samplePortion>=1 ? 1.0 : confidence*samplePortion);
	  for(Symbol s=0 ; s<alphabetSize ; ++s)
	    {
	      char c=alphabet.lookup(s);
	      TigrString ngram=longHistory+c;
	      const char *p=ngram.c_str(); int len=ngram.length();
	      if(!count->isDefined(p,len)) continue;
	      newModel->lookup(p,len)=
		lambda * model->lookup(p,len) +
		(1-lambda) * prevIMM->lookup(p+1,len-1);
	    }
	}
    }
  for(int i=1 ; i<=N ; ++i) delete (*models)[i];
  *models=interpolated;
}



double IMM::chiTest(TigrVector<int> &counts1,TigrVector<int> &counts2)
{
  int sum1=0, sum2=0, n1=counts1.size(), n2=counts2.size();
  for(int i=0 ; i<n1 ; ++i) sum1+=counts1[i];
  for(int i=0 ; i<n2 ; ++i) sum2+=counts2[i];
  TigrVector<int> expectedCounts;
  double denom=sum2;
  for(int i=0 ; i<n2 ; ++i) 
    expectedCounts.push_back(int(counts2[i]/denom*sum1+0.5));
  TigrChi2FitTest test(counts1,expectedCounts,chiSquaredTable);
  return test.getP();
}



void IMM::interpolate_rev()
{
  TigrVector<TigrStringMap<double>*> interpolated;
  interpolated.push_back((*models)[0]);
  
  for(int order=1 ; order<=N ; ++order)
    {
      TigrStringMap<double> *newModel=
	new TigrStringMap<double>(hashTableSize(order));
      interpolated.push_back(newModel);
      TigrStringMap<double> *prevIMM=interpolated[order-1];
      TigrStringMap<double> *model=(*models)[order];
      TigrStringMap<int> *count=(*counts)[order];
      TigrStringMap<int> *smallerCount=(*counts)[order-1];
      NthOrderStringIterator iterator(order,alphabet);
      while(!iterator.done())
	{
	  TigrString longFuture=iterator.getNextString();
	  TigrString shortFuture=longFuture.substr(0,order-1);
	  TigrVector<int> longCounts, shortCounts; // distributions
	  int longTotal=0;
	  for(Symbol s=0 ; s<alphabetSize ; ++s)
	    {
	      char c=alphabet.lookup(s);
	      TigrString ngram=c+longFuture;
	      const char *p=ngram.c_str(); int len=ngram.length();
	      if(!count->isDefined(p,len)) continue;
	      int longCount=count->lookup(p,len);
	      longCounts.push_back(longCount);
	      longTotal+=longCount;
	      ngram=c+shortFuture;
	      int shortCount=
		smallerCount->lookup(ngram.c_str(),ngram.length());
	      shortCounts.push_back(shortCount);
	    }
	  
	  //TigrChi2FitTest test(longCounts,shortCounts,chiSquaredTable);
	  //double p=test.getP();
	  double p=chiTest(longCounts,shortCounts);

	  double confidence=1-p;// confidence that the distributions differ
	  if(confidence<0.5) confidence=0.0;
	  double samplePortion=longTotal/double(400);
	  double lambda=
	    (samplePortion>=1 ? 1.0 : confidence*samplePortion);
	  for(Symbol s=0 ; s<alphabetSize ; ++s)
	    {
	      char c=alphabet.lookup(s);
	      TigrString ngram=c+longFuture;
	      const char *p=ngram.c_str(); int len=ngram.length();
	      if(!count->isDefined(p,len)) continue;
	      newModel->lookup(p,len)=
		lambda * model->lookup(p,len) +
		(1-lambda) * prevIMM->lookup(p,len-1);
	    }
	}
    }

  for(int i=1 ; i<=N ; ++i) delete (*models)[i];
  *models=interpolated;
}

