/****************************************************************
 IntronQueue.C

 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "tigr++/Constants.H"
#include "IntronQueue.H"
#include <iostream>
#include "tigrscan.H"


IntronQueue::IntronQueue(ContentSensor &sensor,
			 DiscreteDistribution &distribution,
			 int capacity)
  : SignalQueue(sensor,distribution),
    comp0(0,sensor.getContentType(),distribution),
    comp1(1,sensor.getContentType(),distribution),
    comp2(2,sensor.getContentType(),distribution)
{
  lists[0]=new FixedSizePriorityList<SignalPtr >(capacity,comp0);
  lists[1]=new FixedSizePriorityList<SignalPtr >(capacity,comp1);
  lists[2]=new FixedSizePriorityList<SignalPtr >(capacity,comp2);

  accumulator[1]=accumulator[2]=NEGATIVE_INFINITY;
}



IntronQueue::~IntronQueue()
{
  delete lists[0];
  delete lists[1];
  delete lists[2];

  // do not delete signals here!  they are deleted elsewhere!
}



void IntronQueue::addSignal(SignalPtr signal)
{
  ContentType contentType=getContentType();
  Propagator &propagator=signal->getPropagator(contentType);
  int propagatorPosition=propagator.getPosition();
  if(propagatorPosition <= accumulator.getPosition())
    // This only happens for the "left terminus" signals
    //mainQueue.push_back(signal);
    for(int i=0 ; i<3 ; ++i)
      {
	lists[i]->push_back(signal);
	membershipCounter.increment(signal);
      }
  else
    holdingQueue.push_back(signal);
}



void IntronQueue::updateHoldingQueue(int position)
{
  // Checks whether any signals in holdingQueue have been passed by
  // the accumulator.  If so, calls flushAccumulator() and then moves
  // those signals into the main queue.

  ContentType contentType=getContentType();
  int accumulatorPosition=accumulator.getPosition();

  // Iterate through the holding queue
  bool changes=false;
  TigrList<SignalPtr >::iterator cur=holdingQueue.begin(), 
    end=holdingQueue.end();
  bool first=true;
  while(cur!=end)
    {
      SignalPtr signal=*cur;
      Propagator &newProp=signal->getPropagator(contentType);
      if(newProp.getPosition() <= accumulatorPosition)
	{
	  changes=true;
	  // This signal is ready to be shifted into the main queue.

	  // Make sure any signals already in the main queue are propagated
	  // up to this point before proceding with the promotion (so that
	  // inductive scores will be comparable):
	  if(first) {flushAccumulator();first=false;}

	  // Remove it from the holding queue
	  TigrList<SignalPtr >::iterator graduate=cur;
	  ++cur;
	  holdingQueue.erase(graduate);

	  // Add it to the three phase-specific priority queues
	  for(int phase=0 ; phase<3 ; ++phase)
	    {
	      double newScore=newProp[phase];
	      if(isinf(newScore)) continue;
	      FixedSizePriorityList<SignalPtr > *q=lists[phase];
	      SignalPtr displaced;
	      if(q->size()==q->getCapacity()) displaced=*q->begin();
	      bool inserted=q->insert(signal);//may reject it if score is too low
	      if(inserted) 
		{
		  membershipCounter.increment(signal);
		  if(displaced) 
		      membershipCounter.decrement(displaced);
		}
	    }
	}
      else ++cur;
    }
}



TigrIterator<SignalPtr > &IntronQueue::begin()
{
  iter=IntronQueueIterator(lists,false); // false = not past-the-end
  return iter;
}



TigrIterator<SignalPtr > &IntronQueue::end()
{
  iterEnd=IntronQueueIterator(lists,true); // true = past-the-end
  return iterEnd;
}



int IntronQueue::numElements()
{
  int n=0;
  for(int i=0 ; i<3 ; ++i)
    n+=lists[i]->size();
  return n;
}



void IntronQueue::drop(TigrIterator<SignalPtr > &victim)
{
  SignalPtr signal=*victim;
  membershipCounter.decrement(signal);
  IntronQueueIterator &iqi=static_cast<IntronQueueIterator&>(victim);
  lists[iqi.getIndex()]->erase(iqi.getNativeIterator());
}



void IntronQueue::flushAccumulator()
{
  // Adds accumulator contents to propagators of all signals in the
  // main queue and then resets accumulator scores to 0.

  // Update the signal propagators:
  ContentType contentType=getContentType();
  Strand strand=getStrand();

  TigrMap<SignalPtr ,int>::iterator cur=membershipCounter.begin(),
    end=membershipCounter.end();
  for(; cur!=end ; ++cur)
    {
      SignalPtr signal=(*cur).first;
      Propagator &propagator=signal->getPropagator(contentType);
      propagator.update(accumulator,strand,false,signal);      
    }

  // Reset accumulator scores:
  accumulator.resetScoresForNoncoding();       // [0,-inf,-inf]
}



//==============================================================
//               IntronQueueIterator methods
//==============================================================

IntronQueueIterator::IntronQueueIterator(FixedSizePriorityList<SignalPtr > 
					 *lists[3],
					 bool pastTheEnd)
{
  if(pastTheEnd) 
    {
      index=2;
      iterators[2]=lists[2]->end();
    }
  else
    {
      index=0;
      for(int i=0 ; i<3 ; ++i)
	{
	  iterators[i]=lists[i]->begin();
	  ends[i]=lists[i]->end();
	}
    }
}



void IntronQueueIterator::operator++()
{
  increment();
}



void IntronQueueIterator::increment()  
{
  FixedSizePriorityList<SignalPtr>::iterator &iter=iterators[index];
  ++iter;
  if(iter==ends[index] && index<2) ++index;
}



void IntronQueueIterator::operator++(int)
{
  increment();
}



SignalPtr &IntronQueueIterator::operator*()
{
  return *iterators[index];
}



bool IntronQueueIterator::operator==(const TigrIterator<SignalPtr> &other) const
{
  const IntronQueueIterator &iqiOther=
    dynamic_cast<const IntronQueueIterator&>(other);
  return iterators[index]==iqiOther.iterators[index];
}



FixedSizePriorityList<SignalPtr>::iterator 
   IntronQueueIterator::getNativeIterator()
  const 
{
  return iterators[index];
}



int IntronQueueIterator::getIndex() const
{
  return index;
}



TigrIterator<SignalPtr> &IntronQueueIterator::clone()
{
  return *new IntronQueueIterator(*this);
}




//==============================================================
//               SinglePhaseComparator methods
//==============================================================
SinglePhaseComparator::SinglePhaseComparator(int phase,ContentType contentType,
					     DiscreteDistribution &distr)
  : phase(phase),
    contentType(contentType),
    distr(distr)
{
}



double SinglePhaseComparator::getLengthPenalty(SignalPtr signal,int position)
{
  int length=position-signal->posOfBaseFollowingConsensus();
  double lengthPenalty=distr.getLogP(length);
  return lengthPenalty;
}



void SinglePhaseComparator::getScores(SignalPtr a,SignalPtr b,
				      double &scoreA,double &scoreB)
{
  Propagator &propA=a->getPropagator(contentType);
  Propagator &propB=b->getPropagator(contentType);

  int pos=max(propA.getPosition(),propB.getPosition())+1;

  scoreA=propA[phase]+getLengthPenalty(a,pos);
  scoreB=propB[phase]+getLengthPenalty(b,pos);
}



bool SinglePhaseComparator::equal(SignalPtr &a,SignalPtr &b)
{
  double scoreA, scoreB;
  getScores(a,b,scoreA,scoreB);
  return scoreA==scoreB;
}



bool SinglePhaseComparator::greater(SignalPtr &a,SignalPtr &b)
{
  double scoreA, scoreB;
  getScores(a,b,scoreA,scoreB);
  return scoreA>scoreB;
}



bool SinglePhaseComparator::less(SignalPtr &a,SignalPtr &b)
{
  double scoreA, scoreB;
  getScores(a,b,scoreA,scoreB);
  return scoreA<scoreB;
}



