/****************************************************************
 IntronQueue.H

 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/

#ifndef INCL_IntronQueue_H
#define INCL_IntronQueue_H
#include "tigr++/TigrMap.H"
#include "tigr++/TigrSet.H"
#include "SignalQueue.H"


class SinglePhaseComparator : public TigrComparator<SignalPtr>
{
protected:
  int phase;
  ContentType contentType;
  DiscreteDistribution &distr;

  virtual double getLengthPenalty(SignalPtr,int position);
  virtual void getScores(SignalPtr,SignalPtr,double &scoreA,double &scoreB);
public:
  SinglePhaseComparator(int phase,ContentType,DiscreteDistribution &);
  virtual bool equal(SignalPtr &,SignalPtr &);
  virtual bool greater(SignalPtr &,SignalPtr &);
  virtual bool less(SignalPtr &,SignalPtr &);
};



class IntronQueueIterator : public TigrIterator<SignalPtr>
{
  FixedSizePriorityList<SignalPtr >::iterator iterators[3];
  FixedSizePriorityList<SignalPtr >::iterator ends[3];
  int index;

  void increment();
public:
  IntronQueueIterator() {}
  IntronQueueIterator(FixedSizePriorityList<SignalPtr > *lists[3],
		      bool pastTheEnd=false);
  virtual void operator++();
  virtual void operator++(int);
  virtual SignalPtr &operator*();
  virtual bool operator==(const TigrIterator<SignalPtr > &) const;
  virtual TigrIterator<SignalPtr > &clone();
  FixedSizePriorityList<SignalPtr >::iterator getNativeIterator() const;
  int getIndex() const;
};



/*************************************************************************
  This is a special type of SignalQueue in which there can be multiple
  (overlapping) signals in the holding queue, but there can be at most
  three signals in the main queue -- a best signal for each phase.  When 
  a signal is ready to graduate from the holding queue to the main queue 
  we see (in each phase) if it is better than one in the main queue (i.e., 
  has a better inductive score in that phase), and if so, move it there; 
  otherwise we discard it.
**************************************************************************/

class IntronQueue : public SignalQueue
{
  class MembershipCounter // counts # "main" queues a signal is a member of (0-3)
  {
    TigrMap<SignalPtr ,int> counter;
  public: 
    void increment(SignalPtr s) 
      {if(counter.isDefined(s)) ++counter[s]; else counter[s]=1;}
    void decrement(SignalPtr s) {--counter[s];if(!counter[s]) counter.undefine(s);}
    TigrMap<SignalPtr ,int>::iterator begin() {return counter.begin();}
    TigrMap<SignalPtr ,int>::iterator end() {return counter.end();}
  } membershipCounter;

  FixedSizePriorityList<SignalPtr> *lists[3]; // replaces 'mainQueue' in SignalQueue
  SinglePhaseComparator comp0, comp1, comp2;
  IntronQueueIterator iter, iterEnd;
public:
  IntronQueue(ContentSensor &sensor,DiscreteDistribution &,int capacity);
  virtual ~IntronQueue();
  virtual void addSignal(SignalPtr );
  virtual void updateHoldingQueue(int position);
  virtual TigrIterator<SignalPtr > &begin();
  virtual TigrIterator<SignalPtr > &end();
  TigrMap<SignalPtr ,int>::iterator beginUniq() {return membershipCounter.begin();}
  TigrMap<SignalPtr ,int>::iterator endUniq() {return membershipCounter.end();}
  virtual int numElements();
  virtual void drop(TigrIterator<SignalPtr >&); // removes a signal from the queue
  virtual void flushAccumulator();
};


#endif
