/****************************************************************
 MarkovChain.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "MarkovChain.H"
#include <iostream>
#include <fstream>
#include <math.h>
#include "tigr++/TigrProteinTrans.H"
#include "tigr++/Constants.H"
#include "tigr++/Alphabet.H"
#include "tigr++/TigrStacktrace.H"
#include "NthOrderStringIterator.H"

extern Alphabet alphabet;
#define HASH_TABLE_SIZE /*99991*/ 24967


MarkovChain::MarkovChain(TigrVector<Sequence*> &v,int order,
			 int minSampleSize,int phase,ContentType contentType)
  : N(order), alphabetSize(alphabet.getNumElements()),
    minSampleSize(minSampleSize),  phase(phase)
{
  setContentType(contentType);
  setStrand(PLUS_STRAND);
  buildModels(v);
}



MarkovChain::MarkovChain(istream &is)
{
  load(is);
}



MarkovChain::MarkovChain(int order,int minSampleSize,int phase,
			 Strand strand,ContentType contentType)
  : N(order), 
    minSampleSize(minSampleSize), 
    phase(phase),
    alphabetSize(alphabet.getNumElements())
{
  setStrand(strand);
  setContentType(contentType);
}



MarkovChain::MarkovChain(const TigrString &filename)
{
  ifstream is(filename.c_str());
  if(!is.good()) throw TigrString("Error opening file ")+filename
		   +" in MC::MC()";
  TigrString modelType;
  is >> modelType;
  if(modelType!="MC")
    throw TigrString("Attempt to load an object of type ")+modelType+
      " into an MC (MarkovChain)";
  load(is);
}



MarkovChain::~MarkovChain()
{
  int n=models.size();
  for(int i=0 ; i<n ; ++i)
    delete models[i];
}



double MarkovChain::scoreSubsequence(Sequence &seq,const TigrString &str,
				     int begin,int length,int)
{
  // ### This is quick and dirty -- will be optimizing this class next week
  //     to use a suffix-tree-like data structure so that each base is
  //     read only once

  Symbol dummySymbol;
  char dummyChar;
  double score=0;
  int end=begin+length;
  for(int pos=begin ; pos<end ; ++pos)
    score+=scoreSingleBase(seq,str,pos,dummySymbol,dummyChar);
  return score;
}



double MarkovChain::scoreSingleBase(Sequence &seq,const TigrString &str,
				    int index,Symbol s,char c)
{
  const char *p=str.c_str();
  switch(getStrand())
    {
    case PLUS_STRAND:
      {
	int maxOrder=(index>N ? N : index);
	for(int order=maxOrder ; order>=0 ; --order)
	  {
	    TigrStringMap<double> &model=*models[order];
	    if(model.isDefined(p,index-order,order+1))
	      {
		double d=model.lookup(p,index-order,order+1);
		return d;
	      }
	  }
	throw TigrString("NthOrderMarkovChain::scoreSingleBase('+',")+
	  ","+index+",strlen="+strlen(p)+",str="+p+")";
      }

    case MINUS_STRAND:
      {
	/*
	  On the minus strand we have to take our contexts from the
	  right (but only because we trained the model that way)
	 */
	int seqLen=seq.getLength();
	int maxOrder=seqLen-index;
	if(maxOrder>N) maxOrder=N;
	for(int order=maxOrder ; order>=0 ; --order)
	  {
	    TigrStringMap<double> &model=*models[order];
	    if(model.isDefined(p,index,order+1)) 
	      {
		double d=model.lookup(p,index,order+1);
		return d;
	      }
	  }
	throw TigrStacktrace(
          TigrString("NthOrderMarkovChain::scoreSingleBase('-',")+
	    +index+",strlen="+strlen(p)+",str="+p+")");
      }

    default: throw TigrString(__FILE__)+__LINE__;
    }
}



void MarkovChain::scoreSingleBase(Sequence &seq,const TigrString &str,
				  int index,Symbol s,char c,
				  double &scorePhase0,double &scorePhase1,
				  double &scorePhase2)
{
  // I'm not a 3-periodic markov chain -- I'm just a plain, ordinary,
  // homogeneous markov chain:
  scorePhase0=scorePhase1=scorePhase2=scoreSingleBase(seq,str,index,s,c);
}



ContentSensor *MarkovChain::reverseComplement()
{
  MarkovChain *other=
    new MarkovChain(N,minSampleSize,phase,complement(getStrand()),
		    ::reverseComplement(getContentType()));

  for(int order=0 ; order<=N ; ++order)
    {
      TigrStringMap<double> &model=*models[order];
      TigrStringMap<double> *otherModel=
	new TigrStringMap<double>(HASH_TABLE_SIZE);
      other->models.push_back(otherModel);
      TigrStringMap<double>::iterator cur=model.begin(), end=model.end();
      for(; cur!=end ; ++cur)
	{
	  TigrStringMap<double>::ElemType &e=*cur;
	  TigrString revComp=TigrProteinTrans::reverseComplement(e.first);
	  otherModel->lookup(revComp.c_str(),0,revComp.length())=e.second;
	}
    }
  return other;
}



bool MarkovChain::save(const TigrString &filename)
{
  ofstream os(filename.c_str());
  if(!os.good()) throw TigrString("Error creating file ")+filename+
		   "in MarkovChain::save()";
  return save(os);
}



bool MarkovChain::save(ostream &os)
{
  os << "MC" << endl;
  os << getContentType() << endl;
  os << N << "\t" << phase << "\t" << minSampleSize << endl;
  int numModels=models.size();
  os << numModels << endl;
  for(int i=0 ; i<numModels ; ++i)
    {
      TigrStringMap<double> &model=*models[i];
      os << model.size() << endl;
      TigrStringMap<double>::iterator cur=model.begin(), end=model.end();
      for(; cur!=end ; ++cur)
	os << (*cur).first << endl << (*cur).second << endl;
    }
  return true;
}



void MarkovChain::load(istream &is)
{
  int numModels, numElements;
  TigrString str, pStr;
  ContentType contentType;
  
  is >> contentType >> N >> phase >> minSampleSize >> numModels;
  setContentType(contentType);
  setStrand(::getStrand(contentType));

  for(int i=0 ; i<numModels ; ++i)
    {
      models.push_back(new TigrStringMap<double>(HASH_TABLE_SIZE));
      TigrStringMap<double> &model=*models[i];
      is >> numElements;
      for(int j=0 ; j<numElements ; ++j)
	{
	  is >> str >> pStr;
	  model.lookup(str.c_str(),str.length())=pStr.asDouble();
	}
    }
}



void MarkovChain::buildModels(TigrVector<Sequence*> &v)
{
  for(int i=0 ; i<=N ; ++i)
    models.push_back(new TigrStringMap<double>(HASH_TABLE_SIZE));

  // ### Install pseudocounts
  for(int order=0 ; order<=N ; ++order)
    {
      TigrStringMap<double> &model=*models[order];
      NthOrderStringIterator iterator(order+1,alphabet);
      while(!iterator.done())
	{
	  TigrString ngram=iterator.getNextString();
	  model.lookup(ngram.c_str(),ngram.length())=1;
	}
    }

  // Extract all windows up to size N and count the different
  // combinations that actually occur
  TigrVector<Sequence*>::iterator cur=v.begin(), end=v.end();
  for(; cur!=end ; ++cur)
    {
      Sequence &seq=**cur;
      TigrString &str=*seq.toString(alphabet);
      for(int order=0 ; order<=N ; ++order)
	updateCounts(str,order,seq.getPhase());
      delete &str;
    }

  // Now convert the counts into conditional probabilities
  for(int order=0 ; order<=N ; ++order)
    {
      TigrStringMap<double> &model=*models[order];
      NthOrderStringIterator iterator(order,alphabet);
      while(!iterator.done())
	{
	  // Within this state (history), consider all emissions
	  TigrString history=iterator.getNextString();
	  int sampleSize=0;
	  for(Symbol s=0 ; s<alphabetSize ; ++s)
	    {
	      TigrString ngram=history+alphabet.lookup(s);
	      if(model.isDefined(ngram.c_str(),ngram.length()))
		sampleSize+=int(model.lookup(ngram.c_str(),ngram.length()));
	    }
	  
	  // If sample size is insufficient, remove from the model
	  // (a lower-order model will always be available)
	  if(sampleSize<minSampleSize)
	    undefine(history,model);

	  // Otherwise, normalize counts by sample size to produce a
	  // probability
	  else
	    for(Symbol s=0 ; s<alphabetSize ; ++s)
	      {
		TigrString ngram=history+alphabet.lookup(s);
		if(model.isDefined(ngram.c_str(),ngram.length()))
		  {
		    double &d=model.lookup(ngram.c_str(),ngram.length());
		    d=log(d/sampleSize);
		  }
		else
		  model.lookup(ngram.c_str(),ngram.length())=NEGATIVE_INFINITY;
	      }
	}
    }
}



void MarkovChain::undefine(TigrString &history,
			   TigrStringMap<double> &model)
{
  for(Symbol s=0 ; s<alphabetSize ; ++s)
    {
      TigrString ngram=history+alphabet.lookup(s);
      if(model.isDefined(ngram.c_str(),ngram.length())) 
	model.remove(ngram.c_str(),ngram.length());
    }
}



void MarkovChain::updateCounts(TigrString &str,int order,int seqPhase)
{
  TigrStringMap<double> &model=*models[order];
  int windowSize=order+1;
  int len=str.length();

  int firstPos, periodicity;
  if(phase==NO_PHASE) 
    {
      firstPos=0;
      periodicity=1;
    }
  else 
    {
      if(seqPhase==-1) 
	throw "seq phase=-1 in MarkovChain::updateCounts(), "
	  "called as part of ThreePeriodicMarkovChain";
      firstPos=(phase-seqPhase+3)%3;
      periodicity=3;
    }

  for(int pos=firstPos ; pos<len ; pos+=periodicity)
    {
      int begin=pos-order;
      if(begin<0) continue;
      if(model.isDefined(str.c_str(),begin,windowSize)) 
	model.lookup(str.c_str(),begin,windowSize)+=1;
      else 
	model.lookup(str.c_str(),begin,windowSize)=1;
    }
}


