/****************************************************************
 MarkovChainCompiler.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "MarkovChainCompiler.H"
#include <math.h>
#include <iostream>
#include "tigr++/Alphabet.H"
#include "NthOrderStringIterator.H"

extern Alphabet alphabet;
const double LOG_OF_1=0.0;


FastMarkovChain *MarkovChainCompiler::compile(MarkovChain &mc)
{
  int order=mc.getOrder();
  FastMarkovChain *fmc=new FastMarkovChain(order,mc.getContentType());
  int alphabetSize=alphabet.getNumElements();

  // ===========================
  // Install state probabilities
  // ===========================
  Sequence seq;
  Symbol s;
  fmc->setProb(0,LOG_OF_1);
  int state=1;
  for(int i=0 ; i<=order ; ++i)
    {
      NthOrderStringIterator stateGen(i+1,alphabet);
      while(!stateGen.done())
	{
	  TigrString stateString=stateGen.getNextString();
	  char c=stateString[i];
	  double score=mc.scoreSingleBase(seq,stateString,i,s,c);
	  fmc->setProb(state,score);
	  //cout<<state<<". "<<stateString<<" : "<<c<<"="<<score<<endl;
	  ++state;
	}
    }  

  // ===================
  // Install transitions
  // ===================

  for(int i=0 ; i<alphabetSize ; ++i)
    {
      fmc->setTrans(0,i,i+1);
      //cout<<'\t'<<stateToString(0,*fmc)<<"x"<<alphabet.lookup(i)<<"->"<<stateToString(i+1,*fmc)<<endl;
    }
  state=1;
  for(int i=0 ; i<=order ; ++i)
    {
      NthOrderStringIterator stateGen(i+1,alphabet);
      while(!stateGen.done())
	{
	  TigrString stateString=stateGen.getNextString();
	  TigrString history=
	    (stateString.length()<=order ? stateString :
	     stateString.substring(1,order));
	  for(Symbol s=0 ; s<alphabetSize ; ++s)
	    {
	      TigrString nextStateString=history+alphabet.lookup(s);
	      int nextState=stringToState(nextStateString,*fmc);
	      fmc->setTrans(state,s,nextState);
	      //cout<<stateString<<"x"<<alphabet.lookup(s)<<"->"<<nextStateString<<endl;
	      //cout<<'\t'<<stateToString(state,*fmc)<<"x"<<alphabet.lookup(s)<<"->"<<stateToString(nextState,*fmc)<<endl;
	    }
	  ++state;
	}
    }  

  if(mc.getStrand()==FORWARD_STRAND && mc.getContentType()!=INTERGENIC)
    fmc->revComp=compile(static_cast<MarkovChain&>(*mc.reverseComplement()));
  return fmc;
}



int MarkovChainCompiler::stringToState(const TigrString &s,
				       FastMarkovChain &fmc)
{
  int len=s.length();
  int state=1;
  for(int i=0 ; i<len-1 ; ++i)
    state+=fmc.statesOfOrder(i);
  int alphabetSize=alphabet.getNumElements();
  int factor=1;
  for(int i=0 ; i<len ; ++i)
    {
      state+=(alphabet.lookup(s[len-i-1]) * factor);
      factor*=alphabetSize;
    }
  return state;
}



TigrString MarkovChainCompiler::stateToString(int state,FastMarkovChain &fmc)
{
  int alphabetSize=alphabet.getNumElements();
  int order=0;
  int firstStateOfOrder=1;
  for(; 1 ; ++order)
    {
      if(state<firstStateOfOrder) break;
      firstStateOfOrder+=fmc.statesOfOrder(order);
    }
  --order;
  firstStateOfOrder=1;
  for(int i=0 ; i<order ; ++i)
    firstStateOfOrder+=fmc.statesOfOrder(i);
  state-=firstStateOfOrder;
  TigrString str;
  for(int i=0 ; i<=order ; ++i)
    {
      int factor=static_cast<int>(pow((float)alphabetSize,order-i));
      for(int s=alphabetSize-1 ; s>=0 ; --s)
	{
	  if(state>=s*factor)
	    {
	      str+=alphabet.lookup(s);
	      state-=s*factor;
	      break;
	    }
	}
    }
  return str;
}
