/****************************************************************
 NoncodingQueue.C

 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "NoncodingQueue.H"
#include <iostream>
#include "tigr++/Constants.H"
#include "tigrscan.H"


NoncodingQueue::NoncodingQueue(ContentSensor &sensor,
			       DiscreteDistribution &distribution,
			       int capacity)
  : SignalQueue(sensor,distribution,capacity,noncodingComparator),
    logP1(distribution.getLogP(1)),
    noncodingComparator(sensor.getContentType(),distribution)
{
  // base-class ctor initializes all 3 entries to zero
  accumulator[1]=accumulator[2]=NEGATIVE_INFINITY;
}



void NoncodingQueue::updateHoldingQueue(int position)
{
  // Checks whether any signals in holdingQueue have been passed by
  // the accumulator.  If so, calls flushAccumulator() and then moves
  // those signals into the main queue.

  ContentType contentType=getContentType();
  int accumulatorPosition=accumulator.getPosition();

  // Iterate through the holding queue
  TigrList<SignalPtr >::iterator cur=holdingQueue.begin(), 
    end=holdingQueue.end();
  bool first=true;
  while(cur!=end)
    {
      SignalPtr signal=*cur;
      Propagator &newProp=signal->getPropagator(contentType);
      if(newProp.getPosition() <= accumulatorPosition)
	{
	  // This signal is ready to be shifted into the main queue.

	  // Make sure any signal already in the main queue is propagated
	  // up to this point before proceding with the promotion (so that
	  // inductive scores will be comparable):
	  if(first) {flushAccumulator();first=false;}

	  // Remove it from the holding queue
	  TigrList<SignalPtr >::iterator graduate=cur;
	  ++cur;
	  holdingQueue.erase(graduate);
	  SignalPtr displaced;
	  if(mainQueue.size()==mainQueue.getCapacity()) 
	    displaced=*mainQueue.begin();

	  // Add it to the main queue
	  bool inserted=mainQueue.insert(signal);
	}
      else ++cur;
    }
}



//=================================================================
//                  NoncodingComparator methods
//=================================================================

NoncodingQueue::NoncodingComparator::NoncodingComparator(ContentType 
							 contentType,
						  DiscreteDistribution &distr)
  : SinglePhaseComparator(-1,contentType,distr)
{
}



void NoncodingQueue::NoncodingComparator::getScores(SignalPtr a,SignalPtr b,
						    double &scoreA,
						    double &scoreB)
{
  Propagator &propA=a->getPropagator(contentType);
  Propagator &propB=b->getPropagator(contentType);

  int pos=propA.getPosition()+1; // both propagators are at the same position!

  int phaseA=(a->getStrand()==FORWARD_STRAND ? 0 : 2);
  int phaseB=(b->getStrand()==FORWARD_STRAND ? 0 : 2);

  scoreA=propA[phaseA]+getLengthPenalty(a,pos);
  scoreB=propB[phaseB]+getLengthPenalty(b,pos);
}

