/****************************************************************
 ParseGraph.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "ParseGraph.H"
#ifdef EXPLICIT_GRAPHS
#include <iostream>
#include "SignalTypeProperties.H"
#include "tigr++/TigrVector.H"
#include "tigr++/TigrStack.H"
#include "tigr++/Constants.H"
#include "TigrScan.H"


ParseGraph::ParseGraph(TigrScan &tigrscan)
  : tigrscan(tigrscan),
    vertices(new TigrVector<SignalPtr >)
{
}



ParseGraph::~ParseGraph()
{
  TigrVector<SignalPtr >::iterator cur=vertices->begin(), end=vertices->end();
  for(; cur!=end ; ++cur) delete *cur;
  delete vertices;
}



void ParseGraph::addVertex(SignalPtr s)
{
  /*
    ### This is terribly inefficient.  It implements an insertion
        sort without even using binary search.  There is definitely a
        better way.
   */

  TigrVector<SignalPtr > &vertices=*this->vertices;
  int numVertices=vertices.size();
  if(numVertices==0) vertices.push_back(s);
  else
    {
      int sPos=s->getConsensusPosition();
      for(int i=numVertices-1 ; i>=0 ; --i)
	{
	  SignalPtr t=vertices[i];
	  if(sPos>t->getConsensusPosition())
	    {
	      int sIndex=i+1;
	      vertices.push_back(NULL);
	      for(int j=numVertices ; j>sIndex ; --j)
		vertices[j]=vertices[j-1];
	      vertices[sIndex]=s;
	      return;
	    }
	}
      vertices.push_front(s);
    }
}



int ParseGraph::numVertices()
{
  return vertices->size();
}



SignalPtr ParseGraph::getIthVertex(int i)
{
  return (*vertices)[i];
}



TigrStack<SignalPtr > *ParseGraph::findOptimalPath(double &parseScore)
{
  TigrVector<SignalPtr> &vertices=*this->vertices;
  TigrStack<SignalPtr> *path=new TigrStack<SignalPtr>;

  //===================================================================
  // Use Dynamic programming to set optimal predecessors at each vertex
  //===================================================================

  TigrVector<SignalPtr>::iterator cur=vertices.begin(), end=vertices.end();
  if(vertices.size()==0) return path;
  for(; cur!=end ; ++cur)
    {
      SignalPtr signal=*cur;
      SignalType signalType=signal->getSignalType();
      double signalScore=signal->contextWindowScore();
      Strand signalStrand=signal->getStrand();
      int defaultSignalPhase=(signalStrand==FORWARD_STRAND ? 0 : 2);
      TigrSet<Edge*> &edges=signal->getEdgesIn();
      int numEdges=edges.size();
      if(numEdges==0) // left terminus... DP initialization step
	{
	  for(int i=0 ; i<3 ; ++i) 
	    {
	      signal->setPredecessor(i,NULL);
	      signal->getInductiveScore(i)=0;
	    }
	  continue;
	}
      for(int i=0 ; i<3 ; ++i)
	{
	  signal->setPredecessor(i,NULL);
	  signal->getInductiveScore(i)=NEGATIVE_INFINITY;
	}
      TigrSet<Edge*>::iterator eCur=edges.begin(), eEnd=edges.end();
      for(; eCur!=eEnd ; ++eCur)
	{
	  Edge *edge=*eCur;
	  SignalPtr pred=edge->getLeft();
	  if(edge->isPhased())
	    {
	      PhasedEdge *pe=static_cast<PhasedEdge*>(edge);
	      if(pe->isCoding()) // exon------------------------------------
		for(int edgePhase=0 ; edgePhase<3 ; ++edgePhase)
		  {
		    double score=pe->getInductiveScore(edgePhase)+signalScore;
		    int signalPhase=pe->propagateForward(edgePhase);
		    if(signalType==TAG && signalPhase!=0 ||
		       signalType==NEG_ATG && signalPhase!=2)
		      continue;
		    double &bestScore=signal->getInductiveScore(signalPhase);
		    if(score>bestScore)
		      {
			bestScore=score;
			signal->setPredecessor(signalPhase,pred);
		      }
		  }
	      else // intron------------------------------------------------
		for(int phase=0 ; phase<3 ; ++phase)
		  {
		    double score=pe->getInductiveScore(phase)+signalScore;
		    double &bestScore=signal->getInductiveScore(phase);
		    if(score>bestScore)
		      {
			bestScore=score;
			signal->setPredecessor(phase,pred);
		      }
		  }
	    }
	  else // intergenic or UTR-----------------------------------------
	    {
	      NonPhasedEdge *ne=static_cast<NonPhasedEdge*>(edge);
	      double score=ne->getInductiveScore()+signalScore;
	      double &bestScore=signal->getInductiveScore(defaultSignalPhase);
	      if(score>bestScore)
		{
		  bestScore=score;
		  signal->setPredecessor(defaultSignalPhase,pred);
		}
	    }
	}
    }

  //=======================================
  // Trace back to reconstruct optimal path
  //=======================================

  TigrVector<SignalPtr >::reverse_iterator rCur=vertices.rbegin(), 
    rEnd=vertices.rend();
  double bestScore;
  SignalPtr bestTerminus=NULL;
  int bestPhase;
  for(; rCur!=rEnd ; ++rCur) // iterate through right termini
    {
      SignalPtr signal=*rCur;
      if(!signal->getEdgesOut().isEmpty()) break;// not a right terminus
      for(int phase=0 ; phase<3 ; ++phase)
	{
	  double score=signal->priorInductiveScore(phase);
	  if(!bestTerminus || score>bestScore)
	    {
	      bestScore=score;
	      bestTerminus=signal;
	      bestPhase=phase;
	    }
	}
    }
  parseScore=bestScore;
  int phase=bestPhase;
  SignalPtr signal=bestTerminus;
  while(true)
    {
      path->push(signal);
      SignalPtr pred=signal->getPredecessor(phase);
      signal->setPredecessor((phase+1)%3,NULL);
      signal->setPredecessor((phase+2)%3,NULL);
      if(!pred) break;
      phase=tigrscan.mapPhaseBack(phase,signal,pred);
      signal=pred;
    }
  return path;
}



void ParseGraph::deleteVertices(TigrArray1D<bool> &deleteIfTrue)
{
  TigrVector<SignalPtr > &oldArray=*vertices;
  TigrVector<SignalPtr > *newArray=new TigrVector<SignalPtr >;
  int n=oldArray.size();
  for(int i=0 ; i<n ; ++i)
    {
      SignalPtr signal=oldArray[i];
      if(deleteIfTrue[i])
	delete signal;
      else
	newArray->push_back(signal);
    }
  
  delete vertices;
  vertices=newArray;
}



void ParseGraph::setVertexIndices()
{
  TigrVector<SignalPtr > &array=*vertices;
  int n=array.size();
  for(int i=0 ; i<n ; ++i)
    array[i]->setIndex(i);
}



int ParseGraph::findSignal(SignalType type,int consensusPos) // -1 = !found
{
  // First, find any signal with the same consensusPos
  TigrVector<SignalPtr> &array=*vertices;
  int arraySize=array.size();
  int min=0, max=arraySize;
  while(max-min>1) // binary search
    {
      int mid=(min+max)/2;
      SignalPtr signal=array[mid];
      int signalPos=signal->getConsensusPosition();
      if(consensusPos<signalPos) max=mid;
      else min=mid;
    }

  // Back up to the leftmost signal with this consensusPos
  while(min>0)
    {
      SignalPtr pred=array[min-1];
      if(pred->getConsensusPosition()!=consensusPos) break;
      --min;
    }
  
  // Scan through the signals having this consensusPos to find one of the
  // right type
  while(min<arraySize)
    {
      SignalPtr signal=array[min];
      if(signal->getConsensusPosition()!=consensusPos) break;
      if(signal->getSignalType()==type) return min;
      ++min;
    }

  // None found
  return -1;
}



int ParseGraph::findFirstSignalAfter(int pos) // -1 if no signals follow pos
{
  TigrVector<SignalPtr> &array=*vertices;
  int arraySize=array.size();
  int min=0, max=arraySize;
  while(min<max) // binary search
    {
      int mid=(min+max)/2;
      SignalPtr signal=array[mid];
      int signalPos=signal->getConsensusPosition();
      if(signalPos<pos) min=mid+1;
      else max=mid;
    }
  if(min>=arraySize) return -1;
  //if(array[min]->getConsensusPosition()<pos) {++min;cout<<"NECESSARY!"<<endl;}
  //if(min && array[min-1]->getConsensusPosition()>=pos) throw "OK";
  return min;
}



int ParseGraph::countLeftTermini()
{
  int i=0, n=numVertices();
  for(; i<n ; ++i)
    if(!getIthVertex(i)->isLeftTerminus())
      break;
  return i;
}



int ParseGraph::countRightTermini()
{
  int n=numVertices();
  int i=n-1;
  for(; i>0 ; --i)
    if(!getIthVertex(i)->isRightTerminus())
      break;
  return n-1-i;
}



#endif
