/****************************************************************
 SignalQueue.H
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#ifndef INCL_SignalQueue_H
#define INCL_SignalQueue_H
#include "tigr++/FixedSizePriorityList.H"
#include "tigr++/Constants.H"
#include "ContentSensor.H"
#include "Signal.H"
#include "Accumulator.H"
#include "DiscreteDistribution.H"

/************************************************************************
 Each SignalQueue contains only signals which are followed by the same
 type of content in a valid gene model; thus, each queue has its own
 content sensor for evaluating the content following all the signals
 in the queue.

 The accumulator moves along the sequence accumulating scores from
 individual bases as they are evaluated by this queue's content sensor.
 At certain key times the contents of the accumulator are added to the
 propagators of all the signals in this queue (and the accumulator score is
 then reset to zero).  Thus we maintain the invariant that a signal's
 propagator, appropriately added to the accumulator (i.e., accounting
 for phase/frame), gives the inductive score of the best gene model
 up to this point, passing through that signal.
 
 The main queue contains all signals of a given type that have been
 completely passed by this queue's accumulator; they therefore are
 eligible to be linked to any subsequent signal that is found.
 
 The holdingQueue is for those signals which have been very recently
 added to the queue, so that their propagators (by virtue of having
 been initialized to include the scores in the entire context window)
 are currently ahead of this queue's accumulator.  When that is no
 longer true they are moved to the main queue and are then eligible
 to be the predecessor of a subsequent signal.
***********************************************************************/

class IgnorantComparator : public TigrComparator<SignalPtr >
{
public:
  virtual bool equal(SignalPtr  &a,SignalPtr  &b) {return false;}
  virtual bool greater(SignalPtr  &a,SignalPtr  &b) {return false;}
  virtual bool less(SignalPtr  &a,SignalPtr  &b) {return true;}
};



class SignalQueue
{
  static IgnorantComparator ignorantComparator;
protected:
  ContentSensor &contentSensor;
  Accumulator accumulator;
  FixedSizePriorityList<SignalPtr> mainQueue;
  TigrList<SignalPtr> holdingQueue;
  DiscreteDistribution &distribution; // length disribution
  FSPLIterator<SignalPtr> mainIter, mainEnd;
  TigrList<SignalPtr>::iterator holdingIter,holdingEnd;
public:
  inline SignalQueue(ContentSensor &sensor,DiscreteDistribution &,
		     int capacity=LARGEST_INTEGER,
		     TigrComparator<SignalPtr > &comp=ignorantComparator);
  virtual ~SignalQueue() {/* signals are deleted by GarbageCollector, not here! */ delete &contentSensor;}
  FixedSizePriorityList<SignalPtr> &getMainQueue() {return mainQueue;}
  inline bool precedesExon();
  inline bool precedesIntron();
  inline bool precedesIntergenic();
  inline bool precedesUTR();
  inline Strand getStrand();
  inline ContentType getContentType();
  virtual void addSignal(SignalPtr );
  inline Accumulator &getAccumulator();
  virtual TigrIterator<SignalPtr> &begin();
  virtual TigrIterator<SignalPtr> &end();
  inline TigrList<SignalPtr>::iterator &getHoldingQueue();
  inline TigrList<SignalPtr>::iterator &holdingQueueEnd();
  inline ContentSensor &getContentSensor();
  virtual int numElements();
  inline bool isEmpty();
  virtual void drop(TigrIterator<SignalPtr> &); // removes a signal from the queue
  inline void dropFromHoldingQueue(TigrList<SignalPtr>::iterator &);
  inline DiscreteDistribution &getDistribution();

  inline void addToAccumulator(double scorePhase0,double scorePhase1,
			double scorePhase2,int positionOnSubstrate);
  inline void addToAccumulator(double score);
  // ^--Adds values to accumulator according to absolute (substrate) frame,
  //    advances accumulator's position, and then calls updateHoldingQueue().

  virtual void updateHoldingQueue(int position);
  // ^--Checks whether any signals in holdingQueue have been passed by
  //    the accumulator.  If so, calls flushAccumulator() and then moves
  //    those signals into the main queue.

  virtual void flushAccumulator();
  // ^--Adds accumulator contents to propagators of all signals in the
  //    main queue (after adjusting for frame/phase) and then resets
  //    accumulator scores to 0.
};



inline SignalQueue::SignalQueue(ContentSensor &sensor,
				DiscreteDistribution &distribution,
				int capacity,
				TigrComparator<SignalPtr > &comp)
  : contentSensor(sensor),
    accumulator(-1),
    distribution(distribution),
    mainQueue(capacity,comp)
{
}



inline bool SignalQueue::precedesExon()
{
  return isCoding(contentSensor.getContentType());
}



inline bool SignalQueue::precedesIntron()
{
  return isIntron(contentSensor.getContentType());
}



inline bool SignalQueue::precedesIntergenic()
{
  return isIntergenic(contentSensor.getContentType());
}



inline bool SignalQueue::precedesUTR()
{
  return isUTR(contentSensor.getContentType());
}



inline ContentType SignalQueue::getContentType()
{
  return contentSensor.getContentType();
}



inline Accumulator &SignalQueue::getAccumulator()
{
  return accumulator;
}



inline TigrList<SignalPtr >::iterator &SignalQueue::getHoldingQueue()
{
  holdingIter=holdingQueue.begin();
  return holdingIter;
}



inline TigrList<SignalPtr >::iterator &SignalQueue::holdingQueueEnd()
{
  holdingEnd=holdingQueue.end();
  return holdingEnd;
}



inline void SignalQueue::addToAccumulator(double scorePhase0,
					  double scorePhase1,
					  double scorePhase2,int position)
{
  // CODING VERSION

  // Adds values to accumulator according to absolute (substrate) frame,
  // advances accumulator's position, and then calls updateHoldingQueue().

  int absoluteFrame=position%3;
  switch(getStrand())
    {
    case FORWARD_STRAND:
      // ### THIS CAN BE OPTIMIZED FURTHER BY COMPUTING POSMOD(0-ABSOLUTEFRAME)
      //     AND THEN DUPLICATING THE CODE FOR THE THREE POSSIBLE RESULTS,
      //     THEREBY ALLOWING A POINTER TO BE USED FOR ACCUMULATOR[...]

      accumulator[posmod(0-absoluteFrame)]+=scorePhase0;
      accumulator[posmod(1-absoluteFrame)]+=scorePhase1;
      accumulator[posmod(2-absoluteFrame)]+=scorePhase2;      
      /*
      {
	int i=posmod(-absoluteFrame);
	switch(i)
	  {
	  case 0:
	    {
	      double *as=accumulator.scores;
	      (*as++)+=scorePhase0;
	      (*as++)+=scorePhase1;
	      (*as++)+=scorePhase2;      
	    }
	    break;
	  case 1:
	    {
	      double *as=accumulator.scores;
	      (*as++)+=scorePhase2;
	      (*as++)+=scorePhase0;
	      (*as++)+=scorePhase1;      
	    }
	    break;
	  case 2:
	    {
	      double *as=accumulator.scores;
	      (*as++)+=scorePhase1;
	      (*as++)+=scorePhase2;
	      (*as++)+=scorePhase0;      
	    }
	    break;
	  }
      }
      */
      break;

    case REVERSE_STRAND:
      accumulator[posmod(0+absoluteFrame)]+=scorePhase0;
      accumulator[posmod(1+absoluteFrame)]+=scorePhase1;
      accumulator[posmod(2+absoluteFrame)]+=scorePhase2;      
      /*
      {
	int i=posmod(absoluteFrame);
	switch(i)
	  {
	  case 0:
	    {
	      double *as=accumulator.scores;
	      (*as++)+=scorePhase0;
	      (*as++)+=scorePhase1;
	      (*as++)+=scorePhase2;      
	    }
	    break;
	  case 1:
	    {
	      double *as=accumulator.scores;
	      (*as++)+=scorePhase2;
	      (*as++)+=scorePhase0;
	      (*as++)+=scorePhase1;      
	    }
	    break;
	  case 2:
	    {
	      double *as=accumulator.scores;
	      (*as++)+=scorePhase1;
	      (*as++)+=scorePhase2;
	      (*as++)+=scorePhase0;      
	    }
	    break;
	  }
      }
      */
      break;
    }
  accumulator.incrementPosition();
  updateHoldingQueue(accumulator.getPosition());
}



inline Strand SignalQueue::getStrand()
{
  return contentSensor.getStrand();
}



inline void SignalQueue::addToAccumulator(double score)
{
  // NONCODING VERSION

  // Adds values to accumulator, advances accumulator's position, and 
  // then calls updateHoldingQueue().

  accumulator.addFrameZeroAndAdvance(score);
  updateHoldingQueue(accumulator.getPosition());
}



inline void SignalQueue::flushAccumulator()
{
  // Adds accumulator contents to propagators of all signals in the
  // main queue (after adjusting for frame/phase) and then resets
  // accumulator scores to 0.

  // Update the signal propagators:
  ContentType contentType=getContentType();
  Strand strand=getStrand();
  bool isCoding=precedesExon();
  FixedSizePriorityList<SignalPtr >::iterator cur=mainQueue.begin(), 
    end=mainQueue.end();
  for(; cur!=end ; ++cur)
    {
      SignalPtr signal=*cur;
      Propagator &propagator=signal->getPropagator(contentType);
      propagator.update(accumulator,strand,isCoding,signal);      
    }

  // Reset accumulator scores:
  if(isCoding)
    accumulator.resetScoresForCoding();       // [0,0,0]
  else
    accumulator.resetScoresForNoncoding();    // [0,-inf,-inf]
}



inline ContentSensor &SignalQueue::getContentSensor()
{
  return contentSensor;
}



inline void SignalQueue::dropFromHoldingQueue(TigrList<SignalPtr>::iterator &victim)
{
  holdingQueue.erase(victim);
  SignalPtr signal=*victim;
}



inline DiscreteDistribution &SignalQueue::getDistribution()
{
  return distribution;
}



inline bool SignalQueue::isEmpty()
{
  return numElements()==0;
}


#endif
