/****************************************************************
 TataCapModel.C

 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "TataCapModel.H"
#include <iostream>
#include <fstream>
#include "tigr++/Constants.H"
#include "TataCapSignal.H"


TataCapModel::TataCapModel(GarbageCollector &gc,const TigrString &filename)
  : SignalSensor(gc)
{
  load(filename);
}



TataCapModel::TataCapModel(GarbageCollector &gc,istream &is)
  : SignalSensor(gc)
{
  load(is);
}



TataCapModel::TataCapModel(GarbageCollector &gc,
			   SignalSensor *tataModel,SignalSensor *capModel,
			   int minConsensusSeparation,
			   int maxConsensusSeparation,
			   ContentSensor *contentSensor,
			   SignalType signalType)
  : SignalSensor(gc),
    tataModel(tataModel),
    capModel(capModel),
    contentSensor(contentSensor),
    tataLength(tataModel->getContextWindowLength()),
    capLength(capModel->getContextWindowLength()),
    minConsensusSeparation(minConsensusSeparation),
    maxConsensusSeparation(maxConsensusSeparation)
{
  // Set some things in the superclass:
  setSignalType(signalType);
  Strand strand=::getStrand(signalType);
  setStrand(strand);

  // Convert consensus separation bounds to context window separation bounds:
  switch(strand)
    {
    case FORWARD_STRAND:
      {
	int tataRightContext=
	  tataLength-
	  tataModel->getConsensusOffset()-
	  tataModel->getConsensusLength();
	int capLeftContext=capModel->getConsensusOffset();
	minSeparation=minConsensusSeparation-tataRightContext-capLeftContext;
	maxSeparation=maxConsensusSeparation-tataRightContext-capLeftContext;
      }
      break;
    case REVERSE_STRAND:
      {
	int capRightContext=
	  capLength-
	  capModel->getConsensusOffset()-
	  capModel->getConsensusLength();
	int tataLeftContext=tataModel->getConsensusOffset();
	minSeparation=minConsensusSeparation-capRightContext-tataLeftContext;
	maxSeparation=maxConsensusSeparation-capRightContext-tataLeftContext;
      }
      break;
    }

  // Compute the parameter for the uniform duration distribution:
  int totalLengthRange=maxSeparation+1-minSeparation;
  durationScore=log(1.0/totalLengthRange);
  
  // Set some default sizes (these are rarely used by the system, because
  // we override methods in TataCapSignal so as not to request them from this
  // class) -- they should only be used at the termini:
  int minConsensusLength=tataModel->getConsensusLength()+
    minConsensusSeparation+capModel->getConsensusLength();
  int minContextWindowLength=tataLength+minSeparation+capLength;
  setSizes(minConsensusLength,
	   strand==FORWARD_STRAND ? tataModel->getConsensusOffset() :
	   capModel->getConsensusOffset(),
	   minContextWindowLength);
}



SignalPtr TataCapModel::detect(const Sequence &seq,const TigrString &str,
			     int contextWindowPosition)
{
  switch(getStrand())
    {
    case FORWARD_STRAND: return detect_fw(seq,str,contextWindowPosition);
    case REVERSE_STRAND: return detect_rev(seq,str,contextWindowPosition);
    }
}



SignalPtr TataCapModel::detect_fw(const Sequence &seq,const TigrString &str,
			       int contextWindowPosition)
{
  // First of all, there must be a TATA box here:
  SignalPtr tata=tataModel->detect(seq,str,contextWindowPosition);
  if(!tata) return NULL;

  // Secondly, there must be a CAP site 30-36bp upstream:
  int tataEnd=contextWindowPosition+tataLength;
  int begin=tataEnd+minSeparation;
  int end=tataEnd+maxSeparation;
  SignalPtr bestCap=NULL;
  double bestScore;
  for(int pos=begin ; pos<=end ; ++pos)
    {
      SignalPtr cap=capModel->detect(seq,str,pos);
      if(cap)
	{
	  double score=cap->contextWindowScore();
	  if(!bestCap)
	    {
	      bestCap=cap;
	      bestScore=score;
	    }
	  else if(score>bestScore)
	    {
	      //delete bestCap;
	      bestCap=cap;
	      bestScore=score;
	    }
	}
    }
  if(!bestCap) {/*delete tata;*/ return NULL;}

  // Compute scores and lengths, etc.
  int separationLength=bestCap->getContextWindowPosition()-tataEnd;
  double interveningScore=
    contentSensor->scoreSubsequence(seq,str,tataEnd,separationLength,
				    ContentSensor::NO_PHASE);
  double signalScore=
    tata->contextWindowScore()+bestCap->contextWindowScore()+interveningScore+
    durationScore;
  int signalLength=tataLength+capLength+separationLength;
  int bogusConsensusLength=
    bestCap->posOfBaseFollowingConsensus()-tata->getConsensusPosition();

  // Create a new signal, delete the tata & cap signals, and return
  SignalPtr signal=new TataCapSignal(getGC(),contextWindowPosition,signalLength,
				   bogusConsensusLength,signalScore,*this);

  cout<<"TATA/CAP:"<<getStrand()<<" "<<contextWindowPosition<<" "<<signalLength<<" "<<bogusConsensusLength<<" "<<signalScore<<" "<<str.substring(contextWindowPosition,signalLength)<<endl;

  //delete tata;
  //delete bestCap;
  return signal;
}



SignalPtr TataCapModel::detect_rev(const Sequence &seq,const TigrString &str,
			       int contextWindowPosition)
{
  // First of all, there must be a (reverse) CAP site here:
  SignalPtr cap=capModel->detect(seq,str,contextWindowPosition);
  if(!cap) return NULL;

  // Secondly, there must be a TATA box 30-36bp upstream:
  int capEnd=contextWindowPosition+capLength;
  int begin=capEnd+minSeparation;
  int end=capEnd+maxSeparation;
  SignalPtr bestTata=NULL;
  double bestScore;
  for(int pos=begin ; pos<=end ; ++pos)
    {
      SignalPtr tata=tataModel->detect(seq,str,pos);
      if(tata)
	{
	  double score=tata->contextWindowScore();
	  if(!bestTata)
	    {
	      bestTata=tata;
	      bestScore=score;
	    }
	  else if(score>bestScore)
	    {
	      //delete bestTata;
	      bestTata=tata;
	      bestScore=score;
	    }
	}
    }
  if(!bestTata) {/*delete cap;*/ return NULL;}

  // Compute scores and lengths, etc.
  int separationLength=bestTata->getContextWindowPosition()-capEnd;
  double interveningScore=
    contentSensor->scoreSubsequence(seq,str,capEnd,separationLength,
				    ContentSensor::NO_PHASE);
  double signalScore=
    cap->contextWindowScore()+bestTata->contextWindowScore()+interveningScore+
    durationScore;
  int signalLength=capLength+tataLength+separationLength;
  int bogusConsensusLength=
    bestTata->posOfBaseFollowingConsensus()-cap->getConsensusPosition();

  // Create a new signal, delete the tata & cap signals, and return
  SignalPtr signal=new TataCapSignal(getGC(),contextWindowPosition,signalLength,
				   bogusConsensusLength,signalScore,*this);
  //delete bestTata;
  //delete cap;

  cout<<"TATA/CAP:"<<getStrand()<<" "<<contextWindowPosition<<" "<<signalLength<<" "<<bogusConsensusLength<<" "<<signalScore<<" "<<str.substring(contextWindowPosition,signalLength)<<endl;

  return signal;
}



SignalSensor *TataCapModel::reverseComplement()
{
  TataCapModel *model=
    new TataCapModel(getGC(),
		     tataModel->reverseComplement(),
		     capModel->reverseComplement(),
		     minConsensusSeparation,
		     maxConsensusSeparation,
		     contentSensor->reverseComplement(),
		     ::reverseComplement(getSignalType()));
  return model;
}



double TataCapModel::getLogP(const Sequence &seq,const TigrString &str,int begin)
{
  // ### This is a hack, but I don't think this method ever gets called

  SignalPtr signal=detect(seq,str,begin);
  if(!signal) return NEGATIVE_INFINITY;
  double score=signal->contextWindowScore();
  //delete signal;
  return score;
}



bool TataCapModel::save(const TigrString &filename)
{
  ofstream os(filename.c_str());
  if(!os.good())
    throw TigrString("Error creating file ")+filename+
      "in TataCapModel::save()";
  return save(os);  
}



bool TataCapModel::save(ostream &os)
{
  os << "TataCapModel" << endl;
  os << getSignalType() << endl;
  os << minConsensusSeparation << " "<< maxConsensusSeparation << endl;
  os << minSeparation << " " << maxSeparation << endl;
  os << tataLength << " " << capLength << endl;
  os << durationScore << endl;
  tataModel->save(os);
  capModel->save(os);
  contentSensor->save(os);
  
  TigrStringMap<char> &tataConsensuses=tataModel->getConsensuses();
  os << tataConsensuses.size() << endl;
  TigrStringMap<char>::iterator cur=tataConsensuses.begin(),
    end=tataConsensuses.end();
  for(; cur!=end ; ++cur)
    os << (*cur).first << endl;

  TigrStringMap<char> &capConsensuses=capModel->getConsensuses();
  os << capConsensuses.size() << endl;
  cur=capConsensuses.begin();
  end=capConsensuses.end();
  for(; cur!=end ; ++cur)
    os << (*cur).first << endl;

  return true;
}



void TataCapModel::load(istream &is)
{
  SignalType signalType;
  is >> signalType;
  setSignalType(signalType);
  setStrand(::getStrand(signalType));
  is >> minConsensusSeparation >> maxConsensusSeparation;
  is >> minSeparation >> maxSeparation;
  is >> tataLength >> capLength;
  is >> durationScore;

  tataModel=SignalSensor::load(is,getGC());
  capModel=SignalSensor::load(is,getGC());
  contentSensor=ContentSensor::load(is);

  int numTataCons;
  is >> numTataCons;
  TigrString consensus;
  for(int i=0 ; i<numTataCons ; ++i)
    {
      is >> consensus;
      tataModel->addConsensus(consensus);
    }

  int numCapCons;
  is >> numCapCons;
  for(int i=0 ; i<numCapCons ; ++i)
    {
      is >> consensus;
      capModel->addConsensus(consensus);
    }
}



void TataCapModel::load(const TigrString &filename)
{
  ifstream is(filename.c_str());
  TigrString modelType;
  is >> modelType;
  if(modelType!="TataCapModel")
    throw TigrString("Attempt to load a model of type ")+modelType
      +" into TataCapModel";
  load(is);
}


