/****************************************************************
 ThreePeriodicIMM.C

 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "ThreePeriodicIMM.H"
#include <iostream>
#include <fstream>
#include "tigrscan.H"


ThreePeriodicIMM::ThreePeriodicIMM(const TigrString &filename)
{
  ifstream is(filename.c_str());
  if(!is.good()) throw TigrString("Error opening file ")+filename
		   +" in ThreePeriodicIMM()";
  TigrString modelType;
  is >> modelType;
  if(modelType!="3PIMM")
    throw TigrString("Attempt to load an object of type ")+modelType+
      " into a ThreePeriodicIMM (3PIMM)";
  load(is);  
}



ThreePeriodicIMM::ThreePeriodicIMM(istream &is)
{
  load(is);
}



ThreePeriodicIMM::ThreePeriodicIMM(Strand strand,ContentType contentType)
{
  setStrand(strand);
  setContentType(contentType);
}



ThreePeriodicIMM::ThreePeriodicIMM(TigrVector<TrainingSequence*> &
				   sequences,int order,
				   int minSampleSize,
				   ContentType contentType)
{
  setContentType(contentType);
  setStrand(PLUS_STRAND);
  buildModels(sequences,minSampleSize,order);
}



ThreePeriodicIMM::~ThreePeriodicIMM()
{
  for(int i=0 ; i<3 ; ++i)
    delete chains[i];
}



double ThreePeriodicIMM::scoreSingleBase(const Sequence &seq,
						 const TigrString &str,
						 int index,
						 Symbol s,char c)
{
  throw "noncoding version of scoreSingleBase() not applicable for 3PIMM!";
}



void ThreePeriodicIMM::scoreSingleBase(const Sequence &seq,
				       const TigrString &str,
				       int index,
				       Symbol s,char c,
				       double &scorePhase0,
				       double &scorePhase1,
				       double &scorePhase2)
{
  scorePhase0=chains[0]->scoreSingleBase(seq,str,index,s,c);
  scorePhase1=chains[1]->scoreSingleBase(seq,str,index,s,c);
  scorePhase2=chains[2]->scoreSingleBase(seq,str,index,s,c);
}



double ThreePeriodicIMM::scoreSubsequence(const Sequence &seq,
					  const TigrString &str,
					  int begin,int length,
					  int seqPhase)
{
  // ### This is quick and dirty -- will be optimizing this class next week
  //     to use a suffix-tree-like data structure so that each base is
  //     read only once

  double score=0;
  int end=begin+length;
  switch(getStrand())
    {
    case FORWARD_STRAND:
      for(int pos=begin ; pos<end ; ++pos)
	score+=chains[(seqPhase+pos-begin)%3]->
	  scoreSingleBase(seq,str,pos,seq[pos],str[pos]);
      break;
    case REVERSE_STRAND:
      for(int pos=begin ; pos<end ; ++pos)
	score+=chains[posmod(seqPhase-(pos-begin))]->
	  scoreSingleBase(seq,str,pos,seq[pos],str[pos]);
      break;
    }
  return score;
}



bool ThreePeriodicIMM::save(const TigrString &filename)
{
  ofstream os(filename.c_str());
  if(!os.good()) throw TigrString("Error creating file ")+filename+
		   "in ThreePeriodicIMM::save()";
  return save(os);
}



void ThreePeriodicIMM::buildModels(TigrVector<TrainingSequence*> &sequences,
					   int minSampleSize,int order)
{
  for(int i=0 ; i<3 ; ++i)
    chains[i]=new IMM(sequences,order,minSampleSize,i,
			      getContentType());
}



bool ThreePeriodicIMM::save(ostream &os)
{
  os.precision(8);
  os << "3PIMM" << endl;
  for(int i=0 ; i<3 ; ++i)
    chains[i]->save(os);
}



void ThreePeriodicIMM::load(istream &is)
{
  TigrString dummy;
  for(int i=0 ; i<3 ; ++i)
    {
      is >> dummy;
      chains[i]=new IMM(is);
    }
  ContentType contentType=chains[0]->getContentType();
  setContentType(contentType);
  setStrand(::getStrand(contentType));
}



ContentSensor *ThreePeriodicIMM::reverseComplement()
{
  ThreePeriodicIMM *other=
    new ThreePeriodicIMM(complement(getStrand()),
				 ::reverseComplement(getContentType()));
  for(int i=0 ; i<3 ; ++i)
    other->chains[i]=
      static_cast<IMM*>(chains[i]->reverseComplement());
  return other;
}



void ThreePeriodicIMM::useLogOdds(ContentSensor &nullModel)
{
  ThreePeriodicIMM &null3P=
    dynamic_cast<ThreePeriodicIMM&>(nullModel);
  for(int i=0 ; i<3 ; ++i)
    chains[i]->useLogOdds(*null3P.chains[i]);
}



void ThreePeriodicIMM::useLogOdds_anonymous(ContentSensor &nullModel)
{
  for(int i=0 ; i<3 ; ++i)
    chains[i]->useLogOdds_anonymous(nullModel);
}


