/****************************************************************
 ThreePeriodicMarkovChain.C

 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "ThreePeriodicMarkovChain.H"
#include <iostream>
#include <fstream>
#include "tigrscan.H"


ThreePeriodicMarkovChain::ThreePeriodicMarkovChain(const TigrString &
						   filename)
{
  ifstream is(filename.c_str());
  if(!is.good()) throw TigrString("Error opening file ")+filename
		   +" in ThreePeriodicMarkovChain()";
  TigrString modelType;
  is >> modelType;
  if(modelType!="3P")
    throw TigrString("Attempt to load an object of type ")+modelType+
      " into a ThreePeriodicMarkovChain (3P)";
  load(is);  
}



ThreePeriodicMarkovChain::ThreePeriodicMarkovChain(istream &is)
{
  load(is);
}



ThreePeriodicMarkovChain::ThreePeriodicMarkovChain(Strand strand,
						   ContentType contentType)
{
  setStrand(strand);
  setContentType(contentType);
}



ThreePeriodicMarkovChain::ThreePeriodicMarkovChain(TigrVector<TrainingSequence*> &
						   sequences,int order,
						   int minSampleSize,
						   ContentType contentType)
{
  setContentType(contentType);
  setStrand(PLUS_STRAND);
  buildModels(sequences,minSampleSize,order);
}



ThreePeriodicMarkovChain::~ThreePeriodicMarkovChain()
{
  for(int i=0 ; i<3 ; ++i)
    delete chains[i];
}



double ThreePeriodicMarkovChain::scoreSingleBase(const Sequence &seq,
						 const TigrString &str,
						 int index,
						 Symbol s,char c)
{
  throw "noncoding version of scoreSingleBase() not applicable for 3PMC!";
}



void ThreePeriodicMarkovChain::scoreSingleBase(const Sequence &seq,
					       const TigrString &str,
					       int index,
					       Symbol s,char c,
					       double &scorePhase0,
					       double &scorePhase1,
					       double &scorePhase2)
{
  scorePhase0=chains[0]->scoreSingleBase(seq,str,index,s,c);
  scorePhase1=chains[1]->scoreSingleBase(seq,str,index,s,c);
  scorePhase2=chains[2]->scoreSingleBase(seq,str,index,s,c);
}



double ThreePeriodicMarkovChain::scoreSubsequence(const Sequence &seq,
						  const TigrString &str,
						  int begin,int length,
						  int seqPhase)
{
  // ### This is quick and dirty -- will be optimizing this class next week
  //     to use a suffix-tree-like data structure so that each base is
  //     read only once

  double score=0;
  int end=begin+length;

  switch(getStrand())
    {
    case FORWARD_STRAND:
      for(int pos=begin ; pos<end ; ++pos)
	score+=chains[(seqPhase+pos-begin)%3]->
	  scoreSingleBase(seq,str,pos,seq[pos],str[pos]);
      break;
    case REVERSE_STRAND:
      for(int pos=begin ; pos<end ; ++pos)
	score+=chains[posmod(seqPhase-(pos-begin))]->
	  scoreSingleBase(seq,str,pos,seq[pos],str[pos]);
      break;
    }
  return score;
}



bool ThreePeriodicMarkovChain::save(const TigrString &filename)
{
  ofstream os(filename.c_str());
  if(!os.good()) throw TigrString("Error creating file ")+filename+
		   "in ThreePeriodicMarkovChain::save()";
  return save(os);
}



void ThreePeriodicMarkovChain::buildModels(TigrVector<TrainingSequence*> &sequences,
					   int minSampleSize,int order)
{
  for(int i=0 ; i<3 ; ++i)
    chains[i]=new MarkovChain(sequences,order,minSampleSize,i,
			      getContentType());
}



bool ThreePeriodicMarkovChain::save(ostream &os)
{
  os.precision(8);
  os << "3P" << endl;
  for(int i=0 ; i<3 ; ++i)
    chains[i]->save(os);
}



void ThreePeriodicMarkovChain::load(istream &is)
{
  TigrString dummy;
  for(int i=0 ; i<3 ; ++i)
    {
      is >> dummy;
      chains[i]=new MarkovChain(is);
    }
  ContentType contentType=chains[0]->getContentType();
  setContentType(contentType);
  setStrand(::getStrand(contentType));
}



ContentSensor *ThreePeriodicMarkovChain::reverseComplement()
{
  ThreePeriodicMarkovChain *other=
    new ThreePeriodicMarkovChain(complement(getStrand()),
				 ::reverseComplement(getContentType()));
  for(int i=0 ; i<3 ; ++i)
    other->chains[i]=
      static_cast<MarkovChain*>(chains[i]->reverseComplement());
  return other;
}



void ThreePeriodicMarkovChain::useLogOdds(ContentSensor &nullModel)
{
  ThreePeriodicMarkovChain &null3P=
    dynamic_cast<ThreePeriodicMarkovChain&>(nullModel);
  for(int i=0 ; i<3 ; ++i)
    chains[i]->useLogOdds(*null3P.chains[i]);
}



void ThreePeriodicMarkovChain::useLogOdds_anonymous(ContentSensor &nullModel)
{
  for(int i=0 ; i<3 ; ++i)
    chains[i]->useLogOdds_anonymous(nullModel);
}


