/****************************************************************
 TigrScan
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "TigrScan.H"
#include "tigr++/TigrFastaReader.H"
#ifdef REPORT_PROGRESS
#include "tigr++/TigrProgress.H"
#endif

TigrScan::TigrScan(const TigrString &PROGRAM_NAME,
		   const TigrString &VERSION,EdgeFactory &edgeFactory,
		   int &transcriptId)
  : intergenicSums(0),
    transcriptId(transcriptId),
#ifdef USE_BAYES
    classificationCase(4),
    initialBayes(0),internalBayes(0),finalBayes(0),singleBayes(0),
    useBayes(true),
#endif
    edgeFactory(edgeFactory),
#ifdef EXPLICIT_GRAPHS
    parseGraph(*this),
#endif
    PROGRAM_NAME(PROGRAM_NAME),
    VERSION(VERSION),
    configFile(NULL),
    transitionProbs(NULL)
  {
    // ctor
    
    ++transcriptId;
  }



TigrScan::~TigrScan()
{
  // dtor

  // Delete the signal sensors
  TigrVector<SignalSensor*>::iterator sscur=signalSensors.begin(), 
    ssend=signalSensors.end();
  for(; sscur!=ssend ; ++sscur)
      delete *sscur;

  // Delete the config file
  delete configFile;

  // Delete the Transitions table
  delete transitionProbs;

  // Delete the signal queues
  TigrVector<SignalQueue*>::iterator sqCur=signalQueues.begin(),
    sqEnd=signalQueues.end();
  for(; sqCur!=sqEnd ; ++sqCur) delete *sqCur;

  // Delete the length distributions
  TigrVector<DiscreteDistribution*>::iterator dCur=distributions.begin(),
    dEnd=distributions.end();
  for(; dCur!=dEnd ; ++dCur)
    delete *dCur;

#ifdef USE_BAYES
  // Delete Bayes models
  delete initialBayes; delete internalBayes; 
  delete finalBayes; delete singleBayes;
#endif
}



#ifdef EXPLICIT_GRAPHS
ParseGraph &TigrScan::parse(const TigrString &fastaFilename,
			    const TigrString &isochoreFilename,
			    float &gcContent)
{
  Sequence *seq;
  TigrString *seqStr;
  ParseGraph &graph=parse(fastaFilename,isochoreFilename,seq,seqStr,
			  gcContent);
  delete seq;
  delete seqStr;
  return graph;
}
#endif



#ifdef EXPLICIT_GRAPHS
ParseGraph &TigrScan::parse(const TigrString &fastaFilename,
			    const TigrString &isochoreFilename,
			    Sequence *&substrate,
			    TigrString *&substrateString,
			    float &gcContent)
{
    cerr << "Loading sequence from " << fastaFilename 
	 << "........................" << endl;
    alphabet=DnaAlphabet::global;
    substrate=Sequence::load(fastaFilename,alphabet,substrateId);
    substrateString=substrate->toString(alphabet);
    seqLen=substrate->getLength();

    cerr << "Identifying isochore..." << endl;
    gcContent=getGCcontent(*substrateString);
    cerr << "\t(G+C)/(A+T+C+G)=" << gcContent << endl;

    cerr << "Processing config file..." << endl;
    configFile=IsochoreFile::parse(isochoreFilename,gcContent,gffIsochoreMsg);
    processConfigFile(*configFile);

    cerr << "Running the TigrScan algorithm..." << endl;
    cerr << "\tsequence length=" << substrate->getLength() << endl;
    buildParseGraph(*substrate,*substrateString);

#ifdef REPORT_MEMORY_USAGE
    MemoryProfiler::report("TigrScan TOTAL MEMORY USAGE: ",cerr);
#endif
    
    return parseGraph;
}
#endif



#ifdef EXPLICIT_GRAPHS
ParseGraph &TigrScan::parse(const TigrString &isochoreFilename,
			    const Sequence &substrate,
			    const TigrString &substrateString,
			    float &gcContent)
{
    alphabet=DnaAlphabet::global;
    seqLen=substrate.getLength();

    cerr << "Identifying isochore..." << endl;
    gcContent=getGCcontent(substrateString);
    cerr << "\t(G+C)/(A+T+C+G)=" << gcContent << endl;

    cerr << "Processing config file..." << endl;
    configFile=IsochoreFile::parse(isochoreFilename,gcContent,gffIsochoreMsg);
    processConfigFile(*configFile);

    cerr << "Running the TigrScan algorithm..." << endl;
    cerr << "\tsequence length=" << substrate.getLength() << endl;
    buildParseGraph(substrate,substrateString);

#ifdef REPORT_MEMORY_USAGE
    MemoryProfiler::report("                            TigrScan TOTAL MEMORY USAGE: ",cerr);
#endif
    
    return parseGraph;
}
#endif



#ifdef EXPLICIT_GRAPHS
void TigrScan::buildParseGraph(const Sequence &seq,const TigrString &str)
{
  // Instantiate one signal of each type at the left terminus to act as
  // anchors to which real signals can link back
  instantiateLeftTermini();

  // Make a single left-to-right pass across the sequence
  intergenicSums.resize(seqLen);
  const char *charPtr=str.c_str();
  computeIntergenicSums(seq,str,charPtr);
  for(int pos=0 ; pos<seqLen ; ++pos, ++charPtr)
    {
      Symbol base=seq[pos];
      
      // Check whether any signals occur here
      TigrVector<SignalSensor*>::iterator cur=signalSensors.begin(),
	end=signalSensors.end();
      for(; cur!=end ; ++cur )
	{
	  SignalSensor &sensor=**cur;
	  if(pos+sensor.getContextWindowLength()>seqLen) continue;

	  SignalPtr signal=sensor.detect(seq,str,pos);
#ifdef FORCE_SPECIFIC_SIGNALS
	  if(!signal && forcedSignalCoords.isMember(pos+sensor.getConsensusOffset()))
	    signal=sensor.detectWithNoCutoff(seq,str,pos);
#endif
	  if(signal)
	    {
	      // Find optimal predecessor for this signal in all 3 phases
	      linkBack(str,signal);

	      // Add this signal to the appropriate queue(s)
	      enqueue(signal);
	    }
	}
      
      // Check for stop codons & terminate reading frames when they are found.
      // This check lags behind by two bases so that any stop codon we find 
      // won't overlap with a signal yet to be identified 
      // (consider TAGT=TAG+GT; the TAG should not stop any reading frames 
      // for the GT because when GT is used as a donor only the TA would be 
      // present during translation)
      if(pos>1) handleStopCodons(str,pos-2);

      // Propagate scores of all non-eclipsed signals up to this base
      updateAccumulators(seq,str,pos,base,*charPtr);
    }

  // Instantiate an anchor signal of each type at the right terminus
  // and link them back, to complete the dynamic programming evaluation:
  double parseScore;
  TigrStack<SignalPtr> *path=instantiateRightTermini(str,seqLen,parseScore);
  delete path;

  // Run the garbage collector & build the parse graph
  TigrVector<SignalPtr>::iterator rCur=rightTermini.begin(), 
    rEnd=rightTermini.end();
  for(; rCur!=rEnd ; ++rCur)
    {
      SignalPtr s=*rCur;
      garbageCollector.markLeft(s);
    }
  TigrVector<SignalPtr>::iterator lCur=leftTermini.begin(), 
    lEnd=leftTermini.end();
  for(; lCur!=lEnd ; ++lCur)
    {
      SignalPtr s=*lCur;
      garbageCollector.markRight(s);
    }
  garbageCollector.sweep();
  TigrSet<SignalPtr>::iterator sCur=garbageCollector.signalsBegin(),
    sEnd=garbageCollector.signalsEnd();
  while(sCur!=sEnd)
    {
      SignalPtr s=*sCur;
      parseGraph.addVertex(s);
      ++sCur;
      garbageCollector.drop(s);
    }
}
#endif



int TigrScan::main(int argc,char *argv[])
  {
    // Process command line
    TigrCommandLine cmd(argc,argv,"r:");
    if(cmd.numArgs()!=2)
      throw string(
"\ntigrscan <*.iso> <*.fasta> [-r <*.gff>]\n\
          where -r <*.gff> specifies a GFF file to load and score\n\n");
    TigrString isochoreFilename=cmd.arg(0);
    TigrString fastaFilename=cmd.arg(1);

    cerr << "Loading sequence from " << fastaFilename 
	 << "........................" << endl;
    alphabet=DnaAlphabet::global;

    TigrFastaReader fastaReader(fastaFilename);
    TigrString defline, substrateString;
    fastaReader.nextSequence(defline,substrateString);
    Sequence substrate(substrateString,DnaAlphabet::global);

    seqLen=substrate.getLength();

    cerr << "Identifying isochore..." << endl;
    float gcContent=getGCcontent(substrateString);
    cerr << "\t(G+C)/(A+T+C+G)=" << gcContent << endl;

    cerr << "Processing config file..." << endl;
    configFile=IsochoreFile::parse(isochoreFilename,gcContent,gffIsochoreMsg);
    processConfigFile(*configFile);

    /*
    if(cmd.option('r')) 
      loadGFF(cmd.optParm('r'),substrate,substrateString);
    */

    cerr << "Running the TigrScan algorithm..." << endl;
    cerr << "\tsequence length=" << substrate.getLength() << endl;
    TigrStack<SignalPtr> *path=mainAlgorithm(substrate,substrateString);
    delete path;

#ifdef REPORT_MEMORY_USAGE
    MemoryProfiler::report("TOTAL MEMORY USAGE: ",cerr);
#endif

    return 0;
  }



TigrStack<SignalPtr> * TigrScan::mainAlgorithm(const Sequence &substrate,
					       const TigrString &substrateString,
					       const TigrString &isoFilename,
					       const TigrString &substrateId)
{
    seqLen=substrate.getLength();
    this->substrateId=substrateId;

    cerr << "Identifying isochore..." << endl;
    float gcContent=getGCcontent(substrateString);
    cerr << "\t(G+C)/(A+T+C+G)=" << gcContent << endl;

    cerr << "Processing config file..." << endl;
    configFile=IsochoreFile::parse(isoFilename,gcContent,gffIsochoreMsg);
    processConfigFile(*configFile);

    cerr << "Running the TigrScan algorithm..." << endl;
    cerr << "\tsequence length=" << seqLen << endl;
    return mainAlgorithm(substrate,substrateString);
}



float TigrScan::getGCcontent(const TigrString &seq)
{
  int n=seq.length(), ATCG=0, GC=0;
  const char *p=seq.c_str();
  for(int i=0 ; i<n ; ++i)
    switch(*p++)
      {
      case 'G':
      case 'C':
	++GC;
	// no break...fall through:
      case 'A':
      case 'T':
	++ATCG;
      }
  return GC/float(ATCG);
}



void TigrScan::loadSubmodels(TigrConfigFile &configFile)
{
  useSignalThresholds=configFile.getBoolOrDie("use-signal-thresholds");
  signalThresholdMultiplier=
    configFile.getDoubleOrDie("signal-threshold-multiplier");
  
  invertSignalProbs=configFile.getBoolOrDie("invert-signal-probabilities");

  loadSignalSensor(configFile,"donor-model","donor-consensus");
  loadSignalSensor(configFile,"acceptor-model","acceptor-consensus");
  loadSignalSensor(configFile,"start-codon-model","start-codon-consensus");
  loadSignalSensor(configFile,"stop-codon-model","stop-codon-consensus");
  loadSignalSensor(configFile,"polya-model","polya-consensus");
  loadSignalSensor(configFile,"promoter-model","promoter-consensus");

  loadContentSensor(configFile,"initial-exons",EMPIRICAL_DISTRIBUTION,
		    "initial-exon-lengths");
  loadContentSensor(configFile,"internal-exons",EMPIRICAL_DISTRIBUTION,
		    "internal-exon-lengths");
  loadContentSensor(configFile,"final-exons",EMPIRICAL_DISTRIBUTION,
		    "final-exon-lengths");
  loadContentSensor(configFile,"single-exons",EMPIRICAL_DISTRIBUTION,
		    "single-exon-lengths");
  loadContentSensor(configFile,"introns",GEOMETRIC_DISTRIBUTION,
		    "mean-intron-length");
  loadContentSensor(configFile,"intergenic",GEOMETRIC_DISTRIBUTION,
		    "mean-intergenic-length");
  loadContentSensor(configFile,"3'-UTR",GEOMETRIC_DISTRIBUTION,
		    "mean-3'-UTR-length");
  loadContentSensor(configFile,"5'-UTR",GEOMETRIC_DISTRIBUTION,
		    "mean-5'-UTR-length");

#ifdef USE_BAYES
  if(useBayes)
    {
      TigrString filename=configFile.lookup("bayes-initial-exons");
      initialBayes=filename.isEmpty() ? 
	NULL :
	initialBayes=new BayesModel(filename);

      filename=configFile.lookup("bayes-internal-exons");
      internalBayes=filename.isEmpty() ? 
	NULL :
	internalBayes=new BayesModel(filename);

      filename=configFile.lookup("bayes-final-exons");
      finalBayes=filename.isEmpty() ? 
	NULL :
	finalBayes=new BayesModel(filename);

      filename=configFile.lookup("bayes-single-exons");
      singleBayes=filename.isEmpty() ? 
	NULL :
	singleBayes=new BayesModel(filename);
    }
#endif
}



void TigrScan::loadContentSensor(TigrConfigFile &configFile,
				 const TigrString &modelLabel,
				 DistributionType distrType,
				 const TigrString &distrLabel)
{
  // Load the model
  TigrString filename=configFile.lookupOrDie(modelLabel);
  ContentSensor *sensor=ContentSensor::load(filename);
  
  // Load or create the distribution
  DiscreteDistribution *distribution;
  TigrString distrParm=configFile.lookupOrDie(distrLabel);
  if(distrType==EMPIRICAL_DISTRIBUTION)
    distribution=new EmpiricalDistribution(distrParm);
  else
    distribution=new GeometricDistribution(distrParm.asInt());
  distributions.push_back(distribution);
  //distribution->useLogLikelihoodRatios();//### doesn't work

  // Create a signal queue for it
  ContentType contentType=sensor->getContentType();
  SignalQueue *queue, *negQueue=NULL;
  switch(contentType)
    {
    case INTERGENIC:
      queue=new NoncodingQueue(*sensor,*distribution,queueCapacity);
      signalQueues.push_back(queue);
      contentToQueue[contentType]=queue;
      delete sensor->reverseComplement();
      break;
    case FIVE_PRIME_UTR:
    case THREE_PRIME_UTR:
      queue=new NoncodingQueue(*sensor,*distribution,queueCapacity);
      signalQueues.push_back(queue);
      contentToQueue[contentType]=queue;
#ifndef FORWARD_STRAND_ONLY
      negQueue=new NoncodingQueue(*sensor->reverseComplement(),
				  *distribution,queueCapacity);
      signalQueues.push_back(negQueue);
      contentToQueue[reverseComplement(contentType)]=negQueue;
#endif
      break;
    case INTRON:
      queue=new IntronQueue(*sensor,*distribution,queueCapacity);
      signalQueues.push_back(queue);
      contentToQueue[contentType]=queue;
#ifndef FORWARD_STRAND_ONLY
      negQueue=new IntronQueue(*sensor->reverseComplement(),*distribution,
			       queueCapacity);
      signalQueues.push_back(negQueue);
      contentToQueue[reverseComplement(contentType)]=negQueue;
#endif
      break;
    default:
      queue=new SignalQueue(*sensor,*distribution);
      signalQueues.push_back(queue);
      contentToQueue[contentType]=queue;
#ifndef FORWARD_STRAND_ONLY
      negQueue=new SignalQueue(*sensor->reverseComplement(),*distribution);
      signalQueues.push_back(negQueue);
      contentToQueue[reverseComplement(contentType)]=negQueue;
#endif
      break;
    }

  // Keep a separate list of coding queues (both strands separately)
  // so we can tell when a stop codon eclipses an ORF
  if(queue->precedesExon())
    {
      forwardCodingQueues.push_back(queue);
#ifndef FORWARD_STRAND_ONLY
      reverseCodingQueues.push_back(negQueue);
#endif
    }
}



void TigrScan::loadSignalSensor(TigrConfigFile &configFile,
				const TigrString &modelLabel,
				const TigrString &consensusLabel)
{
  // Load the model
  TigrString filename=configFile.lookupOrDie(modelLabel);
  SignalSensor *sensor=SignalSensor::load(filename,garbageCollector);
  if(!useSignalThresholds) sensor->ignoreCutoff();
  sensor->multiplyCutoffBy(signalThresholdMultiplier);

  // Load inverse signal probability histogram, if applicable
  if(invertSignalProbs)
    {
      TigrString histogramFile=filename;
      histogramFile=histogramFile.substitute(".model",".isp");
      if(TigrFile::exists(histogramFile))
	{
	  Histogram<double> *hist=new Histogram<double>(histogramFile);
	  sensor->useInverseProbabilities(hist);
	}
      else cerr<<"WARNING: "<<histogramFile<<" does not exist -- skipping"
	       <<endl;
    }

  // Load consensus strings
  TigrString consensuses=configFile.lookupOrDie(consensusLabel);
  TigrVector<TigrString> *fields=consensuses.getFields(",|;:/.&_-+");
  TigrVector<TigrString>::iterator cur=fields->begin(), end=fields->end();
  for(; cur!=end ; ++cur)
    sensor->addConsensus(*cur);
  delete fields;

  // Reverse-complement the model
#ifndef FORWARD_STRAND_ONLY
  SignalSensor *negSensor=sensor->reverseComplement();
#endif

  // Add model to global model list
  signalSensors.push_back(sensor);
#ifndef FORWARD_STRAND_ONLY
  signalSensors.push_back(negSensor);
#endif
  signalToSensor[sensor->getSignalType()]=sensor;
#ifndef FORWARD_STRAND_ONLY
  signalToSensor[negSensor->getSignalType()]=negSensor;
#endif

  if(sensor->getSignalType()==TAG)
    {
      stopCodonSensor=sensor;
#ifndef FORWARD_STRAND_ONLY
      negStopCodonSensor=negSensor;
#endif
    }
}



void TigrScan::processConfigFile(TigrConfigFile &configFile)
{
  queueCapacity=configFile.getIntOrDie("queue-capacity");

  pI=configFile.getDoubleOrDie("probability-start-in-intron");
  pN=configFile.getDoubleOrDie("probability-start-in-intergenic");
  pF=configFile.getDoubleOrDie("probability-start-in-5pUTR");
  pT=configFile.getDoubleOrDie("probability-start-in-3pUTR");
  double pSum=pI+pN+pF+pT;
  pI/=pSum; pN/=pSum; pF/=pSum; pT/=pSum;

  TigrString topologyFile=configFile.lookupOrDie("metamodel-topology");
  if(!SignalTypeProperties::global.hasBeenLoaded())
    TopologyLoader::load(topologyFile);

  cerr << "Loading submodels..." << endl;
  loadSubmodels(configFile);

  cerr << "Loading transition probabilities..." << endl;
  TigrString transFile=configFile.lookupOrDie("transition-probabilities");
  float optimism=configFile.getFloatOrDie("exon-optimism");
  float intronOptimism=configFile.getFloatOrDie("intron-optimism");
  loadTransProbs(transFile,optimism,intronOptimism);
}



void TigrScan::loadTransProbs(const TigrString &transFile,float optimism,
			      float intronOptimism)
{
  ifstream is(transFile.c_str());
  if(!is.good()) throw TigrString("Can't open file ")+transFile;
  int numSignalTypes;
  is >> numSignalTypes;
  transitionProbs=new Transitions(numSignalTypes,is,optimism,intronOptimism);
}



void TigrScan::instantiateLeftTermini()
{
  TigrVector<SignalSensor*>::iterator cur=signalSensors.begin(),
    end=signalSensors.end();
  for(; cur!=end ; ++cur)
    {
      SignalSensor &s=**cur;
      TigrSet<ContentType> &queues=s.belongsInWhichQueues();
      TigrSet<ContentType>::iterator cur=queues.begin(), end=queues.end();
      for(; cur!=end ; ++cur)
	{
	  ContentType contentType=*cur;

	  double initialTransitionScore;
	  switch(contentType)
	    {
	    case INITIAL_EXON:
	    case INTERNAL_EXON:
	    case FINAL_EXON:
	    case SINGLE_EXON:
	    case NEG_INITIAL_EXON:
	    case NEG_INTERNAL_EXON:
	    case NEG_FINAL_EXON:
	    case NEG_SINGLE_EXON:
	      continue; // force GHMM to begin in "D-type" state
	    case NEG_INTRON:
	    case INTRON: 
	      initialTransitionScore=pI;
	      if(pI==0.0) continue;
	      break;
	    case INTERGENIC: 
	      initialTransitionScore=pN;
	      break;
	    case NEG_FIVE_PRIME_UTR:
	    case FIVE_PRIME_UTR: 
	      initialTransitionScore=pF;
	      if(pF==0.0) continue; // ### 11/12/03
	      break;
	    case NEG_THREE_PRIME_UTR:
	    case THREE_PRIME_UTR: 
	      initialTransitionScore=pT;
	      if(pF==0.0) continue; // ### 11/12/03
	      break;
	    }

	  SignalPtr terminus=s.getLeftTerminus(initialTransitionScore);
	  SignalQueue *queue=contentToQueue[contentType];
	  queue->addSignal(terminus);
	  for(int i=0 ; i<3 ;++i) terminus->getInductiveScore(i)=0;

#ifdef EXPLICIT_GRAPHS
	  leftTermini.push_back(terminus);
#endif
	}
    }
}



TigrStack<SignalPtr> *TigrScan::instantiateRightTermini(const TigrString &str,
						      int seqLen,
						      double &bestScore)
{
  int bestPhase;
  SignalPtr bestSignal;
  bestScore=NEGATIVE_INFINITY;
  bestSignal=NULL;

  TigrSet<ContentType> instantiated;
  TigrVector<SignalSensor*>::iterator cur=signalSensors.begin(),
    end=signalSensors.end();
  for(; cur!=end ; ++cur)
    {
      // NOTE: we consider only one content-type below because all
      // propagators of the right terminus are initialized identically,
      // and since no propagation will occur past the right terminus, 
      // they will remain identical.
      SignalSensor &s=**cur;
      TigrSet<ContentType> &queues=s.linksBackToWhichQueues();
      ContentType contentType=*queues.begin();

      double finalTransitionScore;
      switch(contentType)
	{
	case INITIAL_EXON:
	case INTERNAL_EXON:
	case FINAL_EXON:
	case SINGLE_EXON:
	case NEG_INITIAL_EXON:
	case NEG_INTERNAL_EXON:
	case NEG_FINAL_EXON:
	case NEG_SINGLE_EXON:
	  continue; // force GHMM to end in "D-type" state
	case NEG_INTRON:
	case INTRON: 
	  finalTransitionScore=pI;
	  if(pI==0.0) continue;
	  break;
	case INTERGENIC: 
	  finalTransitionScore=pN;
	  break;
	case NEG_FIVE_PRIME_UTR:
	case FIVE_PRIME_UTR: 
	  finalTransitionScore=pF;
	  break;
	case NEG_THREE_PRIME_UTR:
	case THREE_PRIME_UTR: 
	  finalTransitionScore=pT;
	  break;
	}

      SignalPtr rightTerminus=s.getRightTerminus(seqLen,finalTransitionScore);
      linkBack(str,rightTerminus);
      Propagator &prop=rightTerminus->getPropagator(contentType);
      for(int i=0 ; i<3 ; ++i)
	{
	  double thisScore=prop[i];
	  if(thisScore>bestScore)
	    {
	      bestScore=thisScore;
	      bestSignal=rightTerminus;
	      bestPhase=i;
	    }
	}
#ifdef EXPLICIT_GRAPHS
      rightTermini.push_back(rightTerminus);
#endif
    }

  cerr<<"Optimal Path Score: "<<bestScore<<endl;
  return traceBack(bestSignal,bestPhase);
}



TigrStack<SignalPtr> * TigrScan::mainAlgorithm(const Sequence &seq,
					       const TigrString &str)
{
  // Compute cumulative intergenic score at each base
  const char *charPtr=str.c_str();
  intergenicSums.resize(seqLen);
  computeIntergenicSums(seq,str,charPtr);

#ifndef EXPLICIT_GRAPHS
  // Instantiate one signal of each type at the left terminus to act as
  // anchors to which real signals can link back
  instantiateLeftTermini();

#ifdef REPORT_PROGRESS
  TigrProgress progress;
  progress.start(seqLen);
#endif

  // Make a single left-to-right pass across the sequence
  for(int pos=0 ; pos<seqLen ; ++pos, ++charPtr)
    {
      Symbol base=seq[pos];
      
      // Check whether any signals occur here
      TigrVector<SignalSensor*>::iterator cur=signalSensors.begin(),
	end=signalSensors.end();
      for(; cur!=end ; ++cur )
	{
	  SignalSensor &sensor=**cur;
	  if(pos+sensor.getContextWindowLength()>seqLen) continue;
#ifdef FORCE_SPECIFIC_SIGNALS
	  SignalPtr signal=
	    (forcedSignalCoords.isMember(pos+sensor.getConsensusOffset()) ? 
	     sensor.detectWithNoCutoff(seq,str,pos) :
	     sensor.detect(seq,str,pos));
#else
	  SignalPtr signal=sensor.detect(seq,str,pos);
#endif
	  if(signal)
	    {
	      // Find optimal predecessor for this signal in all 3 phases
	      linkBack(str,signal);

	      // Add this signal to the appropriate queue(s)
	      enqueue(signal);
	    }
	}
      
      // Check for stop codons & terminate reading frames when they 
      // are found.  This check lags behind by two bases so that any 
      // stop codon we find won't overlap with a signal yet to be 
      // identified (consider TAGT=TAG+GT; the TAG should not stop 
      // any reading frames for the GT because when GT is used as a 
      // donor only the TA would be present during translation)
      if(pos>1) handleStopCodons(str,pos-2);

      // Propagate scores of all non-eclipsed signals up to this base
      updateAccumulators(seq,str,pos,base,*charPtr);

#ifdef REPORT_PROGRESS
      if((pos+1)%5000==0) cerr<<progress.getProgress(pos)<<endl;
#endif
    }

  // Instantiate an anchor signal of each type at the right terminus
  // and link them back, to complete the dynamic programming evaluation:
  double parseScore;
  TigrStack<SignalPtr> *path=instantiateRightTermini(str,seqLen,parseScore);

  // Output gene prediction in GFF format
  generateGff(path,seqLen,parseScore);
#endif

#ifdef EXPLICIT_GRAPHS
  // ### DEBUGGING: do the prediction using the graph instead of the "trellis"
  //const char *charPtr=str.c_str();
  //intergenicSums.resize(seqLen);
  //computeIntergenicSums(seq,str,charPtr);
  buildParseGraph(seq,str);
  double parseScore;
  TigrStack<SignalPtr> *path=parseGraph.findOptimalPath(parseScore);
  generateGff(path,seqLen,parseScore);
#endif

#ifdef REPORT_MEMORY_USAGE
    MemoryProfiler::report("TigrScan TOTAL MEMORY USAGE: ",cerr);
#endif

  return path;
}



void TigrScan::updateAccumulators(const Sequence &seq,
					 const TigrString &str,
					 int pos,Symbol base,char c)
{
  TigrVector<SignalQueue*>::iterator cur=signalQueues.begin(),
    end=signalQueues.end();
  for(; cur!=end ; ++cur)
    {
      SignalQueue &queue=**cur;
      ContentSensor &contentSensor=queue.getContentSensor();
      if(contentSensor.isCoding())
	{
	  double scorePhase0, scorePhase1, scorePhase2;
	  contentSensor.scoreSingleBase(seq,str,pos,base,c,scorePhase0,
					scorePhase1,scorePhase2);
	  queue.addToAccumulator(scorePhase0,scorePhase1,scorePhase2,pos);
	}
      else
	{
	  //### Hmmmm... I think the INTERGENIC model should be turned off here...
	  double score=contentSensor.scoreSingleBase(seq,str,pos,base,c);
	  queue.addToAccumulator(score);
	}
    }
}



void TigrScan::linkBack(const TigrString &str,SignalPtr newSignal)
{
  SignalPtr bestPred[3];
  bestPred[0]=bestPred[1]=bestPred[2]=NULL;
  double bestScore[3];
  bestScore[0]=bestScore[1]=bestScore[2]=NEGATIVE_INFINITY;
  int newConsPos=newSignal->getConsensusPosition();
  Strand strand=newSignal->getStrand();
  SignalType signalType=newSignal->getSignalType();
  if(::endsCoding(signalType)) observeRecentStopCodons(str,newSignal);

  // Consider all queues that newSignal could link back to
  TigrSet<ContentType> &queues=newSignal->linksBackToWhichQueues();
  TigrSet<ContentType>::iterator cur=queues.begin(), end=queues.end();
  for(; cur!=end ; ++cur)
    {
      // Make sure the queue's signals are all propagated up to this point
      ContentType contentType=*cur;
      SignalQueue *queue=contentToQueue[contentType];
      queue->flushAccumulator();

      // Consider all of the queue's signals as potential predecessors
      selectPredecessors(newConsPos,*queue,contentType,strand,
			 bestScore,bestPred,signalType,str,newSignal);
    }

  // Install the selected predecessors as dyn. prog. links and update
  // inductive scores for all propagators of this signal (identically)
  TigrSet<ContentType> &nextQueues=newSignal->belongsInWhichQueues();
  cur=nextQueues.begin();
  end=nextQueues.end();
  for(; cur!=end ; ++cur)
    {
      ContentType contentType=*cur;
      Propagator &prop=newSignal->getPropagator(contentType);
      prop[0]+=bestScore[0];
      prop[1]+=bestScore[1];
      prop[2]+=bestScore[2];
      
      // At this point the propagators are all the same, so if we overwrite
      // one set of inductive scores with those for another content type, it
      // won't matter, because they are the same:
      newSignal->getInductiveScore(0)=prop[0];
      newSignal->getInductiveScore(1)=prop[1];
      newSignal->getInductiveScore(2)=prop[2];
    }
  newSignal->setPredecessor(0,bestPred[0]);
  newSignal->setPredecessor(1,bestPred[1]);
  newSignal->setPredecessor(2,bestPred[2]);
}



inline void TigrScan::selectPredecessors(int newConsPos,
					 SignalQueue &queue,
					 ContentType contentType,
					 Strand strand,
					 double bestScore[3],
					 SignalPtr bestPred[3],
					 SignalType toType,
					 const TigrString &substrate,
					 SignalPtr signal)
{
  switch(contentType)
    {
    case FIVE_PRIME_UTR:
    case THREE_PRIME_UTR:
    case NEG_FIVE_PRIME_UTR:
    case NEG_THREE_PRIME_UTR:
    case INTERGENIC:
      selectIntergenicPred(newConsPos,queue,strand,bestScore,bestPred,
			   contentType,toType,signal);
      break;
    case INTRON:
    case NEG_INTRON:
      selectIntronPred(newConsPos,queue,strand,bestScore,bestPred,
		       contentType,toType,substrate,signal);
      break;
    default:
      selectCodingPred(newConsPos,queue,strand,bestScore,bestPred,
		       contentType,toType,signal);
    }
}



#ifdef USE_BAYES
BayesModel *TigrScan::getBayesModel(ContentType contentType)
{
  switch(contentType)
    {
    case INITIAL_EXON:      
    case NEG_INITIAL_EXON:  
      return initialBayes;
    case INTERNAL_EXON:     
    case NEG_INTERNAL_EXON: 
      return internalBayes;
    case FINAL_EXON:        
    case NEG_FINAL_EXON:    
      return finalBayes;
    case SINGLE_EXON:       
    case NEG_SINGLE_EXON:   
      return singleBayes;
    }
}
#endif



void TigrScan::selectCodingPred(int newConsPos,
				SignalQueue &queue,
				Strand strand,double bestScore[3],
				SignalPtr bestPred[3],
				ContentType contentType,
				SignalType toType,
				SignalPtr signal)
{
  // Version #1 of 3: CODING.  Phase is translated across the exon.

#ifdef USE_BAYES
  BayesModel *bayesModel=getBayesModel(contentType);
#endif
  TigrIterator<SignalPtr> &cur=queue.begin(), &end=queue.end();

  for(; cur!=end ; ++cur)
    {
      SignalPtr pred=*cur;

#ifdef USE_BAYES
      if(strand==FORWARD_STRAND)
	{
	  classificationCase.setContinuous(1,pred->contextWindowScore());
	  classificationCase.setContinuous(2,signal->contextWindowScore());
	}
      else
	{
	  classificationCase.setContinuous(1,signal->contextWindowScore());
	  classificationCase.setContinuous(2,pred->contextWindowScore());
	}
#endif

      Propagator &predProp=pred->getPropagator(contentType);
      SignalType predType=pred->getSignalType();
      int oldPos=pred->posOfBaseFollowingConsensus();
      int length=newConsPos-oldPos;
      double lengthScore=
	queue.getDistribution().getLogP(length+pred->getConsensusLength());
#ifdef USE_BAYES
      classificationCase.setContinuous(0,lengthScore);
#endif
      int frameDelta=length % 3;
#ifdef EXPLICIT_GRAPHS
      double linkScores[3];
#endif
      for(int oldPhase=0 ; oldPhase<3 ; ++oldPhase)
	{
	  if(isinf(predProp[oldPhase])) 
	    {
#ifdef EXPLICIT_GRAPHS
	      linkScores[oldPhase]=NEGATIVE_INFINITY;
#endif
	      continue; // 2/5/2004 WHM
	    }

	  int newPhase=(strand==FORWARD_STRAND ?
			(oldPhase+frameDelta) % 3 :
			posmod(oldPhase-frameDelta));
	  if(recentlyEclipsedPhases[newPhase])
	    {
#ifdef EXPLICIT_GRAPHS
	      linkScores[oldPhase]=NEGATIVE_INFINITY;
#endif
	      continue;
	    }

	  // Implement phase-specific intron states:
	  double transScore=
	    scoreIntronPhases(predType,toType,oldPhase,newPhase);
	  
	  double predScore=
	    predProp[oldPhase] + lengthScore + transScore;
	  double &bestPredScore=bestScore[newPhase];

#ifdef EXPLICIT_GRAPHS
	  linkScores[oldPhase]=
	    predScore-pred->posteriorInductiveScore(oldPhase);
#endif

#ifdef USE_BAYES
	  if(bayesModel) 
	    {
	      double hmmScore=
		predProp[oldPhase]-pred->posteriorInductiveScore(oldPhase);
	      int begin=
		pred->getContextWindowPosition()+pred->getContextWindowLength();
	      int end=signal->getContextWindowPosition();
	      double nullScore=intergenicSums[end]-intergenicSums[begin];
	      classificationCase.setContinuous(3,hmmScore/nullScore);

	      float exonBayesScore=
		bayesModel->getScore(classificationCase,1);
	      float nonExonBayesScore=
		bayesModel->getScore(classificationCase,0);

	      predScore+=log(exonBayesScore/nonExonBayesScore);
	    }
#endif

	  if(finite(predScore) && predScore>bestPredScore)
	    {
	      bestPredScore=predScore;
	      bestPred[newPhase]=pred;
	    }
	}
#ifdef EXPLICIT_GRAPHS
      if(finite(linkScores[0]) || finite(linkScores[1]) || 
	 finite(linkScores[2]))
	{
	  Edge *edge=
	    edgeFactory.newPhasedEdge(linkScores[0],linkScores[1],
				      linkScores[2],pred,signal);
	  pred->addEdgeOut(edge);
	  signal->addEdgeIn(edge);
	}
#endif
    }
}



inline void TigrScan::selectIntronPred(int newConsPos,SignalQueue &q,
				       Strand strand,double bestScore[3],
				       SignalPtr bestPred[3],
				       ContentType contentType,
				       SignalType toType,
				       const TigrString &substrate,
				       SignalPtr signal)
{
  // Version #2 of 3: INTRONS.  Phase is not translated across the intron.
  // Also, we look for stop codons interrupted by the intron and close their
  // open reading frames.

  int seqLen=substrate.length();
  bool phaseIsOpen[3];
  for(int i=0 ; i<3 ; ++i) phaseIsOpen[i]=true;
  TigrString nextCodingBase,nextTwoCodingBases;
  if(newConsPos<seqLen-2)
    {
      nextCodingBase=substrate.substring(newConsPos+2,1);
      nextTwoCodingBases=substrate.substring(newConsPos+2,2);
    }
  IntronQueue &queue=(IntronQueue&)q;
  TigrMap<SignalPtr,int>::iterator cur=queue.beginUniq(), end=queue.endUniq();
  for(; cur!=end ; ++cur)
    {
      SignalPtr pred=(*cur).first;
      int oldConsPos=pred->getConsensusPosition();

      // Handle stop codons interrupted by the intron
      if(newConsPos<seqLen-2 && oldConsPos>1)
	{
	  TigrString prevCodingBase=substrate.substring(oldConsPos-1,1);
	  TigrString prevTwoCodingBases=substrate.substring(oldConsPos-2,2);
	  TigrString codon12=prevCodingBase+nextTwoCodingBases;
	  TigrString codon21=prevTwoCodingBases+nextCodingBase;
	  phaseIsOpen[0]=
	    strand==FORWARD_STRAND ?
	    true :
	    !negStopCodonSensor->consensusOccursAt(codon21,0);
	  phaseIsOpen[1]=
	    strand==FORWARD_STRAND ?
	    !stopCodonSensor->consensusOccursAt(codon12,0) :
	    !negStopCodonSensor->consensusOccursAt(codon12,0);
	  phaseIsOpen[2]=
	    strand==FORWARD_STRAND ?
	    !stopCodonSensor->consensusOccursAt(codon21,0) :
	    true;
	}

      double transScore=transitionProbs->getLogP(pred->getSignalType(),toType);
      int oldPos=pred->posOfBaseFollowingConsensus();
      int length=newConsPos-oldPos;
      double lengthScore=
	queue.getDistribution().getLogP(length+pred->getConsensusLength());
      Propagator &predProp=pred->getPropagator(contentType);
#ifdef EXPLICIT_GRAPHS
      double linkScores[3];
#endif
      for(int phase=0 ; phase<3 ; ++phase)
	{
	  if(!phaseIsOpen[phase]) 
	    {
#ifdef EXPLICIT_GRAPHS
	      linkScores[phase]=NEGATIVE_INFINITY;
#endif
	      continue;
	    }
	  double predScore=predProp[phase] + lengthScore + transScore;
#ifdef EXPLICIT_GRAPHS
	  linkScores[phase]=predScore-pred->posteriorInductiveScore(phase);
#endif
	  double &bestPredScore=bestScore[phase];
	  if(finite(predScore) && predScore>bestPredScore)
	    {
	      bestPredScore=predScore;
	      bestPred[phase]=pred;
	    }
	}
#ifdef EXPLICIT_GRAPHS
      if(finite(linkScores[0]) || finite(linkScores[1]) || 
	 finite(linkScores[2]))
	{
	  Edge *edge=
	    edgeFactory.newPhasedEdge(linkScores[0],linkScores[1],
				      linkScores[2],pred,signal);
	  pred->addEdgeOut(edge);
	  signal->addEdgeIn(edge);
	}
#endif
    }
}



inline void TigrScan::selectIntergenicPred(int newConsPos,SignalQueue &queue,
					   Strand strand,double bestScore[3],
					   SignalPtr bestPred[3],
					   ContentType contentType,
					   SignalType toType,
					   SignalPtr signal)
{
  // Version #3 of 3: INTERGENIC.  Phase 0 or 2 is used depending on strand.

  int newPhase=(strand==FORWARD_STRAND ? 0 : 2);
  TigrIterator<SignalPtr> &cur=queue.begin(), &end=queue.end();
  for(; cur!=end ; ++cur)
    {
      SignalPtr pred=*cur;
      double transScore=transitionProbs->getLogP(pred->getSignalType(),toType);
      int oldPos=pred->posOfBaseFollowingConsensus();
      int length=newConsPos-oldPos;
      double lengthScore=
	queue.getDistribution().getLogP(length+pred->getConsensusLength());
      Propagator &predProp=pred->getPropagator(contentType);
      int oldPhase=(pred->getStrand()==FORWARD_STRAND ? 0 : 2);
      double predScore=predProp[oldPhase] + lengthScore + transScore;
      double &bestPredScore=bestScore[newPhase];
      if(finite(predScore))
	{
#ifdef EXPLICIT_GRAPHS
	  Edge *edge=
	    edgeFactory.newNonPhasedEdge(
	      predScore-pred->posteriorInductiveScore(oldPhase),pred,signal);
	  pred->addEdgeOut(edge);
	  signal->addEdgeIn(edge);
#endif
	  if(predScore>bestPredScore)
	    {
	      bestPredScore=predScore;
	      bestPred[newPhase]=pred;
	    }
	}
    }
}



void TigrScan::enqueue(SignalPtr signal)
{
  TigrSet<ContentType> &queues=signal->belongsInWhichQueues();
  TigrSet<ContentType>::iterator cur=queues.begin(), end=queues.end();
  for(; cur!=end ; ++cur)
    {
      ContentType contentType=*cur;
      contentToQueue[contentType]->addSignal(signal);
    }
}



void TigrScan::handleStopCodons(const TigrString &str,int pos)
{
  if(stopCodonSensor->consensusOccursAt(str,pos))
    terminateForwardORFs(pos);
#ifndef FORWARD_STRAND_ONLY
  else if(negStopCodonSensor->consensusOccursAt(str,pos))
    terminateReverseORFs(pos);
#endif
}



void TigrScan::terminateForwardORFs(int TAGposition)
{
  TigrVector<SignalQueue*>::iterator qCur=forwardCodingQueues.begin(),
    qEnd=forwardCodingQueues.end();
  for(; qCur!=qEnd ; ++qCur)
    {
      SignalQueue &queue=**qCur;

      // *********** ITERATE THROUGH THE MAIN QUEUE *************

      TigrIterator<SignalPtr> &sCur=queue.begin(), &sEnd=queue.end();
      while(sCur!=sEnd)
	{
	  // Figure out which phase the TAG occurs in, and terminate it
	  SignalPtr signal=*sCur;
	  int delta=TAGposition-signal->posOfBaseFollowingConsensus();
	  if(delta<0) {++sCur;continue;}
	  int stoppedPhase=posmod(-delta);
	  Propagator &prop=signal->getPropagator(queue.getContentType());

	  prop[stoppedPhase]=NEGATIVE_INFINITY;

	  // If all phases of this signal are stopped, remove it from queue
	  if(isinf(prop[(stoppedPhase+1)%3]) &&
	     isinf(prop[(stoppedPhase+2)%3]))
	    {
	      TigrIterator<SignalPtr> &victim=sCur.clone();
	      ++sCur; // otherwise queue.erase() will invalidate it
	      queue.drop(victim);
	      delete &victim;
	      continue;
	    }
	  ++sCur;
	}

      // ********** NOW DO THE SAME THING FOR THE HOLDING QUEUE *********

      TigrList<SignalPtr>::iterator hsCur=queue.getHoldingQueue();
      TigrList<SignalPtr>::iterator hsEnd=queue.holdingQueueEnd();
      while(hsCur!=hsEnd)
	{
	  // Figure out which phase the TAG occurs in, and terminate it
	  SignalPtr signal=*hsCur;
	  int delta=TAGposition-signal->posOfBaseFollowingConsensus();
	  if(delta<0) {++hsCur;continue;}
	  int stoppedPhase=posmod(-delta);
	  Propagator &prop=signal->getPropagator(queue.getContentType());

	  prop[stoppedPhase]=NEGATIVE_INFINITY;

	  // If all phases of this signal are stopped, remove it from queue
	  if(isinf(prop[(stoppedPhase+1)%3]) &&
	     isinf(prop[(stoppedPhase+2)%3]))
	    {
	      TigrList<SignalPtr>::iterator victim=hsCur;
	      ++hsCur; // otherwise queue.erase() will invalidate it
	      queue.dropFromHoldingQueue(victim);
	      continue;
	    }
	  ++hsCur;
	}
    }
}



void TigrScan::terminateReverseORFs(int TAGposition)
{
  TigrVector<SignalQueue*>::iterator qCur=reverseCodingQueues.begin(),
    qEnd=reverseCodingQueues.end();
  for(; qCur!=qEnd ; ++qCur)
    {
      SignalQueue &queue=**qCur;

      // *********** ITERATE THROUGH THE MAIN QUEUE *************

      TigrIterator<SignalPtr> &sCur=queue.begin(), &sEnd=queue.end();
      while(sCur!=sEnd)
	{
	  // Figure out which phase the TAG occurs in, and terminate it
	  SignalPtr signal=*sCur;
	  int delta=TAGposition-signal->posOfBaseFollowingConsensus();
	  if(delta<0) {++sCur;continue;}
	  int stoppedPhase=posmod(delta-1);
	  Propagator &prop=signal->getPropagator(queue.getContentType());
	  prop[stoppedPhase]=NEGATIVE_INFINITY;

	  // If all phases of this signal are stopped, remove it from queue
	  if(isinf(prop[(stoppedPhase+1)%3]) &&
	     isinf(prop[(stoppedPhase+2)%3]))
	    {
	      TigrIterator<SignalPtr> &victim=sCur.clone();
	      ++sCur; // otherwise queue.erase() will invalidate it
	      queue.drop(victim);
	      delete &victim;
	      continue;
	    }
	  ++sCur;
	}

      // ********** NOW DO THE SAME THING FOR THE HOLDING QUEUE *********

      TigrList<SignalPtr>::iterator hsCur=queue.getHoldingQueue();
      TigrList<SignalPtr>::iterator hsEnd=queue.holdingQueueEnd();
      while(hsCur!=hsEnd)
	{
	  // Figure out which phase the TAG occurs in, and terminate it
	  SignalPtr signal=*hsCur;
	  int delta=TAGposition-signal->posOfBaseFollowingConsensus();
	  if(delta<0) {++hsCur;continue;}
	  int stoppedPhase=posmod(delta-1);
	  Propagator &prop=signal->getPropagator(queue.getContentType());
	  prop[stoppedPhase]=NEGATIVE_INFINITY;

	  // If all phases of this signal are stopped, remove it from queue
	  if(isinf(prop[(stoppedPhase+1)%3]) &&
	     isinf(prop[(stoppedPhase+2)%3]))
	    {
	      TigrList<SignalPtr>::iterator victim=hsCur;
	      ++hsCur; // otherwise queue.erase() will invalidate it
	      queue.dropFromHoldingQueue(victim);
	      continue;
	    }
	  ++hsCur;
	}
    }  
}



TigrStack<SignalPtr> *TigrScan::traceBack(SignalPtr rightTerminus,int phase)
{
  cerr << "Tracing back to find optimal path..." << endl;

  // Trace back and build stack of visited signals
  TigrStack<SignalPtr> *stk=new TigrStack<SignalPtr>;
  stk->push(rightTerminus);
  SignalPtr signal=rightTerminus;
  while(true)
    {
      SignalPtr pred=signal->getPredecessor(phase);
      signal->setPredecessor((phase+1)%3,NULL);
      signal->setPredecessor((phase+2)%3,NULL);
      if(!pred) break;
      int distance=signal->getConsensusPosition() -
	pred->posOfBaseFollowingConsensus();
      stk->push(pred);
      switch(pred->getSignalType())
	{
	case NEG_TAG:
	case NEG_POLYA:
	case NEG_ATG:     
	case NEG_PROM:    
	  phase=2;
	  break;
	case PROM:        
	case ATG:
	case TAG:         
	case POLYA:       
	  phase=0;
	  break;
	case GT:
	case NEG_AG:
	  break;
	case AG:
	  phase=posmod(phase-distance);
	  break;
	case NEG_GT:
	  phase=posmod(phase+distance);
	  break;
	}
      signal=pred;
    }
  return stk;
}



int TigrScan::getPhase(SignalPtr signal)
{
  // ### This needs to be fixed to work for partial genes

  int phase=(signal->getStrand()==FORWARD_STRAND ? 0 : 2);
  for(int i=0 ; i<3 ; ++i)
    if(signal->getPredecessor(i))
      phase=i;
  return phase;
}



double TigrScan::scoreExon(SignalPtr thisSignal,SignalPtr nextSignal,
			   int phase,ContentType &contentType)
{
  SignalType thisType=thisSignal->getSignalType();
  SignalType nextType=nextSignal->getSignalType();
  int begin=thisSignal->posOfBaseFollowingConsensus();
  int end=nextSignal->getConsensusPosition();
  int length=end-begin;
  int nextPhase=getPhase(nextSignal);
  contentType=
    SignalTypeProperties::global.getContentType(thisType,nextType);
  double lengthPenalty=
    contentToQueue[contentType]->getDistribution().getLogP(length);
  double numerator=
    nextSignal->getInductiveScore(nextPhase)-
    nextSignal->contextWindowScore() -
    thisSignal->getInductiveScore(phase) -
    transitionProbs->getLogP(thisType,nextType) -
    lengthPenalty;
  double denominator=
    intergenicSums[begin+length-1]-(begin ? intergenicSums[begin-1] : 0);
  return numerator-denominator;
}



void TigrScan::generateGff(TigrStack<SignalPtr> *path,int seqLen,
			   double parseScore)
{
  // Invariant: the stack always contains at least 2 signals, and both
  // the top signal and the bottom signal are bogus (left & right termini)
  SignalPtr thisSignal=path->pop();
  SignalPtr firstSignal=thisSignal;
  
  ContentSensor &intergenicSensor=
    contentToQueue[INTERGENIC]->getContentSensor();

  cout.precision(4);
  cout << "##gff-version 2" << endl;
  cout << "##source-version " << PROGRAM_NAME << " " << VERSION << endl;
  cout << "##date " << getDateAndTime();
  cout << "##Type DNA" << endl;

  while(!path->isEmpty())
    {
      int phase=getPhase(thisSignal);
      SignalType thisType=thisSignal->getSignalType();
      Strand strand=thisSignal->getStrand();
      SignalPtr nextSignal=path->pop();
      SignalType nextType=nextSignal->getSignalType();
      int thisPos=thisSignal->getConsensusPosition();
      int nextPos=nextSignal->getConsensusPosition();
      if(thisPos<0) thisPos=0;
      double exonScore;

      // Compute exon score
      ContentType contentType;
      if(::beginsCoding(thisType)) 
	exonScore=scoreExon(thisSignal,nextSignal,phase,contentType);

      switch(thisType)
	{
	case POLYA:
	case NEG_POLYA:
	  if(thisSignal!=firstSignal)
	    cout << substrateId << '\t' << PROGRAM_NAME << "\tpoly-A-signal\t"
		 << thisPos+1 << '\t' 
		 << min(thisPos+1+thisSignal->getConsensusLength(),seqLen)
		 << "\t.\t"<<strand<<"\t.\t"
		 << "transgrp=" 
		 << (thisType==POLYA ? transcriptId : transcriptId+1)
		 << ";" << endl;
	  break;
	case PROM:
	case NEG_PROM:
	  if(thisSignal!=firstSignal)
	    cout << substrateId << '\t' << PROGRAM_NAME << "\tpromoter\t"
		 << thisPos+1 << '\t' 
		 << min(thisPos+1+thisSignal->getConsensusLength(),seqLen)
		 << "\t.\t"<<strand<<"\t.\t"
		 << "transgrp=" 
		 << (thisType==PROM ? transcriptId+1 : transcriptId)
		 << ";" << endl;
	  break;
	case ATG: // initial-exon or single-exon
	  ++transcriptId;
	  switch(nextType)
	    {
	    case GT:
	      cout << substrateId << '\t' << PROGRAM_NAME << "\tinitial-exon\t"
		   << thisPos+1 << '\t' 
		   << min(nextPos,seqLen) << "\t" << exonScore << "\t+\t0\t"
		   << "transgrp=" << transcriptId << ";" << endl;
	      break;
	    case TAG:
	      cout << substrateId << '\t' << PROGRAM_NAME << "\tsingle-exon\t"
		   << thisPos+1 << '\t' 
		   << min(nextPos+3,seqLen) << "\t" << exonScore << "\t+\t0\t"
		   << "transgrp=" << transcriptId << ";" << endl;
	      break;
	    }
	  break;
	case AG: // internal-exon or final-exon
	  switch(nextType)
	    {
	    case GT:
	      cout << substrateId << '\t' << PROGRAM_NAME 
		   << "\tinternal-exon\t"
		   << thisPos+3 << '\t' 
		   << min(nextPos,seqLen) << "\t" << exonScore << "\t+\t"
		   << phase << "\t"
		   << "transgrp=" << transcriptId << ";" << endl;
	      break;
	    case TAG:
	      cout << substrateId << '\t' << PROGRAM_NAME << "\tfinal-exon\t"
		   << thisPos+3 << '\t' 
		   << min(nextPos+3,seqLen) << "\t" << exonScore << "\t+\t"
		   << phase << "\t"
		   << "transgrp=" << transcriptId << ";" << endl;
	      break;
	    }
	  break;
	case NEG_TAG: // final-exon or single-exon
	  ++transcriptId;
	  switch(nextType)
	    {
	    case NEG_AG:
	      cout << substrateId << '\t' << PROGRAM_NAME << "\tfinal-exon\t"
		   << thisPos+1 << '\t' 
		   << min(nextPos,seqLen) << "\t" << exonScore << "\t-\t"
		   << ((getPhase(nextSignal)+1)%3) << "\t"
		   << "transgrp=" << transcriptId << ";" << endl;
	      break;
	    case NEG_ATG:
	      cout << substrateId << '\t' << PROGRAM_NAME << "\tsingle-exon\t"
		   << thisPos+1 << '\t' 
		   << min(nextPos+3,seqLen) << "\t" << exonScore << "\t-\t0\t"
		   << "transgrp=" << transcriptId << ";" << endl;
	      break;
	    }
	  break;
	case NEG_GT: // internal-exon or initial-exon
	  switch(nextType)
	    {
	    case NEG_AG:
	      cout << substrateId << '\t' << PROGRAM_NAME 
		   << "\tinternal-exon\t"
		   << thisPos+3 << '\t' 
		   << min(nextPos,seqLen) << "\t" << exonScore << "\t-\t"
		   << ((getPhase(nextSignal)+1)%3) << "\t"
		   << "transgrp=" << transcriptId << ";" << endl;
	      break;
	    case NEG_ATG:
	      cout << substrateId << '\t' << PROGRAM_NAME << "\tinitial-exon\t"
		   << thisPos+3 << '\t' 
		   << min(nextPos+3,seqLen) << "\t" << exonScore << "\t-\t0\t"
		   << "transgrp=" << transcriptId << ";" << endl;
	      break;
	    }
	  break;
	}
      thisSignal=nextSignal;
    }

  cout << gffIsochoreMsg << endl;
  cout << "# sequence length: " << seqLen << endl;
  cout << "# parse score: " << parseScore << endl;
  cout << "# For information about this file format, see:" << endl;
  cout << "#   <http://www.sanger.ac.uk/Software/formats/GFF/GFF_Spec.shtml>" 
     << endl;
}



double TigrScan::scoreIntronPhases(SignalType predType,SignalType toType,
				       int oldPhase,int newPhase)
{
  double transScore;
  switch(toType)
    {
    case NEG_AG:
      switch(predType)
	{
	case NEG_TAG://                      * * * * * Final exon
	  switch(newPhase)
	    {
	    case 0:
	      transScore=transitionProbs->getLogP(NEG_TAG,NEG_AG0);break;
	    case 1:
	      transScore=transitionProbs->getLogP(NEG_TAG,NEG_AG1);break;
	    case 2:
	      transScore=transitionProbs->getLogP(NEG_TAG,NEG_AG2);break;
	    }
	  break;
	case NEG_GT://                    * * * * * Internal exon
	  switch(newPhase)
	    {
	    case 0:
	      switch(oldPhase)
		{
		case 0: 
		  transScore=transitionProbs->getLogP(NEG_GT0,NEG_AG0);break;
		case 1: 
		  transScore=transitionProbs->getLogP(NEG_GT1,NEG_AG0);break;
		case 2: 
		  transScore=transitionProbs->getLogP(NEG_GT2,NEG_AG0);break;
		}
	      break;
	    case 1:
	      switch(oldPhase)
		{
		case 0: 
		  transScore=transitionProbs->getLogP(NEG_GT0,NEG_AG1);break;
		case 1: 
		  transScore=transitionProbs->getLogP(NEG_GT1,NEG_AG1);break;
		case 2: 
		  transScore=transitionProbs->getLogP(NEG_GT2,NEG_AG1);break;
		}
	      break;
	    case 2:
	      switch(oldPhase)
		{
		case 0: 
		  transScore=transitionProbs->getLogP(NEG_GT0,NEG_AG2);break;
		case 1: 
		  transScore=transitionProbs->getLogP(NEG_GT1,NEG_AG2);break;
		case 2: 
		  transScore=transitionProbs->getLogP(NEG_GT2,NEG_AG2);break;
		}
	      break;
	    }
	  break;
	}
      break;
    case NEG_ATG:
      switch(predType)
	{
	case NEG_TAG://                     * * * * * Single exon
	  transScore=transitionProbs->getLogP(NEG_TAG,NEG_ATG);break;
	  break;
	case NEG_GT://                  * * * * * * Initial exon
	  switch(oldPhase)
	    {
	    case 0:
	      transScore=transitionProbs->getLogP(NEG_GT0,NEG_ATG);break;
	    case 1:
	      transScore=transitionProbs->getLogP(NEG_GT1,NEG_ATG);break;
	    case 2:
	      transScore=transitionProbs->getLogP(NEG_GT2,NEG_ATG);break;
	    }
	  break;
	}
      break;
    case GT:
      switch(predType)
	{
	case ATG://                  * * * * * * Initial exon
	  switch(newPhase)
	    {
	    case 0:
	      transScore=transitionProbs->getLogP(ATG,GT0);break;
	    case 1:
	      transScore=transitionProbs->getLogP(ATG,GT1);break;
	    case 2:
	      transScore=transitionProbs->getLogP(ATG,GT2);break;
	    }
	  break;
	case AG://                    * * * * * Internal exon
	  switch(oldPhase)
	    {
	    case 0:
	      switch(newPhase)
		{
		case 0:
		  transScore=transitionProbs->getLogP(AG0,GT0);break;
		case 1:
		  transScore=transitionProbs->getLogP(AG0,GT1);break;
		case 2:
		  transScore=transitionProbs->getLogP(AG0,GT2);break;
		}
	      break;
	    case 1:
	      switch(newPhase)
		{
		case 0:
		  transScore=transitionProbs->getLogP(AG1,GT0);break;
		case 1:
		  transScore=transitionProbs->getLogP(AG1,GT1);break;
		case 2:
		  transScore=transitionProbs->getLogP(AG1,GT2);break;
		}
	      break;
	    case 2:
	      switch(newPhase)
		{
		case 0:
		  transScore=transitionProbs->getLogP(AG2,GT0);break;
		case 1:
		  transScore=transitionProbs->getLogP(AG2,GT1);break;
		case 2:
		  transScore=transitionProbs->getLogP(AG2,GT2);break;
		}
	      break;
	    }
	  break;
	default: throw "bad! (1)";
	}
      break;
    case TAG:  
      switch(predType)
	{
	case ATG://                     * * * * * Single exon
	  transScore=transitionProbs->getLogP(ATG,TAG);break;
	case AG://                      * * * * * Final exon
	  switch(oldPhase)
	    {
	    case 0:
	      transScore=transitionProbs->getLogP(AG0,TAG);break;
	    case 1:
	      transScore=transitionProbs->getLogP(AG1,TAG);break;
	    case 2:
	      transScore=transitionProbs->getLogP(AG2,TAG);break;
	    }
	  break;
	default: throw "bad! (3)";
	}
      break;
    default: throw "bad! (2)";
    }
  return transScore;
}



void TigrScan::observeRecentStopCodons(const TigrString &str,SignalPtr signal)
{
  /*
    This method observes stop codons which occur anywhere from two bases 
    before the start of this signal's context window up to the last full 
    codon before the signal's consensus.  Any stop codons which are thus 
    found are stored in a set of phases which this signal cannot use for
    linking back to a coding signal.  This is necessary because at the
    time that this signal is instantiated, the main code for handling stop 
    codons will have processed only the sequence up to and including three
    bases left of the signal's context window, so any stop codons between
    that and the signal's consensus would otherwise be overlooked.
   */
  bool *p=recentlyEclipsedPhases;
  *p=*(p+1)=*(p+2)=false;
  int consPos=signal->getConsensusPosition();
  int begin=signal->getContextWindowPosition()-2;
  int end=consPos-3;
  switch(signal->getStrand())
    {
    case FORWARD_STRAND:
      for(int pos=begin ; pos<=end ; ++pos)
	if(stopCodonSensor->consensusOccursAt(str,pos))
	  {
	    int phase=posmod(consPos-pos);
	    recentlyEclipsedPhases[phase]=true;
	  }
      break;
    case REVERSE_STRAND:
      for(int pos=begin ; pos<=end ; ++pos)
	if(negStopCodonSensor->consensusOccursAt(str,pos))
	  {
	    int phase=posmod(pos-consPos-1);
	    recentlyEclipsedPhases[phase]=true;
	  }
      break;
    }
}



int TigrScan::mapPhaseBack(int phase,SignalPtr right,SignalPtr left)
{
  SignalType leftType=left->getSignalType();
  SignalType rightType=right->getSignalType();
  ContentType contentType=
    SignalTypeProperties::global.getContentType(leftType,rightType);
  switch(contentType)
    {
    case INITIAL_EXON:
    case INTERNAL_EXON:
    case FINAL_EXON:
    case SINGLE_EXON:
      {
	int length=
	  right->getConsensusPosition()-left->posOfBaseFollowingConsensus();
	return posmod(phase-length);
      }
    case INTRON:
    case NEG_INTRON:
      return phase;
    case INTERGENIC:
      return (left->getStrand()==FORWARD_STRAND ? 0 : 2);
    case FIVE_PRIME_UTR:
    case THREE_PRIME_UTR:
      return 0;
    case NEG_FIVE_PRIME_UTR:
    case NEG_THREE_PRIME_UTR:
      return 2;
    case NEG_INITIAL_EXON:
    case NEG_INTERNAL_EXON:
    case NEG_FINAL_EXON:
    case NEG_SINGLE_EXON:
      {
	int length=
	  right->getConsensusPosition()-left->posOfBaseFollowingConsensus();
	return (phase+length)%3;
      }
    }
}



#ifdef EXPLICIT_GRAPHS
ParseGraph &TigrScan::getParseGraph()
{
  return parseGraph;
}
#endif



TigrConfigFile *TigrScan::getConfigFile()
{
  return configFile;
}



DiscreteDistribution &TigrScan::getDistribution(ContentType t)
{
  return contentToQueue[t]->getDistribution();
}



ContentType TigrScan::exonTypeToContentType(ExonType exonType,
					       Strand strand)
{
  return ::exonTypeToContentType(exonType,strand);
}



SignalSensor &TigrScan::getSignalSensor(SignalType t) 
{
  return *signalToSensor[t];
}



ContentSensor &TigrScan::getContentSensor(ContentType t)
{
  return contentToQueue[t]->getContentSensor();
}



void TigrScan::computeIntergenicSums(const Sequence &seq,
				     const TigrString &str,
				     const char *p)
{
  ContentSensor &sensor=getContentSensor(INTERGENIC);
  double score=0;
  for(int pos=0 ; pos<seqLen ; ++pos)
    {
      score+=sensor.scoreSingleBase(seq,str,pos,seq[pos],*p++);
      intergenicSums[pos]=score;
    }
}


#ifdef EXPLICIT_GRAPHS
TigrVector<SignalPtr> *TigrScan::getPathFromGff(TigrVector<TigrGffTranscript*> &t,
					      Sequence &seq,
					      const TigrString &seqStr,
					      TigrVector<BOOL> &found)
{
  GffPathFromParseGraph g(*this);
  return g.getPathFromGff(t,seq,seqStr,found);
}
#endif


#ifdef FORCE_SPECIFIC_SIGNALS
void TigrScan::forceSignalCoords(TigrVector<int> &signalCoords)
{
  TigrVector<int>::iterator cur=signalCoords.begin(), end=signalCoords.end();
  for(; cur!=end ; ++cur)
    forcedSignalCoords.insert(*cur);
}
#endif
