/****************************************************************
 TigrScan
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#ifndef INCL_TigrScan_H
#include <math.h>
#include <iostream>
#include <fstream>
#include <typeinfo>
#include "tigr++/TigrCommandLine.H"
#include "tigr++/DnaAlphabet.H"
#include "tigr++/TigrConfigFile.H"
#include "tigr++/TigrMap.H"
#include "tigr++/TigrStack.H"
#include "tigr++/Constants.H"
#include "tigr++/TigrTime.H"
#include "tigr++/TigrArray1D.H"
#include "tigr++/TigrGffReader.H"
#include "tigrscan.H"
#include "SignalSensor.H"
#include "SignalQueue.H"
#include "SignalTypeProperties.H"
#include "NoncodingQueue.H"
#include "IntronQueue.H"
#include "EmpiricalDistribution.H"
#include "GeometricDistribution.H"
#include "Transitions.H"
#include "IsochoreFile.H"
#include "GffReader.H"
#include "TopologyLoader.H"
#include "Edge.H"
#include "GarbageCollector.H"
#include "EdgeFactory.H"
#include "GffPathFromParseGraph.H"
#ifdef USE_BAYES
#include "TAIL/bayes/BayesModel.H"
#endif

#ifdef EXPLICIT_GRAPHS
#include "ParseGraph.H"
#endif

#ifdef REPORT_MEMORY_USAGE
#include "tigr++/MemoryProfiler.H"
#endif



/****************************************************************
                        class TigrScan
 ****************************************************************/

class TigrScan
{
public:

/****************************************************************
                        PUBLIC MEMBERS
 ****************************************************************/

  TigrScan(const TigrString &PROGRAM_NAME,const TigrString &VERSION,
	   EdgeFactory &,int &transcriptId);
  virtual ~TigrScan();
  int main(int argc,char *argv[]);
  virtual TigrStack<SignalPtr> * mainAlgorithm(const Sequence &,
					       const TigrString &,
					       const TigrString &
					         isochoreFilename,
					       const TigrString &substrateId);
  int mapPhaseBack(int phase,SignalPtr right,SignalPtr left);
  TigrConfigFile *getConfigFile();
  DiscreteDistribution &getDistribution(ContentType);
  Transitions *getTransitionProbs()            {return transitionProbs;}
  TigrArray1D<double> &getIntergenicSums()      {return intergenicSums;}
  const TigrString &getSubstrateId()               {return substrateId;}
  int getSeqLen()                                       {return seqLen;}
  TigrString getGffIsochoreMsg()                {return gffIsochoreMsg;}
  static ContentType exonTypeToContentType(ExonType,Strand);
  SignalSensor &getSignalSensor(SignalType t);
  ContentSensor &getContentSensor(ContentType);
  GarbageCollector &getGC()                   {return garbageCollector;}
  double scoreExon(SignalPtr left,SignalPtr right,int phase,
		   ContentType &);

#ifdef USE_BAYES
  void dontUseBayesModel() {useBayes=false;}
#endif

#ifdef EXPLICIT_GRAPHS
  ParseGraph &getParseGraph();
  ParseGraph &parse(const TigrString &seqFile,const TigrString &isoFile,
		    float &gcContent);
  ParseGraph &parse(const TigrString &seqFile,const TigrString &isoFile,
		    Sequence *&,TigrString *&seqString,float &gcContent);
  ParseGraph &parse(const TigrString &isoFile,const Sequence &,
		    const TigrString &seqString,float &gcContent);
  int numRightTermini() {return rightTermini.size();}
  TigrVector<SignalPtr> *getPathFromGff(TigrVector<TigrGffTranscript*> &,
				      Sequence &,const TigrString &seqStr,
				      TigrVector<BOOL> &found);
#endif

#ifdef FORCE_SPECIFIC_SIGNALS
  void forceSignalCoords(TigrVector<int> &signalCoords);
#endif

/****************************************************************
                     PRIVATE & PROTECTED MEMBERS
 ****************************************************************/

protected:
#ifdef FORCE_SPECIFIC_SIGNALS
  TigrSet<int> forcedSignalCoords;
#endif

#ifdef EXPLICIT_GRAPHS
  TigrVector<SignalPtr> leftTermini, rightTermini;
  ParseGraph parseGraph;

  void buildParseGraph(const Sequence &seq,const TigrString &str);
#endif

#ifdef USE_BAYES
  bool useBayes;
  BayesModel *initialBayes, *internalBayes, *finalBayes, *singleBayes;
  Case classificationCase;
  BayesModel *getBayesModel(ContentType);
#endif

  bool invertSignalProbs; // use P(signal|score), not P(sequence|signal)
  bool useSignalThresholds;
  double signalThresholdMultiplier;
  int &transcriptId;
  TigrVector<DiscreteDistribution*> distributions;
  GarbageCollector garbageCollector;
  EdgeFactory &edgeFactory;
  TigrString PROGRAM_NAME, VERSION;
  int queueCapacity; // for constraining graph construction
  int seqLen;
  double pI, pN, pF, pT;
  TigrString substrateId;
  TigrVector<SignalSensor*> signalSensors;
  TigrVector<SignalQueue*> signalQueues;
  TigrVector<SignalQueue*> forwardCodingQueues, reverseCodingQueues;
  TigrMap<ContentType,SignalQueue*> contentToQueue;
  SignalSensor *stopCodonSensor, *negStopCodonSensor;
  Transitions *transitionProbs;
  TigrString gffIsochoreMsg;

  bool recentlyEclipsedPhases[3];

  TigrMap<SignalType,SignalSensor*> signalToSensor;
  TigrArray1D<double> intergenicSums;
  TigrConfigFile *configFile;
  float gcContent;

  virtual TigrStack<SignalPtr> * mainAlgorithm(const Sequence &,
					       const TigrString &);
  void computeIntergenicSums(const Sequence &,const TigrString &,
			     const char *);
  float getGCcontent(const TigrString &);
  void processConfigFile(TigrConfigFile &);
  void loadSubmodels(TigrConfigFile &);
  void loadSignalSensor(TigrConfigFile &,const TigrString &modelLabel,
			const TigrString &consensusLabel);
  void loadContentSensor(TigrConfigFile &,const TigrString &modelLabel,
			 DistributionType,const TigrString &distrLabel);
  void instantiateLeftTermini();
  TigrStack<SignalPtr> *instantiateRightTermini(const TigrString &,int seqLen,
					      double &parseScore);
  virtual void updateAccumulators(const Sequence &,const TigrString &,int pos,
				  Symbol,char);
  /*inline*/ void linkBack(const TigrString &,SignalPtr newSignal);
  inline void selectPredecessors(int newConsPos,SignalQueue &queue,
				 ContentType contentType,Strand strand,
				 double bestScore[3],SignalPtr bestPred[3],
				 SignalType toType,
				 const TigrString &substrate,
				 SignalPtr);
  inline void selectIntergenicPred(int newConsPos,SignalQueue &queue,
				   Strand strand,double bestScore[3],
				   SignalPtr bestPred[3],ContentType,
				   SignalType toType,SignalPtr);
  virtual void selectCodingPred(int newConsPos,SignalQueue &queue,
			       Strand strand,double bestScore[3],
			       SignalPtr bestPred[3],ContentType,
			       SignalType toType,SignalPtr);
  inline void selectIntronPred(int newConsPos,SignalQueue &queue,
			       Strand strand,
			       double bestScore[3],SignalPtr bestPred[3],
			       ContentType,SignalType toType,
			       const TigrString &substrate,SignalPtr);
  void enqueue(SignalPtr);
  void handleStopCodons(const TigrString &,int pos);
  void terminateForwardORFs(int position);
  void terminateReverseORFs(int position);
  void loadTransProbs(const TigrString &transFile,float optimism,
		      float intronOptimism);
  TigrStack<SignalPtr> *traceBack(SignalPtr rightTerminus,int phase);
  void generateGff(TigrStack<SignalPtr> *path,int seqLen,double parseScore);
  double scoreIntronPhases(SignalType predType,SignalType toType,
			   int oldPhase,int newPhase);
  int getPhase(SignalPtr);
  void observeRecentStopCodons(const TigrString &,SignalPtr);
};


#endif
