/****************************************************************
 TopologyLoader.C
 bmajoros@tigr.org

 Copyright (c)2003, The Institute for Genomic Research (TIGR),
 Rockville, Maryland, U.S.A.  All rights reserved.  This software
 is governed by the ARTISTIC LICENSE (see www.opensource.org).
 ****************************************************************/
#include "TopologyLoader.H"
#include <iostream>
#include "tigr++/TigrFile.H"


// "GT -> AG : INTRON"
TigrRegex TopologyLoader::transitionRegex(
   "^\\s*(\\S+)\\s*->\\s*(\\S+)\\s*:\\s*(\\S+)\\s*$");

// "GT has phase 0"
TigrRegex TopologyLoader::phaseRegex(
   "^\\s*(\\S+)\\s+has\\s+phase\\s+(\\S+)\\s*$");

// "GT strand +"
TigrRegex TopologyLoader::strandRegex(
   "^\\s*(\\S+)\\s+strand\\s+(\\S+)\\s*$");

// "ATG consensus coding"
TigrRegex TopologyLoader::consensusCodingRegex(
   "^\\s*(\\S+)\\s+consensus\\s+coding\\s*$");

// "# comment"
TigrRegex TopologyLoader::commentRegex("^\\s*#.*$");

TigrRegex TopologyLoader::blankLineRegex("^\\s*$");



SignalTypeProperties &TopologyLoader::load(const TigrString &filename)
{
  SignalTypeProperties &stp=SignalTypeProperties::global;
  int lineNumber=0;
  TigrFile file(filename);
  while(!file.eof())
    {
      TigrString line=file.readLine();
      ++lineNumber;
      if(transitionRegex.search(line))
	{
	  // "GT -> AG : INTRON"
	  SignalType from=stringToSignalType(transitionRegex[1]);
	  SignalType to=stringToSignalType(transitionRegex[2]);
	  ContentType content=stringToContentType(transitionRegex[3]);
	  stp.belongTo(from,content);
	  stp.linkTo(to,content);
	}
      else if(phaseRegex.search(line))
	{
	  // "GT has phase 0"
	  SignalType sig=stringToSignalType(phaseRegex[1]);
	  int phase=phaseRegex[2].asInt();
	  stp.allowPhase(sig,phase);
	}
      else if(strandRegex.search(line))
	{
	  // "GT strand +"
	  SignalType sig=stringToSignalType(strandRegex[1]);
	  char strand=strandRegex[2][0];
	  stp.setStrand(sig,strand);
	}
      else if(consensusCodingRegex.search(line))
	{
	  // "ATG consensus coding"
	  SignalType sig=stringToSignalType(consensusCodingRegex[1]);
	  stp.setConsensusCoding(sig,true);
	}
      else if(!commentRegex.match(line) && !blankLineRegex.match(line))
	throw filename+", line "+TigrString(lineNumber)+", syntax error:\n"+line;
    }
  return stp;
}


