/*******************************************************************
 compute-signal-likelihoods.C

 B. Majoros.
 The Institute for Genomic Research
*******************************************************************/
#include <string>
#include <iostream>
#include <fstream>
#include "tigr++/TigrCommandLine.H"
#include "tigr++/TigrVector.H"
#include "tigr++/TigrFastaReader.H"
#include "tigr++/DnaAlphabet.H"
#include "tigr++/Sequence.H"
#include "tigr++/TigrSummaryStats.H"
#include "tigr++/Constants.H"
#include "tigr++/Histogram.H"
#include "tigr++/TigrRegex.H"
#include "SignalSensor.H"
#include "GarbageCollector.H"


Alphabet alphabet;
TigrRegex filestemRegex("([^/.]+).model");


class Application
{
  TigrVector<double> *loadPosExamples(const TigrString &filename,
				      SignalSensor &);
  TigrVector<double> *loadBackground(const TigrString &filename,
				     SignalSensor &);
  void writeHistogramFile(TigrVector<double> &,const TigrString &filename);
  void getExtrema(TigrVector<double> &scores,double &min,double &max);
  void computeRatios(const TigrString &filestem,Histogram<double> &posHist,
		     Histogram<double> &backgroundHist,double minScore,
		     double maxScore,int numBins);
public:
  Application();
  int main(int argc,char *argv[]);
};


int main(int argc,char *argv[])
  {
    try
      {
	Application app;
	return app.main(argc,argv);
      }
    catch(const char *p)
      {
	cerr << p << endl;
      }
    catch(const string &msg)
      {
	cerr << msg.c_str() << endl;
      }
    catch(const exception &e)
      {
	cerr << "STL exception caught in main:\n" << e.what() << endl;
      }
    catch(...)
      {
	cerr << "Unknown exception caught in main" << endl;
      }
    return -1;
  }



Application::Application()
  {
    // ctor
  }



int Application::main(int argc,char *argv[])
  {
    // Process command line
    TigrCommandLine cmd(argc,argv,"");
    if(cmd.numArgs()!=5)
      throw string(
"\ncompute-signal-likelihoods <*.model> <pos-examples.fasta> \n\
                              <neg-examples.fasta> <consensuses> <#bins>\n\
\n\
example:\n\
  compute-signal-likelihoods tag.model tag.fasta contigs.fasta TAG,TGA,TAA\n\
\n\
\n");
    TigrString modelFilename=cmd.arg(0);
    TigrString posExamples=cmd.arg(1);
    TigrString negExamples=cmd.arg(2);
    TigrString consensusArg=cmd.arg(3);
    int numBins=cmd.arg(4).asInt();
    if(!filestemRegex.search(modelFilename)) 
      throw TigrString("Couldn't parse filestem from filename: ")+
	modelFilename;
    TigrString filestem=filestemRegex[1];

    // Load the model
    alphabet=DnaAlphabet::global;
    GarbageIgnorer GC;
    SignalSensor *model=SignalSensor::load(modelFilename,GC);
    TigrVector<TigrString> *consensuses=consensusArg.getFields(",");
    int numConsensuses=consensuses->size();
    for(int i=0 ; i<numConsensuses ; ++i)
      model->addConsensus((*consensuses)[i]);

    // Load the examples
    TigrVector<double> *posScores=loadPosExamples(posExamples,*model);
    TigrVector<double> *negScores=loadPosExamples(negExamples,*model);

    // Report summary stats
    TigrSummaryStats posStats(*posScores);
    TigrSummaryStats backgroundStats(*negScores);
    cout<<"positives:  "<<posStats.getMean()<<"+/-"<<posStats.getStdDev()<<
      " ("<<posStats.getMin()<<"-"<<posStats.getMax()<<")"<<endl;
    cout<<"background: "<<backgroundStats.getMean()<<"+/-"
	<<backgroundStats.getStdDev()<<" ("<<backgroundStats.getMin()
	<<"-"<<backgroundStats.getMax()<<")"<<endl;
    writeHistogramFile(*posScores,filestem+".pos-hist");
    writeHistogramFile(*negScores,filestem+".neg-hist");
    
    // Construct histograms
    double minScore=POSITIVE_INFINITY, maxScore=NEGATIVE_INFINITY;
    getExtrema(*posScores,minScore,maxScore);
    getExtrema(*negScores,minScore,maxScore);
    Histogram<double> posHist(minScore,maxScore,numBins,0.01);
    Histogram<double> backgroundHist(minScore,maxScore,numBins,1);
    posHist.addCounts(*posScores);
    backgroundHist.addCounts(*negScores);
    backgroundHist.addCounts(*posScores);

    // Compute log-likelihood ratios and write output file
    /*computeRatios(filestem,posHist,backgroundHist,minScore,maxScore,
      numBins);*/
    posHist.divideBy(backgroundHist);
    posHist.useLogs();
    TigrString outfile=filestem+".isp";
    posHist.save(outfile);

    return 0;
  }



TigrVector<double>* Application::loadPosExamples(const TigrString &filename,
						 SignalSensor &model)
{
  int consensusOffset=model.getConsensusOffset();
  int windowLen=model.getContextWindowLength();
  int consensusLen=model.getConsensusLength();
  TigrVector<double> *scores=new TigrVector<double>;
  TigrFastaReader reader(filename);
  TigrString def, seqStr;
  while(reader.nextSequence(def,seqStr))
    {
      Sequence seq(seqStr,DnaAlphabet::global);
      int seqLen=seqStr.length();
      int consensusBegin=(seqLen-consensusLen)/2;
      int windowBegin=consensusBegin-consensusOffset;
      double score=model.getLogP(seq,seqStr,windowBegin);
      if(isFinite(score)) scores->push_back(score);
    }

  return scores;
}



TigrVector<double>* Application::loadBackground(const TigrString &filename,
						SignalSensor &model)
{
  int consensusOffset=model.getConsensusOffset();
  int windowLen=model.getContextWindowLength();
  int consensusLen=model.getConsensusLength();
  TigrVector<double> *scores=new TigrVector<double>;
  TigrFastaReader reader(filename);
  TigrString def, seqStr;
  while(reader.nextSequence(def,seqStr))
    {
      Sequence seq(seqStr,DnaAlphabet::global);
      int seqLen=seqStr.length();
      int lastWindowPos=seqLen-windowLen;
      for(int pos=0 ; pos<lastWindowPos ; ++pos)
	if(model.consensusOccursAt(seqStr,pos+consensusOffset))
	  {
	    double score=model.getLogP(seq,seqStr,pos);
	    if(isFinite(score)) 
	      {
		scores->push_back(score);
	      }
	  }
    }

  return scores;
}



void Application::writeHistogramFile(TigrVector<double> &scores,
				     const TigrString &filename)
{
  ofstream os(filename.c_str());
  TigrVector<double>::iterator cur=scores.begin(), end=scores.end();
  for(; cur!=end ; ++cur)
    os<<*cur<<endl;
}



void Application::getExtrema(TigrVector<double> &scores,
			     double &min,double &max)
{
  TigrVector<double>::iterator cur=scores.begin(), end=scores.end();
  for(; cur!=end ; ++cur)
    {
      double score=*cur;
      if(score<min) min=score;
      if(score>max) max=score;
    }
}



void Application::computeRatios(const TigrString &filestem,
				Histogram<double> &posHist,
				Histogram<double> &backgroundHist,
				double minScore,
				double maxScore,int numBins)
{
  TigrString outfile=filestem+".sigrat";
  ofstream os(outfile.c_str());
  double binSize=(maxScore-minScore)/numBins;
  for(int i=0 ; i<numBins ; ++i)
    {
      double x=minScore+i*binSize;
      double numerator=posHist.getBin(i);
      double denominator=backgroundHist.getBin(i);
      double r=denominator ? numerator/denominator : 0;
      os<<x<<"\t"<<r<<"\t"<<posHist.getBin(i)<<"\t"<<backgroundHist.getBin(i)<<endl;
    }
  cout<<"Ratios written into "<<outfile<<endl;
}




