#!/usr/bin/perl
use strict;
use lib('/home/bmajoros/genomics/perl');
use GffTranscriptReader;
use FastaReader;

my $tigrscan="out";
my $cdna="chunks";
my $gffReader=new GffTranscriptReader;
my $tigrscanFiles=getFiles($tigrscan);
my $cdnaFiles=getFiles($cdna);
my $numTigrScan=@$tigrscanFiles;
my $numCdna=@$cdnaFiles;
my $totalBases;
my $totalCoding;
my @globalPredictions;
my @globalCdnas;
my $out;
my $seqLen;

my $numFiles=min($numTigrScan,$numCdna);

open(PIPE,"grep source-version $tigrscan/* |") || die;
my $line=<PIPE>;
close(PIPE);
$line=~/source-version\s*(\S+)/;
my $programName=$1;
pad(\$programName,9);

print STDERR "#files=$numFiles\n\n";
print STDERR "         |NUCLEOTIDE--|SPLICE-|ATG/TAG|EXONS--|EXACT\n";
print STDERR "         |SN  SP   ACC|SN  SP |SN  SP |SN  SP |GENES\n";

evaluateExons($programName,$tigrscanFiles,$numFiles,$cdnaFiles,$tigrscan,
	      $cdna);
evaluateNuc($programName,$tigrscanFiles,$numFiles,$cdnaFiles,$tigrscan,
	    $cdna);
evaluateGenes($programName,$tigrscanFiles,$numFiles,$cdnaFiles,$tigrscan,
	      $cdna);

#----------------------------------------------------------------
sub getFiles
  {
    my ($dir)=@_;
    my $files=`ls $dir`;
    my @files=split/\s+/,$files;
    @files=sort {$a=~/(\d+)/;my $a1=$1;$b=~/(\d+)/;$a1<=>$b} @files;
    my $n=@files;
    my @list;
    for(my $i=0 ; $i<$n ; ++$i)
      {
	my $file=$files[$i];
	next unless $file=~/(\d+)\.gff/;
	my $id=$1;
	push @list,$file;
	my $fasta="chunks/$id.fasta";
	$seqLen+=FastaReader::getGenomicSize($fasta);
      }
    return \@list;
  }
#----------------------------------------------------------------
sub min
  {
    my ($a,$b)=@_;
    return ($a<$b ? $a : $b);
  }
#----------------------------------------------------------------
sub max
  {
    my ($a,$b)=@_;
    return ($a>$b ? $a : $b);
  }
#----------------------------------------------------------------
sub evaluateNuc
  {
    my ($label)=@_;
    my ($totalTP, $totalTN, $totalFP, $totalFN);

    my $cdnaVector=makeVector(\@globalCdnas);
    my $predVector=makeVector(\@globalPredictions);

    my ($percentIdentity,$numErrors)=
      getPercentIdentity($predVector,$cdnaVector);
    my ($sensitivity,$specificity,$TP,$TN,$FP,$FN)=
      getSensAndSpecificity($predVector,$cdnaVector);

    $totalBases+=($TP+$TN+$FP+$FN);
    $totalCoding+=($TP+$FN);

    my $sensitivity=int(100*$TP/($TP+$FN)+0.5);
    my $specificity=int(100*$TP/($TP+$FP)+0.5);
    my $accuracy=int(100*($TP+$TN)/
		     ($TP+$TN+$FP+$FN)+0.5);
    pad(\$sensitivity,2);
    pad(\$specificity,3);
    pad(\$accuracy,2);
    print STDERR "$label|$sensitivity\% $specificity\% $accuracy\%";
    print STDERR "$out";
  }
#----------------------------------------------------------------
sub countIdenticalExons
  {
    my ($list1,$list2)=@_;
    my $identical;
    my $n1=@$list1;
    my $n2=@$list2;
    for(my $i=0 ; $i<$n1 ; ++$i)
      {
	my $exon1=$list1->[$i];
	for(my $j=0 ; $j<$n2 ; ++$j)
	  {
	    my $exon2=$list2->[$j];
	    if(exonsAreIdentical($exon1,$exon2)) 
	      {
		++$identical;
		$exon1->{matched}=1;
		$exon2->{matched}=1;
	      }
	  }
      }
    my $outOf=@$list1;
    my $outOf2=@$list2;
    return ($identical,$outOf,$outOf2);
  }
#----------------------------------------------------------------
sub exonsAreIdentical
  {
    my ($exon1,$exon2)=@_;
    return 
      $exon1->getBegin()==$exon2->getBegin() &&
      $exon1->getEnd()==$exon2->getEnd();
  }
#----------------------------------------------------------------
sub getExonList
  {
    my ($transcripts)=@_;
    my @list;
    foreach my $transcript (@$transcripts)
      {
	my $n=$transcript->numExons();
	for(my $i=0 ; $i<$n ; ++$i)
	  {
	    my $exon=$transcript->getIthExon($i);
	    push @list,$exon;
	  }
      }
    return \@list;
  }
#----------------------------------------------------------------
sub trimPredictions
  {
    # This function drops any prediction which does not overlap
    # a known gene -- we simply have no way of knowing whether the
    # prediction is right or wrong.

    my $n1=@globalPredictions;
    my $n2=@globalCdnas;
    my @keep;
    for(my $i=0 ; $i<$n1 ; ++$i)
      {
	my $prediction=$globalPredictions[$i];
	for(my $j=0 ; $j<$n2 ; ++$j)
	  {
	    my $cdna=$globalCdnas[$j];
	    if($prediction->overlaps($cdna)) 
	      {
		push @keep,$prediction;
		last;
	      }
	  }
      }
    @globalPredictions=@keep;
  }
#----------------------------------------------------------------
sub evaluateExons
  {
    my ($label,$files,$numFiles,$cdnaFiles,$programDir,$cdnaDir)=@_;
    my ($totalExons,$totalExonsCdna,$totalIdentical);
    my ($spliceTP,$spliceFP,$spliceFN);
    my ($codonTP,$codonFP,$codonFN);
    for(my $i=0 ; $i<$numFiles ; ++$i)
      {
	my $file=$files->[$i];
	my $predictions=$gffReader->loadGFF("$programDir/$file");
	next unless $predictions;
	push @globalPredictions,@$predictions;

	$file=~/(\d+)/ || next;
	my $cdnaFile="$1.gff";
	next unless -e "$cdnaDir/$cdnaFile" && -e "$programDir/$file";
	my $cDNAs=$gffReader->loadGFF("$cdnaDir/$cdnaFile");
	push @globalCdnas,@$cDNAs;
      }
    trimPredictions();
    
    my $cdnaExons=getExonList(\@globalCdnas);
    my $cdnaSplices=getSpliceSites($cdnaExons);
    my $cdnaCodons=getStartAndStopCodons($cdnaExons);
    
    my $predExons=getExonList(\@globalPredictions);
    my $predSplices=getSpliceSites($predExons);
    my $predCodons=getStartAndStopCodons($predExons);
    
    my ($identical,$outOf,$outOfCdna)=
      countIdenticalExons($predExons,$cdnaExons);
    $totalIdentical+=$identical;
    $totalExons+=$outOf;
    $totalExonsCdna+=$outOfCdna;
    
    my $commonSplices=
      countCommonSpliceSites($predSplices,$cdnaSplices);
    $spliceTP+=$commonSplices;
    $spliceFP+=(@$predSplices-$commonSplices);
    $spliceFN+=(@$cdnaSplices-$commonSplices);
    
    my $commonCodons=
      countCommonSpliceSites($predCodons,$cdnaCodons);
    $codonTP+=$commonCodons;
    $codonFP+=(@$predCodons-$commonCodons);
    $codonFN+=(@$cdnaCodons-$commonCodons);

    my $spliceSens=int(100*$spliceTP/($spliceTP+$spliceFN)+0.5);
    my $spliceSpec=int(100*$spliceTP/($spliceTP+$spliceFP)+0.5);

    my $codonSens=int(100*$codonTP/($codonTP+$codonFN)+0.5);
    my $codonSpec=int(100*$codonTP/($codonTP+$codonFP)+0.5);

    my $exonSpecificity=int(100*$totalIdentical/$totalExons+0.5);
    my $exonSensitivity=int(100*$totalIdentical/$totalExonsCdna+0.5);
    pad(\$spliceSens,2);
    pad(\$spliceSpec,2);
    pad(\$codonSens,2);
    pad(\$codonSpec,2);
    pad(\$exonSensitivity,2);
    pad(\$exonSpecificity,2);
    $out.="|$spliceSens\% $spliceSpec\%|$codonSens\% $codonSpec\%|$exonSensitivity\% $exonSpecificity\%";
  }
#----------------------------------------------------------------
sub evaluateGenes
  {
    my ($label,$files,$numFiles,$cdnaFiles,$programDir,$cdnaDir)=@_;
    my ($totalIdentical);
    my $numCdnaGenes=@globalCdnas;
    NEXTGENE:
      for(my $ic=0 ; $ic<$numCdnaGenes ; ++$ic)
	{
	  my $cdnaGene=$globalCdnas[$ic];
	  my $cdnaExons=$cdnaGene->{exons};
	  my $numExons=@$cdnaExons;
	  for(my $i=0 ; $i<$numExons ; ++$i) 
	    {
	      if(!$cdnaExons->[$i]->{matched}) {next NEXTGENE}
	    }
	  ++$totalIdentical;
	  next;
	}
    my $percent=int(100*$totalIdentical/$numCdnaGenes+0.5);
    pad(\$percent,2);
    print STDERR "|$percent\%\n";
  }
#----------------------------------------------------------------
sub getSpliceSites
  {
    my ($exons)=@_;
    my (@donors,@acceptors);
    foreach my $exon (@$exons)
      {
	my $exonType=$exon->getType();
	if($exonType eq "initial-exon" || $exonType eq "internal-exon")
	  {
	    push @donors,$exon->getEnd();
	  }
	if($exonType eq "final-exon" || $exonType eq "internal-exon")
	  {
	    push @acceptors,$exon->getBegin();
	  }
      }
    my @spliceSites;
    push @spliceSites,@donors;
    push @spliceSites,@acceptors;
    return \@spliceSites;
  }
#----------------------------------------------------------------
sub getStartAndStopCodons
  {
    my ($exons)=@_;
    my (@starts,@stops);
    foreach my $exon (@$exons)
      {
	my $exonType=$exon->getType();
	if($exonType eq "initial-exon" || $exonType eq "single-exon")
	  {
	    push @starts,$exon->getBegin();
	  }
	if($exonType eq "final-exon" || $exonType eq "single-exon")
	  {
	    push @stops,$exon->getEnd();
	  }
      }
    my @list;
    push @list,@starts;
    push @list,@stops;
    return \@list;
  }
#----------------------------------------------------------------
sub countCommonSpliceSites
  {
    my ($sites1,$sites2)=@_;
    my (%hash,$common);
    foreach my $site (@$sites1) {$hash{$site}=1}
    foreach my $site (@$sites2) {++$common if $hash{$site}}

    return $common;
  }
#-----------------------------------------------------------------
#my $leftTerminus=getLeftTerminus($correctFeatures);
sub getLeftTerminus
  {
    my ($features)=@_;
    my $left;
    foreach my $feature (@$features)
      {
	if(!defined($left) || $feature->getBegin()<$left)
	  { $left=$feature->getBegin() }
      }
    return $left;
  }
#-----------------------------------------------------------------
#my $rightTerminus=getRightTerminus($correctFeatures);
sub getRightTerminus
  {
    my ($features)=@_;
    my $right;
    foreach my $feature (@$features)
      {
	if(!defined($right) || $feature->getEnd()>$right)
	  { $right=$feature->getEnd() }
      }
    return $right;
  }
#----------------------------------------------------------------
#dropSuperfluousFeatures($predictions,$leftTerminus,$rightTerminus);
sub dropSuperfluousFeatures
  {
    my ($predictions,$leftBoundary,$rightBoundary)=@_;
    my @retval;
    my $n=@$predictions;
    for(my $i=0 ; $i<$n ; ++$i)
      {
	my $feature=$predictions->[$i];
	my $begin=$feature->getBegin();
	my $end=$feature->getEnd();
	if($end<=$leftBoundary || $begin>=$rightBoundary) {next}
	push @retval,$feature;

	if(0)
	  {
	    my $ne=$feature->numExons();
	    for(my $i=0 ; $i<$ne ; ++$i)
	      {
		my $exon=$feature->getIthExon($i);
		my $begin=$exon->getBegin();
		my $end=$exon->getEnd();
		if($end<=$leftBoundary || $begin>=$rightBoundary)
		  {
		    $feature->deleteExon($i);
		    --$i;
		    --$ne;
		  }
	      }
	  }
      }
    @$predictions=@retval;
  }
#----------------------------------------------------------------
sub pad
  {
    my ($ptr,$len)=@_;
    my $l=length $$ptr;
    my $extra=$len-$l;
    my $padding=' 'x$extra;
    $$ptr.=$padding;
  }
#----------------------------------------------------------------
#----------------------------------------------------------------
#----------------------------------------------------------------
#----------------------------------------------------------------
#----------------------------------------------------------------
#----------------------------------------------------------------
#----------------------------------------------------------------
#----------------------------------------------------------------
#----------------------------------------------------------------

#-----------------------------------------------------------------
sub max
  {
    my ($a,$b)=@_;
    return ($a>$b ? $a : $b);
  }
#-----------------------------------------------------------------
sub getPercentIdentity
  {
    my ($vector1,$vector2)=@_;
    my $identity;
    my $length;
    for(my $i=0 ; $i<$seqLen ; ++$i)
      {
	my $elem1=$vector1->[$i];
	my $elem2=$vector2->[$i];
	next unless defined($elem1) || defined($elem2);
	if($elem1==$elem2){++$identity}
	++$length;
      }
    return ($identity/$length,$length-$identity);
  }
#-----------------------------------------------------------------
sub getSensAndSpecificity
  {
    my ($hypoth,$correct)=@_;
    my ($TP,$TN,$FP,$FN)=(0,0,0,0);
    for(my $i=0 ; $i<$seqLen ; ++$i)
      {
	my $hypothElem=$hypoth->[$i];
	my $correctElem=$correct->[$i];
	next unless defined($hypothElem) || defined($correctElem);
	if($hypothElem==$correctElem)
	  {
	    if($correctElem==1){++$TP}
	    else {++$TN}
	  }
	else
	  {
	    if($correctElem==1){++$FN}
	    else {++$FP}
	  }
      }
    my $sensitivity=($TP+$FN>0 ? $TP/($TP+$FN) : 1);
    my $specificity=($TP+$FP>0 ? $TP/($TP+$FP) : 1);
    return ($sensitivity,$specificity,$TP,$TN,$FP,$FN);
  }
#-----------------------------------------------------------------
sub makeVector
  {
    my ($transcripts)=@_;
    my $vector=[];
    my $n=@$transcripts;
    for(my $i=0 ; $i<$n ; ++$i)
      {
	my $transcript=$transcripts->[$i];
	my $exons=$transcript->{exons};
	my $n=@$exons;
	my $prevEnd;
	for(my $i=0 ; $i<$n ; ++$i)
	  {
	    my $exon=$exons->[$i];
	    my $begin=$exon->getBegin();
	    my $end=$exon->getEnd();
	    if(defined($prevEnd))
	      {
		for(my $j=$prevEnd ; $j<$begin ; ++$j)
		  {
		    $vector->[$j]=0;
		  }
	      }
	    for(my $j=$begin ; $j<$end ; ++$j)
	      {
		$vector->[$j]=1;
	      }
	    $prevEnd=$end;
	  }
      }
    return $vector;
  }
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
#-----------------------------------------------------------------
