#!/usr/bin/perl
use strict;
use FastaReader;

my $usage="
$0 <*.fasta> <bin-size> <smoothing-window-size> <smoothing-iterations>

  set <bin-size> to 0 to auto-detect

";
die "$usage\n" unless @ARGV==4;
my ($filename,$binSize,$windowSize,$numIterations)=@ARGV;

$filename=~/(.*).fasta/ || die "$filename must have a .fasta extension\n";
my $outfile="$1.distr";

# Read fasta file & collect lengths
my (%counts,$total,@stack,$max);
my $reader=new FastaReader($filename);
while(1)
  {
    my ($defline,$sequence)=$reader->nextSequence();
    last unless defined $defline;
    my $length=length($sequence);
    push @stack,$length;
    if(!defined($max) || $length>$max) {$max=$length}
  }
my $n=@stack;

# Build histogram
my (@histogram,$largestBin,$total);
foreach my $x (@stack)
  {
    my $bin=int($x/$binSize);
    ++@histogram[$bin];
    if($bin>$largestBin){$largestBin=$bin}
    ++$total;
  }

# Normalize histogram & use set of pairs
for(my $i=0 ; $i<=$largestBin ; ++$i) 
  {
    my $count=$histogram[$i];
    my $p=$count/$total;
    $histogram[$i]=[$i*$binSize,$p];
  }

# Smooth histogram
for(my $i=0 ; $i<$numIterations ; ++$i)
  {
    @histogram=@{smooth($windowSize,\@histogram)};
  }

# Compute smoothness
my (@diffs,$prevDiff,$reversals);
for(my $i=1 ; $i<$n ; ++$i)
  {
    next unless $histogram[$i] && $histogram[$i-1];
    my ($thisX,$thisValue)=@{$histogram[$i]};
    my ($prevX,$prevValue)=@{$histogram[$i-1]};
    my $diff=$thisValue-$prevValue;
    #print "$prevValue $thisValue $diff\n";
    push @diffs,$diff;
    if(defined($prevDiff))
      {
	my $prevSign=sign($prevDiff);
	my $sign=sign($diff);
	if($prevSign!=$sign) {++$reversals}
      }
    $prevDiff=$diff;
  }
my $smoothness=($reversals ? 1/$reversals : 1);
print STDERR "SMOOTHNESS: $smoothness\n";

# Output histogram
for(my $i=0 ; $i<=$largestBin ; ++$i)
  {
    #my $x=$i*$binSize;
    #my $y=$histogram[$i];
    my $pair=$histogram[$i];
    next unless defined $pair;
    my ($x,$y)=@$pair;
    print "$x $y\n";
  }




#-------------------------------------------------
sub sign
  {
    my ($x)=@_;
    return ($x<0 ? -1 : 1);
  }
sub smooth
  {
    my ($windowSize,$histogram)=@_;
    my $n=@$histogram;
    --$windowSize;
    my $halfWindow=int($windowSize/2);
    my $otherHalf=$windowSize-$halfWindow; # in case it's an odd number
    my $first=$halfWindow;
    my $last=$n-1-$otherHalf;
    my $newHistogram=[];
    for(my $i=0 ; $i<$first ; ++$i) {$newHistogram->[$i]=$histogram->[$i]}
    for(my $i=$last+1 ; $i<$n ; ++$i) {$newHistogram->[$i]=$histogram->[$i]}
    for(my $i=$first ; $i<=$last ; ++$i)
      {
	my $pair=$histogram->[$i];
	my ($x,$y)=@$pair;
	for(my $j=0 ; $j<$halfWindow ; ++$j)
	  {
	    my $debug=$i-1-$j;
	    my ($leftX,$leftY)=@{$histogram->[$i-1-$j]};
	    $y+=$leftY;
	  }
	for(my $j=0 ; $j<$otherHalf ; ++$j)
	  {
	    my $pair=$histogram->[$i+1+$j];
	    my ($rightX,$rightY)=(defined($pair) ? @$pair : (0,0));
	    $y+=$rightY;
	  }
	$y/=($windowSize+1);
	$newHistogram->[$i]=[$x,$y];
      }
    return $newHistogram;
  }
