
// refine-tata.C

#include <string>
#include <iostream>
#include "tigr++/TigrCommandLine.H"
#include "tigr++/TigrGffReader.H"
#include "tigr++/TigrFastaReader.H"
#include "tigr++/TigrVectorSorter.H"
#include "tigr++/DnaAlphabet.H"
#include "tigr++/Constants.H"
#include "tigr++/TigrProteinTrans.H"
#include "TrainingSequence.H"
#include "SignalSensor.H"
#include "GarbageCollector.H"
#include "WAM.H"
#include "WMM.H"

#define CONSENSUS_OFFSET 5
#define CONSENSUS_LENGTH 6

Alphabet alphabet; // required for linking in Partition
typedef TigrGffReader::TranscriptList TranscriptList;


TigrSet<TigrString> consensuses;
const char *consensusStr[]=
{
  "TATAAA",
  "TATATA",
  "CATAAA", // ### might want to comment this out
  "CATATA"  // ### might want to comment this out
};


class ScoredSequence : public TrainingSequence
{
  double score;
public:
  ScoredSequence(const TigrString &s,Alphabet &a,double score)
    : TrainingSequence(s,a), score(score) {}
  double getScore() {return score;}
};


struct ScoredSeqComp : TigrComparator<TrainingSequence*>
{
  bool equal(TrainingSequence *&a,TrainingSequence *&b)
  {return 
     static_cast<ScoredSequence*>(a)->getScore() ==
     static_cast<ScoredSequence*>(b)->getScore();}
  bool greater(TrainingSequence *&a,TrainingSequence *&b)
  {return 
     static_cast<ScoredSequence*>(a)->getScore() >
     static_cast<ScoredSequence*>(b)->getScore();}
  bool less(TrainingSequence *&a,TrainingSequence *&b)
  {return 
     static_cast<ScoredSequence*>(a)->getScore() <
     static_cast<ScoredSequence*>(b)->getScore();}
};


class Application
{
  int margin;
  TigrVector<TigrString> margins;
  SignalSensor *currentModel;
  int modelLength;
  GarbageIgnorer gc;

  void processContigs(const TigrString &contigsFile,
		      TigrMap<TigrString,TranscriptList*> &transcripts);
  void processForwardFeature(int featureBegin,const TigrString &seq);
  void processReverseFeature(int featureEnd,const TigrString &seq);
  void updateModel(bool echo,float keepFraction);
public:
  Application();
  int main(int argc,char *argv[]);
};


int main(int argc,char *argv[])
  {
    try
      {
	Application app;
	return app.main(argc,argv);
      }
    catch(const char *p)
      {
	cerr << p << endl;
      }
    catch(const string &msg)
      {
	cerr << msg.c_str() << endl;
      }
    catch(const exception &e)
      {
	cerr << "STL exception caught in main:\n" << e.what() << endl;
      }
    catch(...)
      {
	cerr << "Unknown exception caught in main" << endl;
      }
    return -1;
  }



Application::Application()
  {
    // ctor
  }



int Application::main(int argc,char *argv[])
  {
    // Process command line
    TigrCommandLine cmd(argc,argv,"");
    if(cmd.numArgs()!=7)
      throw string("\n\
refine-tata <seed-model> <genes.gff> <contigs.fasta>\n\
             <outfile> <margin> <iterations> <percentile>\n\
");
    TigrString seedModelFile=cmd.arg(0);
    TigrString genesFile=cmd.arg(1);
    TigrString contigsFile=cmd.arg(2);
    TigrString outFile=cmd.arg(3);
    margin=cmd.arg(4).asInt();
    int iterations=cmd.arg(5).asInt();
    float keepFraction=cmd.arg(6).asFloat();

    // Initialization
    alphabet=DnaAlphabet::global;
    int numCons=sizeof(consensusStr)/sizeof(consensusStr[0]);
    for(int i=0 ; i<numCons ; ++i) consensuses.insert(consensusStr[i]);

    // Load GFF
    TigrGffReader gffReader(genesFile);
    TigrMap<TigrString,TranscriptList*> &transcripts=
      *gffReader.loadByContig();

    // Load poly-A model
    currentModel=SignalSensor::load(seedModelFile,gc);
    modelLength=currentModel->getContextWindowLength();

    // Process each contig to extract regions following genes
    processContigs(contigsFile,transcripts);

    // Iteratively process the margins to search for best hits to 
    // current model
    for(int i=0 ; i<iterations ; ++i)
      updateModel(i==iterations-1,keepFraction);

    // Output refined model
    currentModel->save(outFile);

    delete &transcripts;
    return 0;
  }



void Application::processContigs(const TigrString &contigsFile,
	 TigrMap<TigrString,TranscriptList*> &transcriptMap)
{
  TigrFastaReader fastaReader(contigsFile);
  TigrString def, seq;
  while(fastaReader.nextSequence(def,seq))
    {
      TigrString substrate=TigrFastaReader::getId(def);
      if(transcriptMap.isDefined(substrate))
	{
	  TranscriptList &transcripts=*transcriptMap[substrate];
	  int n=transcripts.size();
	  for(int i=0 ; i<n ; ++i)
	    {
	      TigrGffTranscript *feature=transcripts[i];
	      switch(feature->getStrand())
		{
		case '+':
		  processForwardFeature(feature->getBegin(),seq);
		  break;
		case '-':
		  processReverseFeature(feature->getEnd(),seq);
		  break;
		default:
		  continue;
		}
	    }
	}
    }
}



void Application::processForwardFeature(int featureBegin,
					const TigrString &seq)
{
  int end=featureBegin;
  int begin=featureBegin-margin;
  if(begin<0) begin=0;
  TigrString subseq=seq.substr(begin,end-begin);
  margins.push_back(subseq);
}



void Application::processReverseFeature(int featureEnd,
					const TigrString &seq)
{
  int begin=featureEnd;
  int end=begin+margin;
  int len=seq.length();
  if(end>=len) end=len-1;
  TigrString subseq=
    TigrProteinTrans::reverseComplement(seq.substr(begin,end-begin));
  margins.push_back(subseq);
}



void Application::updateModel(bool echo,float keepFraction)
{
  TigrVector<TrainingSequence*> windows;
  TigrVector<double> scores;
  int numMargins=margins.size();
  for(int i=0 ; i<numMargins ; ++i)
    {
      TigrString &margin=margins[i];
      TrainingSequence marginSeq(margin,DnaAlphabet::global);
      int len=margin.length();
      int numPositions=len-modelLength;
      int bestPos;
      double bestScore=NEGATIVE_INFINITY;
      for(int pos=0 ; pos<numPositions ; ++pos)
	{
	  TigrString consensus=
	    margin.substring(pos+CONSENSUS_OFFSET,CONSENSUS_LENGTH);
	  if(!consensuses.isMember(consensus)) continue;
	  double score=currentModel->getLogP(marginSeq,margin,pos);
	  if(score>=bestScore)
	    {
	      bestScore=score;
	      bestPos=pos;
	    }
	}
      if(bestScore==NEGATIVE_INFINITY) continue;



      // ### THESE TWO LINES CAUSE A LATER SEG FAULT...FIND OUT WHY!!!!
      //TrainingSequence *window=new TrainingSequence;
      //marginSeq.getSubsequence(bestPos,modelLength,*window);



      TrainingSequence *window=
	new ScoredSequence(margin.substring(bestPos,modelLength),
			   alphabet,bestScore);

      windows.push_back(window);
    }

  // Sort windows by score
  ScoredSeqComp comparator;
  TigrVectorSorter<TrainingSequence*> sorter(windows,comparator);
  sorter.sortDescendInPlace();

  // Keep only the top X% for retraining
  int numWindows=windows.size();
  int keepNum=int(keepFraction*numWindows);
  windows.resize(keepNum);

  // Retrain model
  cerr <<"retraining on "<<windows.size()<<" sequences..."<<endl;
  delete currentModel;
  //currentModel=new WAM(gc,windows,2,40,POLYA,0,1);
  currentModel=new WMM(gc,windows,POLYA,CONSENSUS_OFFSET,CONSENSUS_LENGTH);

  // Delete training strings
  cerr <<"deleting strings..."<<endl;
  TigrVector<TrainingSequence*>::iterator cur=windows.begin(), 
    end=windows.end();
  for(; cur!=end ; ++cur) delete *cur;
}


