/**************************************************************
test-signal-sensor.C
bmajoros@tigr.org

Copyright (c) 2003, The Institute for Genomic Research (TIGR),
Rockville, Maryland, U.S.A.  All rights reserved.
***************************************************************/

#include <iostream>
#include <fstream>
#include <math.h>
#include "tigr++/TigrString.H"
#include "tigr++/TigrCommandLine.H"
#include "tigr++/TigrVector.H"
#include "tigr++/TigrFastaReader.H"
#include "tigr++/TigrRegex.H"
#include "tigr++/Alphabet.H"
#include "tigr++/DnaAlphabet.H"
#include "ScoreAnalyzer.H"
#include "ModelBuilder.H"
#include "SignalType.H"
#include "TrainingSequence.H"
#include "GarbageCollector.H"

Alphabet alphabet;
int frame=-1; // ### CAUTION: this is required by older code; to be removed

class Application
{
  GarbageCollector gc;
  ModelType modelType;
  int order, minSampleSize, windowSize;
  TigrVector<TrainingSequence*> posTrain, negTrain, posTest, negTest;
  SignalSensor *model;
  float trainTestRatio;
  TigrRegex frameRegex;
  TigrVector<double> posScores, negScores;

  void evaluate(TigrVector<TrainingSequence*> &,TigrVector<double> &scores);
  void load(TigrString filename,TigrVector<TrainingSequence*> &);
  void updateBoostCounts(TigrVector<TrainingSequence*> &seqs,
			 TigrVector<double> &scores,double cutoff);
public:
  Application();
  int main(int argc,char *argv[]);
};


int main(int argc,char *argv[])
  {
    try
      {
	Application app;
	return app.main(argc,argv);
      }
    catch(const char *p)
      {
	cerr << p << endl;
      }
    catch(const TigrString &msg)
      {
	cerr << msg.c_str() << endl;
      }
    catch(const exception &e)
      {
	cerr << "STL exception caught in main:\n" << e.what() << endl;
      }
    catch(...)
      {
	cerr << "Unknown exception caught in main" << endl;
      }
    return -1;
  }



Application::Application()
  : order(5), minSampleSize(175), model(NULL),
    windowSize(5), frameRegex("/frame=(\\d+)")
  {
    // ctor
  }



int Application::main(int argc,char *argv[])
  {
    // ====================
    // Process command line
    // ====================

    TigrCommandLine cmd(argc,argv,"o:s:gr::b:");
    if(cmd.numArgs()!=4)
      throw TigrString("\ntest-signal-sensor <*.model> <pos.fasta> <neg.fasta> <outfile>\n");
    TigrString modelFilename=cmd.arg(0);
    TigrString posFilename=cmd.arg(1);
    TigrString negFilename=cmd.arg(2);
    TigrString outFilestem=cmd.arg(3);
    alphabet=DnaAlphabet::global;

    // =========================
    // Load model
    // =========================
    model=SignalSensor::load(modelFilename,gc);

    // =========================
    // Load test data
    // =========================
    cerr << "loading data..." << endl;
    TigrVector<TrainingSequence*> posExamples, negExamples;
    load(posFilename,posExamples);
    int numPosExamples=posExamples.size();
    load(negFilename,negExamples);
    int numNegExamples=negExamples.size();

    // =========================
    // Apply model
    // =========================
    cout << "testing model on " << posExamples.size() 
	 << " positive examples" << endl;
    evaluate(posExamples,posScores);
    cout << "testing model on " << negExamples.size()
	 << " negative examples" << endl;
    evaluate(negExamples,negScores);
    cout<<posScores.size()<<" pos scores, "<<negScores.size()<<" neg scores"<<endl;

    // =========================
    // Write out .scores file
    // =========================
    cout << "writing .scores file" << endl;
    ofstream os((outFilestem+".scores").c_str());
    TigrVector<double>::iterator cur=posScores.begin(), 
      end=posScores.end();
    
    double minScore=*cur;
    for(; cur!=end ; ++cur) 
      if(!isinf(*cur) && *cur<minScore) minScore=*cur;
    cur=negScores.begin(); end=negScores.end();
    for(; cur!=end ; ++cur) 
      if(!isinf(*cur) && *cur<minScore) minScore=*cur;
    
    cur=posScores.begin(); end=posScores.end();	
    for(; cur!=end ; ++cur) 
      os << "1\t" << (isinf(*cur)?minScore:*cur) << endl;
    os << endl;
    cur=negScores.begin(); end=negScores.end();
    for(; cur!=end ; ++cur) 
      os << "2\t" << (isinf(*cur)?minScore:*cur) << endl;

    // =========================
    // Report accuracy
    // =========================
    ScoreAnalyzer analyzer(posScores,negScores);
    double cutoff=model->getCutoff();//analyzer.getMinRecallCutoff(minSensitivity);
    double prec, rec, percentFP, percentFN;
    analyzer.lookupCutoff(cutoff,prec,rec,percentFP,percentFN);
    updateBoostCounts(posTest,posScores,analyzer.getPercentileCutoff(0.8));
    cout<<"cutoff="<<cutoff<<endl;
    model->setCutoff(cutoff);
    cout << "accuracy="
	 << analyzer.getBestAccuracy() 
	 << " precision=" << prec << " FP="
	 <<int(percentFP*100+0.5)<<"% FN="
	 <<int(percentFN*100+0.5)<<"%"
	 <<endl;
    
    // ==============================================
    // Write out precision-recall graph
    // ==============================================
    cout << "writing precision-recall graph" << endl;
    ofstream os2((outFilestem+".prec-recall").c_str());
    cout << "writing to file " 
	 << (outFilestem+".prec-recall").c_str()
	 <<endl;
    analyzer.outputGraph(os2);
    
    cout << "done." << endl;
    return 0;
  }



void Application::load(TigrString filename,TigrVector<TrainingSequence*> &v)
{
  TigrFastaReader reader(filename);
  TigrString defline, sequence;
  while(reader.nextSequence(defline,sequence))
    {
      if(frameRegex.search(defline))
	frame=frameRegex[1].asInt();
      else
	frame=0;
      TrainingSequence *s=new TrainingSequence(sequence,alphabet);
      s->setPhase(frame);
      v.push_back(s);
    }
}



void Application::evaluate(TigrVector<TrainingSequence*> &sequences,
			   TigrVector<double> &scores)
{
  int n=sequences.size();
  for(int i=0 ; i<n ; ++i)
    {
      TrainingSequence *seq=sequences[i];
      TigrString *str=seq->toString(alphabet);
      frame=sequences[i]->getPhase();
      double score=model->getLogP(*seq,str->c_str(),0);
      delete str;
      scores.push_back(score);
    }
}



void Application::updateBoostCounts(TigrVector<TrainingSequence*> &seqs,
				    TigrVector<double> &scores,
				    double cutoff)
{
  int n=seqs.size();
  for(int i=0 ; i<n ; ++i)
    if(scores[i]<cutoff) 
      seqs[i]->adjustBoostCount(1);
}


