/**************************************************************
train-signal-sensor.C
bmajoros@tigr.org

Copyright (c) 2003, The Institute for Genomic Research (TIGR),
Rockville, Maryland, U.S.A.  All rights reserved.
***************************************************************/

#include <iostream>
#include <fstream>
#include <math.h>
#include "tigr++/TigrString.H"
#include "tigr++/TigrCommandLine.H"
#include "tigr++/TigrVector.H"
#include "tigr++/TigrFastaReader.H"
#include "tigr++/TigrRegex.H"
#include "tigr++/Alphabet.H"
#include "tigr++/DnaAlphabet.H"
#include "ScoreAnalyzer.H"
#include "ModelBuilder.H"
#include "SignalType.H"
#include "TrainingSequence.H"
#include "GarbageCollector.H"

Alphabet alphabet;
int frame=-1; // ### CAUTION: this is required by older code; to be removed

class Application
{
  GarbageCollector gc;
  ModelType modelType;
  int order, minSampleSize, windowSize;
  TigrVector<TrainingSequence*> posTrain, negTrain, posTest, negTest;
  SignalSensor *model, *negModel;
  float trainTestRatio;
  TigrRegex frameRegex;

  void evaluate(TigrVector<TrainingSequence*> &,TigrVector<double> &scores);
  void load(TigrString filename,TigrVector<TrainingSequence*> &);
  void updateBoostCounts(TigrVector<TrainingSequence*> &seqs,
			 TigrVector<double> &scores,double cutoff);
public:
  Application();
  int main(int argc,char *argv[]);
};


int main(int argc,char *argv[])
  {
    try
      {
	Application app;
	return app.main(argc,argv);
      }
    catch(const char *p)
      {
	cerr << p << endl;
      }
    catch(const TigrString &msg)
      {
	cerr << msg.c_str() << endl;
      }
    catch(const exception &e)
      {
	cerr << "STL exception caught in main:\n" << e.what() << endl;
      }
    catch(...)
      {
	cerr << "Unknown exception caught in main" << endl;
      }
    return -1;
  }



Application::Application()
  : order(5), minSampleSize(175), model(NULL),
    negModel(NULL), windowSize(5), frameRegex("/frame=(\\d+)")
  {
    // ctor
  }



int Application::main(int argc,char *argv[])
  {
    // ====================
    // Process command line
    // ====================

    TigrCommandLine cmd(argc,argv,"o:s:gr::b:");
    if(cmd.numArgs()!=9)
      throw TigrString("\n\
train-signal-sensor -g [-o order] [-s MinSampleSize] [-r <null.model>]\n\
   WMM|WAM|WWAM <pos.fasta> <neg.fasta> <output-filestem>\n\
   <%train> <signal-type> <consensus-offset> <consensus-length>\n\
   <min-sensitivity> [-b boosting-threshold]\n\
where -e evaluates performance of model using the given negative examples\n\
      -g outputs a precision-recall graph (requires -e)\n\
      -o sets the Markov order for MC, WAM, or 3P\n\
      -r causes model to use log-odds ratios instead of log-probabilities\n\
         (the model file is optional; w/o it we use the model's rev. comp.)\n\
      WMM = position-specific weight matrix\n\
      WAM = weight array matrix\n\
      WWAM = windowed weight array matrix\n\
      <%train> = portion of training set to use for training (vs. testing)\n\
                 (example: 0.90 = hold out 10% for testing)\n\
      <signal-type> = ATG|TAG|GT|AG|PROMOTER|POLYA\n\
      <min-sensitivity> = (for example) 0.98\n\
\n");
    modelType=stringToModelType(cmd.arg(0));
    TigrString posFilename=cmd.arg(1);
    TigrString negFilename=cmd.arg(2);
    TigrString outFilestem=cmd.arg(3);
    trainTestRatio=cmd.arg(4).asFloat();
    if(cmd.option('o')) order=cmd.optParm('o').asInt();
    if(cmd.option('s')) minSampleSize=cmd.optParm('s').asInt();
    alphabet=DnaAlphabet::global;
    SignalType signalType=stringToSignalType(cmd.arg(5));
    int consensusOffset=cmd.arg(6).asInt();
    int consensusLength=cmd.arg(7).asInt();
    float minSensitivity=cmd.arg(8).asFloat();

    // =========================
    // Load training & test data
    // =========================

    cerr << "loading data..." << endl;
    TigrVector<TrainingSequence*> posExamples, negExamples;
    load(posFilename,posExamples);
    int numPosExamples=posExamples.size();
    int numPosTrain=int(numPosExamples*trainTestRatio);
    int numPosTest=numPosExamples-numPosTrain;
    if(trainTestRatio<1)
      {
	if(cmd.option('b')) throw "-b requires that %train = 1.0";
	posExamples.getSubrange(0,numPosTrain-1,posTrain);
	posExamples.getSubrange(numPosTrain,numPosExamples-1,posTest);
      }
    else
      {
	cout << "Warning -- using training set for testing." << endl;
	posExamples.getSubrange(0,numPosExamples-1,posTrain);
	posExamples.getSubrange(0,numPosExamples-1,posTest);
      }
    
    cerr << "loading negative examples..." << endl;
    load(negFilename,negExamples);
    int numNegExamples=negExamples.size();
    int numNegTrain=int(numNegExamples*trainTestRatio);
    int numNegTest=numNegExamples-numNegTrain;
    if(trainTestRatio<1)
      {
	negExamples.getSubrange(0,numNegTrain-1,negTrain);
	negExamples.getSubrange(numNegTrain,numNegExamples-1,negTest);
      }
    else
      {
	negExamples.getSubrange(0,numNegExamples-1,negTrain);
	negExamples.getSubrange(0,numNegExamples-1,negTest);
      }
    
    int numIterations=1;
    if(cmd.option('b')) numIterations=cmd.optParm('b').asInt();
    for(int i=0 ; i<numIterations ; ++i)
      {
	delete model;

	TigrVector<double> posScores, negScores;

	// =======================
	// Build & train the model
	// =======================
	
	cerr << "training model on " << posTrain.size() 
	     << " positive examples" << endl;
	ModelBuilder modelBuilder(gc,alphabet,minSampleSize,order,
				  windowSize);
	model=modelBuilder.buildSignalSensor(modelType,posTrain,signalType,
					     consensusOffset,
					     consensusLength);
	
	// =============================================================
	// Test the model and establish an acceptable cutoff (threshold)
	// =============================================================
	
	if(cmd.option('r')) 
	  {
	    cerr << "training negative model on " << negTrain.size()
		 << " negative examples" << endl;
	    negModel=modelBuilder.buildSignalSensor(modelType,negTrain,
						    signalType,
						    consensusOffset,
						    consensusLength);
	
	    cerr << "using log odds ratios instead of probabilities" 
		 << endl;
	    TigrString filename=cmd.optParm('r');
	    if(filename=="")
	      model->useLogOdds(*negModel);
	    else
	      {
		ContentSensor *nullModel=ContentSensor::load(filename);
		model->useLogOdds_anonymous(*nullModel);
		delete nullModel;
	      }
	  }
	
	cout << "testing model on " << posTest.size() 
	     << " positive examples" << endl;
	evaluate(posTest,posScores);
	cout << "testing model on " << negTest.size()
	     << " negative examples" << endl;
	evaluate(negTest,negScores);
	cout<<posScores.size()<<" pos scores, "<<negScores.size()
	    <<" neg scores"<<endl;

	if(i==numIterations-1)
	  {
	    cout << "writing .scores file" << endl;
	    ofstream os((outFilestem+".scores").c_str());
	    TigrVector<double>::iterator cur=posScores.begin(), 
	      end=posScores.end();

	    double minScore=*cur;
	    for(; cur!=end ; ++cur) 
	      if(!isinf(*cur) && *cur<minScore) minScore=*cur;
	    cur=negScores.begin(); end=negScores.end();
	    for(; cur!=end ; ++cur) 
	      if(!isinf(*cur) && *cur<minScore) minScore=*cur;

	    cur=posScores.begin(); end=posScores.end();	
	    for(; cur!=end ; ++cur) 
	      os << "1\t" << (isinf(*cur)?minScore:*cur) << endl;
	    os << endl;
	    cur=negScores.begin(); end=negScores.end();
	    for(; cur!=end ; ++cur) 
	      os << "2\t" << (isinf(*cur)?minScore:*cur) << endl;
	  }

	ScoreAnalyzer analyzer(posScores,negScores);
	double cutoff=analyzer.getMinRecallCutoff(minSensitivity);
	double prec, rec, percentFP, percentFN;
	analyzer.lookupCutoff(cutoff,prec,rec,percentFP,percentFN);
	updateBoostCounts(posTest,posScores,
			  analyzer.getPercentileCutoff(0.8));
	cout<<"cutoff="<<cutoff<<endl;
	model->setCutoff(cutoff);
	cout << "************************************* Iteration #" 
	     << i << ": accuracy="
	     << analyzer.getBestAccuracy() 
	     << " precision=" << prec << " FP="
	     <<int(percentFP*100+0.5)<<"% FN="
	     <<int(percentFN*100+0.5)<<"%"
	     <<endl;

	if(i==numIterations-1)
	  {
	    // ==============================================
	    // Write out precision-recall graph, if requested
	    // ==============================================
	    
	    if(cmd.option('g'))
	      {
		cout << "writing precision-recall graph" << endl;
		ofstream os((outFilestem+".prec-recall").c_str());
		cout << "writing to file " 
		     << (outFilestem+".prec-recall").c_str()
		     <<endl;
		analyzer.outputGraph(os);
	      }
	  }
      }

    // ==============
    // Save the model
    // ==============

    cout << "saving model" << endl;
    model->save(TigrString(outFilestem+".model"));

    cout << "done." << endl;
    return 0;
  }



void Application::load(TigrString filename,TigrVector<TrainingSequence*> &v)
{
  TigrFastaReader reader(filename);
  TigrString defline, sequence;
  while(reader.nextSequence(defline,sequence))
    {
      if(frameRegex.search(defline))
	frame=frameRegex[1].asInt();
      else
	frame=0;
      TrainingSequence *s=new TrainingSequence(sequence,alphabet);
      s->setPhase(frame);
      v.push_back(s);
    }
}



void Application::evaluate(TigrVector<TrainingSequence*> &sequences,
			   TigrVector<double> &scores)
{
  int n=sequences.size();
  for(int i=0 ; i<n ; ++i)
    {
      TrainingSequence *seq=sequences[i];
      TigrString *str=seq->toString(alphabet);
      frame=sequences[i]->getPhase();
      double score=model->getLogP(*seq,str->c_str(),0);
      //double negModelScore=negModel->getLogP(*seq,str->c_str(),0);
      //###score-=negModelScore;
      delete str;
      scores.push_back(score);
    }
}



void Application::updateBoostCounts(TigrVector<TrainingSequence*> &seqs,
				    TigrVector<double> &scores,
				    double cutoff)
{
  int n=seqs.size();
  for(int i=0 ; i<n ; ++i)
    if(scores[i]<cutoff) 
      seqs[i]->adjustBoostCount(1);
}


